//-*-c++-*------------------------------------------------------------
//
// File name : newbiogeme.cc
// Author :    Michel Bierlaire and Mamy Fetiarison
// Date :      Wed Apr 29 14:22:58 2009
//
//--------------------------------------------------------------------


#include "bioModel.h"
#include "patError.h"
#include "patDisplay.h"
#include <Python.h>
#include "bioVersion.h"

#include "patAbsTime.h"
#include "patFileNames.h"
#include "bioSample.h"
#include "bioArithSum.h"
#include "bioMetaIterator.h"
#include "bioRowIterator.h"
#include "bioIteratorInfoRepository.h"
#include "bioGenerateLikelihoodFunction.h"
#include "bioGenerateSimulation.h"
#include "bioGenerateStatistics.h"
#include "bioModelParser.h"

#include "bioRandomDraws.h"

int main(int argc, char *argv[]) {
  cout << "~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~" << endl ;
  cout << bioVersion::the()->getVersionInfoDate() << endl ;
  cout << "Michel Bierlaire, EPFL" << endl ;
  cout << bioVersion::the()->getVersionInfoCompiled() << endl ;
  cout << "See http://biogeme.epfl.ch" << endl  ;
  cout << "~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~" << endl  ;
  cout << "\t\"In every non-trivial program there is at least one bug.\"" << endl  ;
  cout << "" << endl  ;

  patAbsTime now ;
  patError *err(NULL) ;
  bioExpression* theFormula ;
  bioModel *theModel ;
  bioModelParser *parser ;
  bioSample* samples ;
  patString filename ;

  if (argc != 3 ) {
    std::cout << "Usage : " << argv[0] << " modelFile sampleFile " << std::endl ;
    return 1 ;
  }
  
  // Reads the sample file
  
  bioSample s(patString(argv[2]),err) ;
  if (err != NULL) {
    WARNING(err->describe()) ;
    exit(-1) ;
  }

  // Init Python
  Py_Initialize() ;

  PyRun_SimpleString("import sys") ;
  PyRun_SimpleString("sys.path.append('./module/')") ;



  // Get model name


  patString fn(argv[1]) ;
  patString pythonFileExtension(".py") ;
  int extensionPos = fn.size()-pythonFileExtension.size() ;

  // Remove .py extension if present
  if (fn.find(pythonFileExtension, extensionPos)!=string::npos) {
    fn.erase(extensionPos) ;
  }

  patFileNames::the()->setModelName(fn) ;

  GENERAL_MESSAGE("Read model specification from " << fn) ;
  parser = new bioModelParser(fn) ;
  parser->setSampleFile(patString(argv[2])) ;
  theModel = parser->readModel(err) ;
  if (err != NULL) {
    WARNING(err->describe()); 
    exit(-1) ;
  }
  GENERAL_MESSAGE("Model read") ;

  s.prepare(theModel->getExcludeExpr(),err) ;
  if (err != NULL) {
    WARNING(err->describe()); 
    exit(-1) ;
  }


  ofstream gslhfile("bioGslBiogeme.h") ;

  gslhfile << "#include \"bioGhFunction.h\"" << endl ;
  gslhfile << "class bioGslParams ;" << endl ;

  gslhfile.close() ;

  patString statCcFile = patFileNames::the()->getCcCode_s(err) ;
  if (err != NULL) {
    WARNING(err->describe()); 
    exit(-1) ;
  }
  DEBUG_MESSAGE("Generate file: " << statCcFile) ;



  ofstream mk("files.mk") ;
  mk << "GRADFILES = " ;
  mk.close() ;

  
  bioGenerateStatistics statGeneration(bioLiteralRepository::the()->getIdsOfParameters(),
				       theModel->getStatistics(),
				       statCcFile) ;
  
  statGeneration.generate(err) ;
  if (err != NULL) {
    WARNING(err->describe()); 
    exit(-1) ;
  }
  patString mainCcFile = patFileNames::the()->getCcCode(err) ;
  if (err != NULL) {
    WARNING(err->describe()); 
    exit(-1) ;
  }
  patString fctCcFile = patFileNames::the()->getCcCode_f(err) ;
  if (err != NULL) {
    WARNING(err->describe()); 
    exit(-1) ;
  }
  patString gradCcFile = patFileNames::the()->getCcCode_g(err) ;
  if (err != NULL) {
    WARNING(err->describe()); 
    exit(-1) ;
  }

  ofstream f(mainCcFile.c_str()) ;
  now.setTimeOfDay() ;
  f << "// This file has automatically been generated." << endl ;
  f << "// " << now.getTimeString(patTsfFULL) << endl ;
  f << "// " << bioVersion::the()->getCopyright() << endl ;
  f << endl ;
  f << "// " << bioVersion::the()->getVersionInfoDate() << endl ;
  f << "// " << bioVersion::the()->getVersionInfoAuthor() << endl ;
  f << endl ;
  f << "#include \"bioMain.h\"" << endl ;
  f << "#include \"bioMyObjects.h\"" << endl ;
  f.close() ;
  
  if (theModel->mustEstimate()) {
    
    DEBUG_MESSAGE("Generate files: " << mainCcFile << ", " << fctCcFile << " and " << gradCcFile) ;
    bioGenerateLikelihoodFunction codeGeneration(bioLiteralRepository::the()->getIdsOfParameters(),
						 theModel->getFormula(),
						 theModel->getConstraints(),
						 mainCcFile,fctCcFile,gradCcFile) ;
    
    
    codeGeneration.generate(err);
    if (err != NULL) {
      WARNING(err->describe()) ;
      exit(-1) ;
    }

  }
  if (theModel->mustSimulate()) {
    DEBUG_MESSAGE("Generate files: " << mainCcFile << ", " << fctCcFile << " and " << gradCcFile) ;
    bioGenerateSimulation codeGeneration(bioLiteralRepository::the()->getIdsOfParameters(),
						 theModel->getSimulation(),
						 mainCcFile,fctCcFile,gradCcFile) ;
    
    
    codeGeneration.generate(err);
    if (err != NULL) {
      WARNING(err->describe()) ;
      exit(-1) ;
    }
    DEBUG_MESSAGE("Code generated") ;
  }
  
  f.open(mainCcFile.c_str(),ios::out|ios::app) ;

  f << "" << endl ;
  f << "// Main program. " << endl ;
  f << "// It sends the pointer to the new object to the DLL and run biogeme" << endl ;
  f << "int main(int argc, char *argv[]) {" << endl ;
  
  f << "  bioMain theMain ;" << endl ;
  f << "  myBioStatistics theStats ; " << endl ;
  f << "  theMain.setStatistics(&theStats) ;" << endl ;
  f << "  myBiogeme theFunction ;" << endl ;
  f << "  theMain.setFunction(&theFunction) ;" << endl ;
  f << " myBioConstraints theContraints ; " << endl ;
  f << "  theMain.setConstraints(&theContraints) ;" << endl ;

  f << "  theMain.run(argc,argv) ;" << endl ;
  f << endl ;
  f << "  " << endl ;
  f << "" << endl ;
  f << "}" << endl ;
  f << "" << endl ;

  f.close() ;
   exit(0) ;



}





