//-*-c++-*------------------------------------------------------------
//
// File name : bioArithBinaryPlus.cc
// Author :    Michel Bierlaire and Mamy Fetiarison
// Date :      Tue Apr 28 15:25:26 2009
//
//--------------------------------------------------------------------

#include <sstream>

#include "patDisplay.h"
#include "patErrNullPointer.h"

#include "bioArithBinaryPlus.h"
#include "bioArithConstant.h"

bioArithBinaryPlus::bioArithBinaryPlus(bioExpression* par,
                                       bioExpression* left, 
                                       bioExpression* right) 
     : bioExpression(par,left,right)
{
}

bioArithBinaryPlus::~bioArithBinaryPlus() {}

bioExpression::bioOperatorType bioArithBinaryPlus::getOperatorType() const {
  return bioExpression::BINARY_OP ;
}

patString bioArithBinaryPlus::getOperatorName() const {
  return ("+") ;
}

patReal bioArithBinaryPlus::getValue(patError*& err) const  {
  if (leftChild == NULL || rightChild == NULL) {
    err = new patErrNullPointer("bioExpression") ;
    WARNING(err->describe()) ;
    return patReal();
  }
  
  patReal left = leftChild->getValue(err) ;
  if (err != NULL) {
    WARNING(err->describe()) ;
    return patReal();
  }
  patReal right = rightChild->getValue(err) ;
  if (err != NULL) {
    WARNING(err->describe()) ;
    return patReal();
  }
  patReal result = left + right ;

  return result ;
}

bioExpression* bioArithBinaryPlus::getDerivative(patULong aLiteralId,
						 patError*& err) const {
  if (leftChild == NULL || rightChild == NULL) {
    err = new patErrNullPointer("bioExpression") ;
    WARNING(err->describe()) ;
    return NULL;
  }
  
  bioExpression* leftResult = leftChild->getDerivative(aLiteralId,err) ;
  if (err != NULL) {
    WARNING(err->describe()) ;
    return NULL ;
  }
  bioExpression* rightResult = rightChild->getDerivative(aLiteralId,err) ;
  if (err != NULL) {
    WARNING(err->describe()) ;
    return NULL ;
  }
  if (leftChild->dependsOf(aLiteralId)) {
    if (rightChild->dependsOf(aLiteralId)) {
      // Left variable, right variable
      bioExpression* result = new bioArithBinaryPlus(NULL,leftResult,rightResult) ;
      return result ;
    }
    else {
      // Left variable, right constant
      return leftResult ;
    }
  }
  else {
    if (rightChild->dependsOf(aLiteralId)) {
      //Left constant, right variable
      return rightResult ;
    }
    else {
      //Left constant, right constant
      bioExpression* result = new bioArithConstant(NULL,0) ;
      return result ;
    }
  }


}

bioArithBinaryPlus* bioArithBinaryPlus::getDeepCopy(patError*& err) {
  bioExpression* leftClone(NULL) ;
  bioExpression* rightClone(NULL) ;
  if (leftChild != NULL) {
    leftClone = leftChild->getDeepCopy(err) ;
    if (err != NULL) {
      WARNING(err->describe()) ;
      return NULL ;
    }
    if (leftClone == NULL) {
      err = new patErrNullPointer("bioExpression") ;
      WARNING(err->describe()) ;
      return NULL ;
    }
  }
  if (rightChild != NULL) {
    rightClone = rightChild->getDeepCopy(err) ;
    if (err != NULL) {
      WARNING(err->describe()) ;
      return NULL ;
    }
    if (rightClone == NULL) {
      err = new patErrNullPointer("bioExpression") ;
      WARNING(err->describe()) ;
      return NULL ;
    }
  }

  bioArithBinaryPlus* theNode = 
    new bioArithBinaryPlus(NULL,leftClone,rightClone) ;
  leftClone->setParent(theNode) ;
  rightClone->setParent(theNode) ;
  return theNode ;
}

patString bioArithBinaryPlus::getExpression(patError*& err) const {
  return bioExpression::getExpression(err);
}

patString bioArithBinaryPlus::getCppCode(patString prefix, patError*& err) {
  if (codeMustBeGenerated) {

    if (leftChild == NULL || rightChild == NULL) {
      err = new patErrNullPointer("bioExpression") ;
      WARNING(err->describe()) ;
      return patString();
    }
  
    patBoolean lSimple = leftChild->isSimple() ;
    patBoolean rSimple = rightChild->isSimple() ;

    patString vLeft ;
    if (lSimple) {
      vLeft = leftChild->getSimpleCppCode(err) ;
      if (err != NULL) {
	WARNING(err->describe()) ;
	return patString();
      }
    }
    else {
	vLeft = leftChild->getCppCode(prefix, err) ;
	if (err != NULL) {
	  WARNING(err->describe()) ;
	  return patString();
	}
      }
      patString vRight ;
      if (rSimple) {
	vRight = rightChild->getSimpleCppCode(err) ;
	if (err != NULL) {
	  WARNING(err->describe()) ;
	  return patString();
	}
      }
      else {
	vRight = rightChild->getCppCode(prefix, err) ;
	if (err != NULL) {
	  WARNING(err->describe()) ;
	  return patString();
	}
      }
      stringstream str ;
      if (!lSimple) {
	str << vLeft << endl ;
      }
      if (!rSimple) {
	str << vRight << endl ;
      }
      str << getCppResultDeclaration(prefix) << "= (" ;
      if (lSimple) {
	str << vLeft ;
      }
      else {
	str << leftChild->getCppResult(prefix) ;
      }
      str << " ) + ( " ;
      if (rSimple) {
	str << vRight ;
      }
      else {
	str << rightChild->getCppResult(prefix) ;
      }
      str << ");" << endl ;
      str << "    if (!finite(" << getCppResult(prefix) << ")) {" << endl ;
      str << "      " << getCppResult(prefix) << " =  patMaxReal ;" << endl ;
      str << "    }" << endl ;
      //      str << "DEBUG_MESSAGE(\"" << getCppResult(prefix) << "= \" << " << getCppResult(prefix) << ");" << endl ;
      codeMustBeGenerated = patFALSE ;
      return patString(str.str()) ;
  }
  else {
    return patString() ;
  }
  
}
  
patString bioArithBinaryPlus::getSimpleCppCode(patError*& err) {
  if (leftChild == NULL || rightChild == NULL) {
    err = new patErrNullPointer("bioExpression") ;
    WARNING(err->describe()) ;
    return patString();
  }
  
  patString vLeft = leftChild->getSimpleCppCode(err) ;
  if (err != NULL) {
    WARNING(err->describe()) ;
    return patString();
  }
  patString vRight = rightChild->getSimpleCppCode(err) ;
  if (err != NULL) {
    WARNING(err->describe()) ;
    return patString();
  }
  stringstream str ;
  str << "(" << vLeft << ")+(" << vRight << ")" ;
  return patString(str.str()) ;
}

patString bioArithBinaryPlus::getExpressionString() const {
  stringstream str ;
  str << '+' ;
  if (leftChild != NULL) {
    str << '{' << leftChild->getExpressionString() << '}' ;
  }
  if (rightChild != NULL) {
    str << '{' << rightChild->getExpressionString() << '}' ;
  }
  return patString(str.str()) ;

}
