//-*-c++-*------------------------------------------------------------
//
// File name : bioArithIntegral.cc
// Author :    Michel Bierlaire
// Date :      Tue Mar 30 17:33:15 2010
//
//--------------------------------------------------------------------

#include <sstream>

#include "patMath.h"
#include "bioParameters.h"
#include "patDisplay.h"
#include "patErrNullPointer.h"
#include "patErrMiscError.h"
#include "bioArithIntegral.h"
#include "bioArithConstant.h"
#include "bioArithLiteral.h"
#include "bioArithMult.h"
#include "bioArithBinaryPlus.h"
#include "bioArithDivide.h"
#include "bioLiteralRepository.h"
#include "bioArithNotEqual.h"

bioArithIntegral::bioArithIntegral(bioExpression* par,
				   bioExpression* left,
				   patString aLiteralName, patError*& err)
  : bioExpression(par,left,NULL), literalName(aLiteralName)
{
  literalId = bioLiteralRepository::the()->getLiteralId(literalName,err) ;
  if (err != NULL) {
    WARNING(err->describe()) ;
  }

  // DEBUG_MESSAGE("Integrate the following expression wrt to " << literalName << " [" << literalId << "]") ;
  //  DEBUG_MESSAGE("Expression: " << *leftChild) ;
   
  
}


bioArithIntegral::~bioArithIntegral() {}

bioExpression::bioOperatorType bioArithIntegral::getOperatorType() const {
  return bioExpression::INTEGRAL_OP ;
}

patString bioArithIntegral::getOperatorName() const {
  return ("Integral") ;
}

patReal bioArithIntegral::getValue(patError*& err) const  {
  err = new patErrMiscError("Not implemented.") ;
  WARNING(err->describe()) ;
  return patReal() ;
}


bioExpression* bioArithIntegral::getDerivative(patULong aLiteralId, 
					       patError*& err) const {

  if (leftChild == NULL) {
    err = new patErrNullPointer("bioExpression") ;
    WARNING(err->describe()) ;
    WARNING(getInfo()) ;
    return NULL;
  }
  patString derivName = bioLiteralRepository::the()->getName(aLiteralId,err) ;

  if (aLiteralId == literalId) {
    stringstream str ;
    str << "Cannot compute the derivative with respect to " << derivName << " of an integral over " << literalName ;
    err = new patErrMiscError(str.str()) ;
    WARNING(err->describe()) ;
    return NULL ;
  }

  if (!dependsOf(aLiteralId) || leftChild->isStructurallyZero()) {
    bioExpression* result = new bioArithConstant(NULL,0.0) ;
    return result ;
  }
  
  bioExpression* leftDeriv = leftChild->getDerivative(aLiteralId,err) ;
  if (err != NULL) {
    WARNING(err->describe()) ;
    return NULL;
  }



  bioExpression* result = new bioArithIntegral(NULL,leftDeriv,literalName,err) ;
  if (err != NULL) {
    WARNING(err->describe()) ;
    return NULL;
  }
  return result ;
}


bioArithIntegral* bioArithIntegral::getDeepCopy(patError*& err) {
  bioExpression* leftClone(NULL) ;
  if (leftChild != NULL) {
    leftClone = leftChild->getDeepCopy(err) ;
  }
  bioArithIntegral* theNode = 
    new bioArithIntegral(NULL,leftClone,literalName,err) ;
  if (err != NULL) {
    WARNING(err->describe()) ;
    return NULL ;
  }
  leftClone->setParent(theNode) ;
  return theNode ;
}

patString bioArithIntegral::getSimpleCppCode(patError*& err) {
  err = new patErrMiscError("No simple c++ code for an integral") ;
  WARNING(err->describe());
  return patString() ;
}


patString bioArithIntegral::getCppCode(patString prefix, patError*& err) {
  if (codeMustBeGenerated) {
    if (leftChild == NULL) {
      err = new patErrNullPointer("bioExpression") ;
      WARNING(err->describe()) ;
      return patString();
    }
    stringstream str ;
    str << "    bioGslParams theParams_" << prefix << getId() << " ;" << endl ;
    str << "    theParams_" << prefix << getId()  << ".__x = __x ;" << endl ;
    str << "    theParams_" << prefix << getId()  << ".__draw = __draw ;" << endl ;
    str << "    theParams_" << prefix << getId()  << ".__beta = __beta ;" << endl ;
    str << "    theParams_" << prefix << getId()  << ".sample = sample ;" << endl ;
    str << "    theParams_" << prefix << getId()  << ".__theThreadSpan = __theThreadSpan ;" << endl ;
    str << "    bioMyGhFunction_"<<prefix << getId()<<" theFunction_"<<prefix << getId()<<"(&theParams_"<<prefix << getId()<<") ;" << endl ;

    str << "    bioGaussHermite gh_"<<prefix << getId()<<"(&theFunction_"<<prefix << getId()<<") ;" << endl ;
    
    str << getCppResultDeclaration(prefix) << " = gh_"<<prefix << getId()<<".integrate() ;" << endl ;
    //    str << "DEBUG_MESSAGE(\"" << getCppResult(prefix) << "= \" << " << getCppResult(prefix) << ");" << endl ;
    codeMustBeGenerated = patFALSE ;
    return patString(str.str()) ;
  }
  else {
    return patString() ;
  }
}

patBoolean bioArithIntegral::isStructurallyZero() const {
  return (leftChild->isStructurallyZero()) ;
}

patString bioArithIntegral::getExpressionString() const {
  stringstream str ;
  str << 'Integ' ;
  if (leftChild != NULL) {
    str << '{' << leftChild->getExpressionString() << '}' ;
  }
  return patString(str.str()) ;

}

void bioArithIntegral::generateGslFunctions(ostream& str_h,patString prefix, patError*& err) {

  // Note: at this point, nested gsl functions are not
  // allowed. Therefore, the call is not propagated to the child.

  if (leftChild == NULL) {
    err = new patErrNullPointer("bioExpression") ;
    WARNING(err->describe()) ;
    return ;
  }
  

  str_h << "    class bioMyGhFunction_"<< prefix << getId() <<": public bioGhFunction {" << endl ;
    str_h << "    public:" << endl ;
    str_h << "      bioMyGhFunction_"<< prefix << getId()<<"(bioGslParams* p): theParams(p) {} ;" << endl ;
    str_h << "      patReal getValue(patReal x) ;" << endl ;
    str_h << "    private:" << endl ;
    str_h << "      bioGslParams* theParams ;" << endl ;
    str_h << "    };" << endl ;
  ofstream mk("files.mk",ios::app) ;
  stringstream f_cc ;
  f_cc << "bioGslBiogeme" << prefix << getId() << ".cc" ;
  patString fileName(f_cc.str()) ;
  mk << fileName << " " ;
  mk.close() ;
  ofstream str_cc(fileName.c_str()) ;
  str_cc << "#include\"patDisplay.h\"" << endl ;
  str_cc << "#include\"bioGslParams.h\"" << endl ;
  str_cc << "#include\"bioGslBiogeme.h\"" << endl ;
  str_cc << "#include\"bioRowIterator.h\"" << endl ;
  str_cc << "#include\"bioSample.h\"" << endl ;
  //  str_cc << " #define DEBUG_MESSAGE(message)" << endl ;
  //str_cc << " #define DEBUG_MESSAGE(message) std::cout << message << endl ;" << endl ;

  patString fct = leftChild->getGslFunction(prefix, literalId, getId() ,err)  ;
  if (err != NULL) {
    WARNING(err->describe()) ;
    return ;
  }
  str_cc << fct << endl ;
  str_cc.close() ;
  return ;
}
