//-*-c++-*------------------------------------------------------------
//
// File name : bioArithIterator.cc
// Author :    Michel Bierlaire and Mamy Fetiarison
// Date :      Tue Jun 16 12:47:18  2009
//
//--------------------------------------------------------------------

#include <sstream>

#include "patDisplay.h"
#include "patErrNullPointer.h"
#include "patErrMiscError.h"
#include "patError.h"

#include "bioArithIterator.h"
#include "bioArithBinaryPlus.h"
#include "bioLiteralRepository.h"
#include "bioParameters.h"

#include "bioIteratorInfoRepository.h"

/*!
*/
bioArithIterator::bioArithIterator(bioExpression* par,
                         bioExpression* left,
                         bioIteratorInfo * it) 
  : bioExpression(par,left,NULL), 
    theIteratorInfo(it)  {

  
}

bioArithIterator::~bioArithIterator() {}


patReal bioArithIterator::getValue(patError*& err) const  {
  err = new patErrMiscError("Cannot compute the value of an iterator") ;
  WARNING(err->describe()) ;
  return patReal() ;
}




patString bioArithIterator::getSimpleCppCode(patError*& err) {
  err = new patErrMiscError("No simple c++ code for an iterator") ;
  WARNING(err->describe());
  return patString() ;
}

patString bioArithIterator::getCppCode(patString prefix, patError*& err) {

  if (codeMustBeGenerated) {
    if (leftChild == NULL) {
      err = new patErrNullPointer("bioExpression") ;
      WARNING(err->describe()) ;
      return patString() ;
    }
    stringstream str ;
    str << endl << "//" << endl << "// " << *this << endl << "//" << endl ;
    if (theIteratorInfo == NULL) {
      err = new patErrNullPointer("bioIteratorInfo") ;
      WARNING(err->describe()) ;
      return patString() ;
    }
    patString cppType = theIteratorInfo->getCppType(err) ;
    patString n = theIteratorInfo->iteratorName ;
    if (err != NULL) {
      WARNING(err->describe()) ;
      return patString() ;
    }

    // Make sure that the expression is regenerated at least once
    // within the loop.

    leftChild->forceCodeGeneration() ;
    bioLiteralRepository::the()->resetAllFlags() ;


    stringstream i ;
    i << getId() ;
    patString uniqueId(i.str()) ;
    theIteratorInfo->idOfCurrentNodeInExpression = uniqueId ;

    if (theIteratorInfo->isTopIterator()) {
    
      str << "bioIteratorSpan " << prefix << "aSpan" << uniqueId << "(patString(\"" << n << "\"),0) ;" << endl ;
    }
    if (theIteratorInfo->isMetaIterator()) {
      str << " bioMetaIterator* " << prefix << "theIter_" << n << "_" << uniqueId << "  = sample->createMetaIterator(" << prefix << "aSpan" ;
      if (theIteratorInfo->isTopIterator()) {
	str << uniqueId; 
      }
      else {
	str << theIteratorInfo->theParent->idOfCurrentNodeInExpression ;
      }
      str <<",__theThreadSpan,err) ;" << endl ;
      str << "      if (err != NULL) {" << endl ;
      str << "	WARNING(err->describe()) ;" << endl ;
      str << " exit(-1) ;" << endl ;
      str << "      }" << endl ;
    }
    else if (theIteratorInfo->isRowIterator()) {
      //      DEBUG_MESSAGE("Create code for row iterator: " << *theIteratorInfo) ;
      bioIteratorInfo* theChild = bioIteratorInfoRepository::the()->getIteratorInfo(theIteratorInfo->childIteratorName) ;
      str << "// Created at row 107 of bioArithIterator.cc" << endl ;
      str << " bioRowIterator* " << prefix << "theIter_" << n <<  "_" << uniqueId<< "  = sample->createRowIterator(" << prefix << "aSpan" ;
      if (theIteratorInfo->isTopIterator()) {
 	str << uniqueId; 
      }
      else {
 	str << theIteratorInfo->theParent->idOfCurrentNodeInExpression ;
      }
      str <<",__theThreadSpan,err) ;" << endl ;
       
      str << "      if (err != NULL) {" << endl ;
      str << "	WARNING(err->describe()) ;" << endl ;
      str << "	exit(-1) ;" << endl ;
      str << "      }" << endl ;
    }
    else if (theIteratorInfo->isDrawIterator()) {
      str << " bioDrawIterator* " << prefix << "theIter_" << n <<  "_" << uniqueId << "  = bioRandomDraws::the()->createIterator() ;" << endl ;

    }
    else {
      stringstream str ;
      str << "Unknown iterator: " << *theIteratorInfo ;
      err = new patErrMiscError(str.str()) ;
      WARNING(err->describe()) ;
      return patString();
    }
 
    str << getCppResultDeclaration(prefix);
    if (isSum()) {
      str << " = 0.0 ;" ;
    }
    else if (isProd()) {
      str << " = 1.0 ;" ;
    }
    else {
      err = new patErrMiscError("Unknown iterator") ;
      WARNING(err->describe()) ;
      return patString() ;
    }
    str <<  endl ;


    str << "for(" << prefix << "theIter_" << n <<  "_" << uniqueId << "->first() ; " << endl ;
    str << "    !" << prefix << "theIter_" << n <<  "_" << uniqueId << "->isDone() ;" << endl ;
    str << "    " << prefix << "theIter_" << n <<  "_" << uniqueId << "->next()) {" << endl ;
    if (theIteratorInfo->isRowIterator()) {
      str << "  patVariables* __x = " << prefix << "theIter_" << n <<  "_" << uniqueId << "->currentItem() ;" 
	  << endl ;
    }
    else if (theIteratorInfo->isMetaIterator()) {
      //      DEBUG_MESSAGE("  " << prefix << "aSpan"<<uniqueId<<" = " << prefix << "theIter_" << n <<  "_" << uniqueId << "->currentItem() ;") ;

      str <<"bioIteratorSpan  " << prefix << "aSpan"<<uniqueId<<" = " << prefix << "theIter_" << n <<  "_" << uniqueId << "->currentItem() ;" << endl ;  
    }
    else if (theIteratorInfo->isDrawIterator()) {
      str << "  __draw = " << prefix << "theIter_" << n <<  "_" << uniqueId << "->currentItem() ;" 
	  << endl ;
    }

    if (!leftChild->isSimple()) {
      str << leftChild->getCppCode(prefix, err) ;
      if (err != NULL) {
	WARNING(err->describe()) ;
	return patString() ;
      }
    }
    if (isSum()) {
      str << getCppResult(prefix) << " += " ;
    }
    else if (isProd()) {
      str << getCppResult(prefix) << " *= " ;
    }
    else {
      err = new patErrMiscError("Unknown iterator") ;
      WARNING(err->describe()) ;
      return patString() ;
    }
      
    if (leftChild->isSimple()) {
      str << leftChild->getSimpleCppCode(err) ;
      if (err != NULL) {
	WARNING(err->describe()) ;
	return patString() ;
      }
    }
    else {
      str << leftChild->getCppResult(prefix) ;

    }
    str << " ; " << endl ;
    str << "if (finite(" << getCppResult(prefix) << ") == 0) {" << endl ; 
    str << "  if (" << getCppResult(prefix) << "< 0) {" << endl ;
    str << "	" << getCppResult(prefix) << " = " << -patMaxReal/2.0 << " ;" << endl ;
    str << "   }" << endl ;
    str << "  else {" << endl ;
    str << "	" << getCppResult(prefix) << " = " << patMaxReal/2.0 << " ;" << endl ;
    str << "   }" << endl ;
    str << "   }" << endl ;

  str << "}" << endl ;
  if (isSum()) {
    if (theIteratorInfo->isDrawIterator()) {
      str << getCppResult(prefix) << " /= patReal(" << bioParameters::the()->getValueInt("NbrOfDraws",err) << ") ; " << endl  ;
      
    }
  }
    str << "DELETE_PTR(" << prefix << "theIter_" << n  << "_" << uniqueId << ") ;" << endl ;
    codeMustBeGenerated = patFALSE ;
    //    str << "DEBUG_MESSAGE(\"" << getCppResult(prefix) << "= \" << " << getCppResult(prefix) << ");" << endl  ;
    codeMustBeGenerated = patFALSE ;
    return patString(str.str()) ;
  }
  else {
    return patString() ;
  }
  
}

patString bioArithIterator::getExpression(patError*& err) const {
  if (theIteratorInfo == NULL) {
    err = new patErrNullPointer("bioIteratorInfo") ;
    WARNING(err->describe()) ;
    return patString() ;
  }
  stringstream ss ;
  ss << theIteratorInfo->getInfo() << bioExpression::getExpression(err) ;
  if (err != NULL) {
    WARNING(err->describe()) ;
    return patString() ;
  }
  return patString(ss.str()) ;
}


patBoolean bioArithIterator::isSimple() const {
  return patFALSE ;
}

bioIteratorInfo* bioArithIterator::theIterator() {
  return theIteratorInfo ;
}

patULong bioArithIterator::getNumberOfOperations() const {
  if (!isSum() && !isProd()) {
    return 0 ;
  }
  if (leftChild == NULL) {
    return 0 ;
  }
  patULong result = leftChild->getNumberOfOperations() ;
  if (theIteratorInfo->isDrawIterator()) {
    patError* err(NULL) ;
    patULong nDraws = bioParameters::the()->getValueInt("NbrOfDraws",err) ;
    result *= nDraws ;
  }
  else {
    result *= theIteratorInfo->mapIds.size() ;
  }
  return result ;
}


void bioArithIterator::simplifyLowerIteratorOrSequence(patError*& err) {
  if (leftChild == NULL) {
    err = new patErrNullPointer("bioExpression") ;
    WARNING(err->describe()) ;
    return ;
  }
  bioExpression* newExpression = simplify(leftChild,err) ;
  if (err != NULL) {
    WARNING(err->describe()) ;
    return ;
  }
  leftChild = newExpression ;
  leftChild->setParent(this) ;
}


