//-*-c++-*------------------------------------------------------------
//
// File name : bioArithMergedSum.cc
// Author :    Michel Bierlaire and Mamy Fetiarison
// Date :      Thu Aug  6 14:22:23 2009
//
//--------------------------------------------------------------------

#include "patDisplay.h"
#include "bioArithLiteral.h"
#include "bioArithMergedSum.h"
#include "patErrMiscError.h"
#include "patErrNullPointer.h"
#include "bioIteratorInfo.h"
#include "bioLiteralRepository.h"
#include "bioParameters.h"

// results contaibns the name of the variables where the results must
// be stored in the generated code
bioArithMergedSum::bioArithMergedSum(bioExpression* par,
				     bioIteratorInfo* aIterator,
				     vector<patString> results,
				     vector<bioExpression*> listOfExpressions) :
  bioExpression(par,NULL,NULL),
  theIteratorInfo(aIterator),
  variablesNames(results),
  allExpressions(listOfExpressions) {
  
}

bioArithMergedSum::~bioArithMergedSum() {
  
}

bioExpression::bioOperatorType bioArithMergedSum::getOperatorType() const {
  return bioExpression::MERGED_OP ;
}

patString bioArithMergedSum::getOperatorName() const {
  return patString("MergedSum") ;
}

patReal bioArithMergedSum::getValue(patError*& err) const {
  err = new patErrMiscError("Cannot compute the value of an iterator") ;
  WARNING(err->describe()) ;
  return patReal() ;
}

bioExpression* bioArithMergedSum::getDerivative(patULong aLiteralId, 
						patError*& err) const {
  vector<bioExpression*> theDerivatives ;
  for (vector<bioExpression*>::const_iterator i = allExpressions.begin() ;
       i != allExpressions.end() ;
       ++i) {
    bioExpression* aDeriv = (*i)->getDerivative(aLiteralId,err) ;
    if (err != NULL) {
      WARNING(err->describe()) ;
      return NULL ;
    }
    theDerivatives.push_back(aDeriv) ;
  }
  vector<patString> theDerivResult ;
  patString literal = bioLiteralRepository::the()->getName(aLiteralId,err) ;
  if (err != NULL) {
    WARNING(err->describe()) ;
    return NULL ;
  }
  for (vector<patString>::const_iterator i = variablesNames.begin() ;
       i != variablesNames.end() ;
       ++i) {
    stringstream str ;
    str << *i << "_" << literal ;
    theDerivResult.push_back(patString(str.str())) ;
  }
  bioExpression* result = new bioArithMergedSum(NULL,
						theIteratorInfo,
						theDerivResult,
						theDerivatives) ;
  return result ;

}

bioArithMergedSum* bioArithMergedSum::getDeepCopy(patError*& err) {
  vector<bioExpression*> theCopies ;
  for (vector<bioExpression*>::iterator i = allExpressions.begin() ;
       i != allExpressions.end() ;
       ++i) {
    bioExpression* aCopy = (*i)->getDeepCopy(err) ;
    if (err != NULL) {
      WARNING(err->describe()) ;
      return NULL ;
    }
    theCopies.push_back(aCopy) ;
  }
  bioArithMergedSum* result = new bioArithMergedSum(NULL,
						theIteratorInfo,
						variablesNames,
						theCopies) ;
  return result ;
  
}

patString bioArithMergedSum::getCppCode(patString prefix, patError*& err) {

  

  if (codeMustBeGenerated) {
    tryToSimplify(err) ;
    stringstream str ;

    str << "    bioIterates theData ;" << endl ;
    str << "    theData.__beta = __beta ;" << endl ;
    str << "    theData.__theThreadSpan = __theThreadSpan ;" << endl ;
    str << "    theData.sample = sample ;" << endl ;
    if (theIteratorInfo == NULL) {
      err = new patErrNullPointer("bioIteratorInfo") ;
      WARNING(err->describe()) ;
      return patString() ;
    }
    patString cppType = theIteratorInfo->getCppType(err) ;
    patString n = theIteratorInfo->iteratorName ;
    if (err != NULL) {
      WARNING(err->describe()) ;
      return patString() ;
    }

    // Make sure that the expression is regenerated at least once
    // within the loop.
    for (vector<bioExpression*>::const_iterator i = allExpressions.begin() ;
	 i != allExpressions.end() ;
	 ++i) {
      (*i)->forceCodeGeneration();
    }
    bioLiteralRepository::the()->resetAllFlags() ;

    stringstream i ;
    i << getId() ;
    patString uniqueId(i.str()) ;
    theIteratorInfo->idOfCurrentNodeInExpression = uniqueId ;

    if (theIteratorInfo->isTopIterator()) {
    
      str << "bioIteratorSpan " << prefix << "aSpan" << uniqueId << "(patString(\"" << n << "\"),0) ;" << endl ;
    }
    if (theIteratorInfo->isMetaIterator()) {
      str << " bioMetaIterator* " << prefix << "theIter_" << n << "_" << uniqueId << "  = sample->createMetaIterator(" << prefix << "aSpan" ;
      if (theIteratorInfo->isTopIterator()) {
	str << uniqueId; 
      }
      else {
	str << theIteratorInfo->theParent->idOfCurrentNodeInExpression ;
      }
    str <<",__theThreadSpan,err) ;" << endl ;
      str << "      if (err != NULL) {" << endl ;
      str << "	WARNING(err->describe()) ;" << endl ;
      str << " exit(-1) ;" << endl ;
      str << "      }" << endl ;
    }
    else if (theIteratorInfo->isRowIterator()) {
      str << " bioRowIterator* " << prefix << "theIter_" << n <<  "_" << uniqueId<< "  = sample->createRowIterator(" << prefix << "aSpan" ;
      if (theIteratorInfo->isTopIterator()) {
	str << uniqueId; 
      }
      else {
	str << theIteratorInfo->theParent->idOfCurrentNodeInExpression ;
      }
      str <<",__theThreadSpan,err) ;" << endl ;
      str << "      if (err != NULL) {" << endl ;
      str << "	WARNING(err->describe()) ;" << endl ;
      str << "	exit(-1) ;" << endl ;
      str << "      }" << endl ;
      
    }
    else if (theIteratorInfo->isDrawIterator()) {
      str << " bioDrawIterator* " << prefix << "theIter_" << n <<  "_" << uniqueId << "  = bioRandomDraws::the()->createIterator() ;" << endl ;

    }
    else {
      stringstream str ;
      str << "Unknown iterator: " << *theIteratorInfo ;
      err = new patErrMiscError(str.str()) ;
      WARNING(err->describe()) ;
      return patString();
    }

    for (patULong i=0 ; i < allExpressions.size() ; ++i) {
      str << variablesNames[i] << " = 0.0 ;" <<  endl ;
    } 

    str << "vector<patReal> oneIterValues(" << allExpressions.size() << ");" << endl ;
    str << "for(" << prefix << "theIter_" << n <<  "_" << uniqueId << "->first() ; " << endl ;
    str << "    !" << prefix << "theIter_" << n <<  "_" << uniqueId << "->isDone() ;" << endl ;
    str << "    " << prefix << "theIter_" << n <<  "_" << uniqueId << "->next()) {" << endl ;
    if (theIteratorInfo->isRowIterator()) {
      str << "  patVariables* __x = " << prefix << "theIter_" << n <<  "_" << uniqueId << "->currentItem() ;" 
	  << endl ;
      str << "      theData.__x = __x ;" << endl ;
    }
    else if (theIteratorInfo->isMetaIterator()) {
      str <<"bioIteratorSpan  " << prefix << "aSpan"<<uniqueId<<" = " << prefix << "theIter_" << n <<  "_" << uniqueId << "->currentItem() ;" << endl ;  ;
      str << "      theData.aSpan = " << prefix << "aSpan" << uniqueId<<" ;" << endl ;
    }
    else if (theIteratorInfo->isDrawIterator()) {
      str << "  __draw = " << prefix << "theIter_" << n <<  "_" << uniqueId << "->currentItem() ;" 
	  << endl ;
      str << "      theData.__draw = __draw ;" << endl ;
    }

    //    DEBUG_MESSAGE("Write files.mk") ;
    ofstream mk("files.mk",ios::app) ;
 
    // Repeated expressions 
    str << "vector<patReal> commonExpressions(" << repeatedExpressions.size() << "); " << endl ;
    if (!repeatedExpressions.empty()) {
      str << "//" << endl ;
      str << "// Expressions common to several entries" << endl ;
      str << "//" << endl ;
      for (patULong j = 0 ; j < repeatedExpressions.size() ; ++j) {
	repeatedExpressions[j]->forceCodeGeneration() ;

	str << "// " << *repeatedExpressions[j] << endl ;
	str << "// SHORT: " << *shortExpressions[j] << endl ;
	if (repeatedExpressions[j]->isSimple()) {
	  str <<  "patReal " << shortExpressions[j]->getSimpleCppCode(err) << "=" << "commonExpressions[" << j << "]="<< repeatedExpressions[j]->getSimpleCppCode(err) << " ; " << endl ;
	  if (err != NULL) {
	    WARNING(err->describe()) ;
	    return patString() ;
	  }
	}
	else {
	  str << repeatedExpressions[j]->getCppCode(prefix,err) ;
	  if (err != NULL) {
	    WARNING(err->describe()) ;
	    return patString() ;
	  }
	  str << endl ;
	  str << "commonExpressions[" << j << "]="<< repeatedExpressions[j]->getCppResult(prefix) << " ; " << endl ;
	str << "patReal " << shortExpressions[j]->getSimpleCppCode(err) << "="<< repeatedExpressions[j]->getCppResult(prefix) << " ; " << endl ;
	}

      }
    }
    for (patULong i=0 ; i < allExpressions.size() ; ++i) {
      allExpressions[i]->forceCodeGeneration() ;
      bioLiteralRepository::the()->resetAllFlags() ;
     stringstream fname ;
      fname << prefix << "__file" << i << ".cc" ;
      mk << fname.str() << " " ;
      ofstream f(fname.str().c_str()) ; 

      f << "#include \"patType.h\"" << endl ;
      f << "#include \"trVector.h\"" << endl ;
      f << "#include \"patVariables.h\"" << endl ;
      f << "#include \"patMath.h\"" << endl ;
      f << "#include \"patConst.h\"" << endl ;
      f << "#include \"bioIterates.h\"" << endl ;
      f << "#include \"bioDrawIterator.h\"" << endl ;
      f << "#include \"bioGslBiogeme.h\"" << endl ;
      f << "#include \"bioGslParams.h\"" << endl ;
      f << "#include \"bioGaussHermite.h\"" << endl ;
      f << "#include \"patDisplay.h\"" << endl ;
      f << "#include \"patErrMiscError.h\"" << endl ;
      f << "#include \"bioRowIterator.h\"" << endl ;
      f << "#include \"bioMetaIterator.h\"" << endl ;
      f << "#include \"bioRandomDraws.h\"" << endl ;
      f << "" << endl ;
      //      f << " #define DEBUG_MESSAGE(message)" << endl ;
      //f << " #define DEBUG_MESSAGE(message) std::cout << message << endl ;" << endl ;

      f << "      patReal " << prefix << "__file" << i << "(bioIterates* theData,vector<patReal>* commonExpressions) {" << endl ;
      
      f << "      patError* err(NULL) ;" << endl ;
      f << "      patVariables* __x = theData->__x ;" << endl ;
      f << "      bioSample* sample = theData->sample ;" << endl ;
      f << "      bioIteratorSpan __theThreadSpan = theData->__theThreadSpan ;" << endl ;
      f << "      bioIteratorSpan " << prefix << "aSpan" << uniqueId << " = theData->aSpan ;" << endl ;
      f << "      vector< vector<patReal> >* __draw = theData->__draw ;" << endl ;
      f << "      trVector* __beta = theData->__beta ;" << endl ;

      for (patULong j = 0 ; j < repeatedExpressions.size() ; ++j) {
	f << "patReal " << shortExpressions[j]->getSimpleCppCode(err) << " = (*commonExpressions)[" << j << "] ; " << endl ;
	if (err != NULL) {
	  WARNING(err->describe()) ;
	  return patString() ;
	}

      }      

      bioLiteralRepository::the()->resetAllFlags() ;
      if (!allExpressions[i]->isSimple()) {
	//	DEBUG_MESSAGE("Expression " << i << " is not simple: " << *(allExpressions[i])) ;
	//	DEBUG_MESSAGE("COde: " << allExpressions[i]->getCppCode(prefix, err)) ; 
	f << allExpressions[i]->getCppCode(prefix, err) ;
	if (err != NULL) {
	  WARNING(err->describe()) ;
	  return patString() ;
	}
      }
      str << "oneIterValues[" << i << "]= " ;
      str << prefix << "__file" << i << "(&theData,&commonExpressions) ;" ;

      

      if (allExpressions[i]->isSimple()) {
	DEBUG_MESSAGE("Expression " << i << " is simple: " << *(allExpressions[i])) ;
	f << "return(" << allExpressions[i]->getSimpleCppCode(err) << ");" ;
	if (err != NULL) {
	  WARNING(err->describe()) ;
	  return patString() ;
	}
      }

      else {
	f << "return(" << allExpressions[i]->getCppResult(prefix) << ");";
      }
      f << "}" << endl ;
      f.close() ;
      str << " ; " << endl ;
      str << variablesNames[i] << " += oneIterValues[" << i << "]" ;
      str << " ; " << endl ;
    }
    mk.close() ;
    str << additionalCode << endl ;
    str << "}" << endl ;
    if (theIteratorInfo->isDrawIterator()) {
      str << getCppResult(prefix) << " /= patReal(" << bioParameters::the()->getValueInt("NbrOfDraws",err) << ") ; " << endl  ;
    
    }
    str << "DELETE_PTR(" << prefix << "theIter_" << n  << "_" << uniqueId << ") ;" << endl ;
    codeMustBeGenerated = patFALSE ;
    return patString(str.str()) ;
  }
  else {
    return patString() ;
  }
  
}

patString bioArithMergedSum::getSimpleCppCode(patError*& err) {
  err = new patErrMiscError("No simple c++ code for an iterator") ;
  WARNING(err->describe());
  return patString() ;

}
patString bioArithMergedSum::getExpression(patError*& err) const {
  if (variablesNames.size() != allExpressions.size()) {
    stringstream str ;
    str << "Inconsistent sizes: " << variablesNames.size() << "!=" << allExpressions.size() ;
    err = new patErrMiscError(str.str()) ;
    WARNING(err->describe()) ;
    return patString() ;
  }
  stringstream str ;
  str << theIteratorInfo->getInfo() << "[" ;
  for (patULong i = 0 ; i < variablesNames.size() ; ++i) {
    if (i != 0) {
      str << ", " ;
    }
    str << variablesNames[i] << "=" << allExpressions[i]->getExpression(err) ;
    if (err != NULL) {
      WARNING(err->describe()) ;
      return patString() ;
    }
  }
  str << "]" ;
  return patString(str.str()) ;
}

patBoolean bioArithMergedSum::isSimple() const {
  return patFALSE ;
}

patBoolean bioArithMergedSum::dependsOf(patULong aLiteralId) const {
  for (vector<bioExpression*>::const_iterator i = allExpressions.begin() ;
       i != allExpressions.end() ;
       ++i) {
    if ((*i)->dependsOf(aLiteralId)) {
      return patTRUE ;
    }
  }
  return patFALSE ;
}
void bioArithMergedSum::forceCodeGeneration() {
  codeMustBeGenerated = patTRUE ;
  bioLiteralRepository::the()->resetAllFlags() ;
  for (vector<bioExpression*>::const_iterator i = allExpressions.begin() ;
       i != allExpressions.end() ;
       ++i) {
    (*i)->forceCodeGeneration();
  }
  
}

void bioArithMergedSum::simplifyZeros(patError*& err) {
  for (vector<bioExpression*>::const_iterator i = allExpressions.begin() ;
       i != allExpressions.end() ;
       ++i) {
    (*i)->simplifyZeros(err);
    if (err != NULL) {
      WARNING(err->describe()) ;
      return ;
    }
  }
}


bioIteratorInfo* bioArithMergedSum::theIterator() {
  return theIteratorInfo ;
}

patULong bioArithMergedSum::getNumberOfOperations() const {
  patULong result(0) ;
  for (vector<bioExpression*>::const_iterator i = repeatedExpressions.begin() ;
       i != repeatedExpressions.end() ;
       ++i) {
    result += (*i)->getNumberOfOperations() ;
  }

  for (vector<bioExpression*>::const_iterator i = allExpressions.begin() ;
       i != allExpressions.end() ;
       ++i) {
    result += (*i)->getNumberOfOperations() ;
  }
  //  result *= theIteratorInfo->mapIds.size() ;
  return result ;
}

patString bioArithMergedSum::getIncludeStatements(patString prefix) const {
  stringstream str ;
  for (vector<bioExpression*>::const_iterator i = allExpressions.begin() ;
       i != allExpressions.end() ;
       ++i) {
    str <<  (*i)->getIncludeStatements(prefix) << endl ;
  }
  for (patULong i=0 ; i < allExpressions.size() ; ++i) {
    str << "patReal "<< prefix <<"__file" << i << "(bioIterates* theData,vector<patReal>* commonExpressions) ;" << endl ;
  }
  return patString(str.str()) ;
}


patBoolean bioArithMergedSum::containsAnIterator() const {
  return patTRUE ;
}

patBoolean bioArithMergedSum::containsASequence() const {
  for (vector<bioExpression*>::const_iterator i = allExpressions.begin() ;
       i != allExpressions.end() ;
       ++i) {
    if  ((*i)->containsASequence()) {
      return patTRUE ;
    }
  }
  return patFALSE ;

}

patBoolean bioArithMergedSum::containsAnIntegral() const {
  for (vector<bioExpression*>::const_iterator i = allExpressions.begin() ;
       i != allExpressions.end() ;
       ++i) {
    if  ((*i)->containsAnIntegral()) {
      return patTRUE ;
    }
  }
  return patFALSE ;

}

bioExpression* bioArithMergedSum::getExpressionFromId(patULong theId) {
  DEBUG_MESSAGE("Look for expression " << theId) ;
  if (theId == getId()) {
    return this ;
  }
  bioExpression* result(NULL) ;
  for (vector<bioExpression*>::iterator i = allExpressions.begin() ;
       i != allExpressions.end() ;
       ++i) {
    result = (*i)->getExpressionFromId(theId) ;
    if (result != NULL) {
      return result ;
    }
  }
  return NULL ;
  
}

patULong bioArithMergedSum::countExpressions(map<patString,patULong>* expr, 
					     map<patString,list<patULong> >* theIds,
					     map<patString,list<bioExpression*> >* thePointers,
					     patError*& err)   {
    stringstream str ;
    str << "Expression to simplify cannot contain an iterator: " << *this;
    err = new patErrMiscError(str.str()) ;
    WARNING(err->describe()) ;
    return 0;
}


void bioArithMergedSum::replaceExpression(bioExpression* theExpression, 
					  bioExpression* anotherExpression, 
					  patError*& err) {
  
  if (*this == *theExpression) {
    stringstream str ;
    str << "Cannot replace the root of an expression" ;
    err = new patErrMiscError(str.str()) ;
    WARNING(err->describe());
    return ;
  }

  for (vector<bioExpression*>::size_type i = 0 ; 
       i < allExpressions.size() ;
       ++i) {
    //    DEBUG_MESSAGE("BEFORE: Expr["<<i<<"]="<<*(allExpressions[i])) ;
    if (*(allExpressions[i]) == *theExpression) {
      //      DELETE_PTR(*i) ;
      allExpressions[i] = anotherExpression->getDeepCopy(err) ;
      if (err != NULL) {
	WARNING(err->describe()) ;
	return ;
      }
      allExpressions[i]->setParent(this) ;
      return ;
    }
    else {
      allExpressions[i]->replaceExpression(theExpression,
					   anotherExpression,
					   err) ;
      if (err != NULL) {
	WARNING(err->describe()) ;
	return ;
      }
    }
    //    DEBUG_MESSAGE("AFTER: Expr["<<i<<"]="<<*(allExpressions[i])) ;
  }
}

patString bioArithMergedSum::getExpressionString() const {
  stringstream str ;
  str << '$Z' ;
  str << theIteratorInfo->iteratorName ;
  str << '[' ;
  for (vector<bioExpression*>::const_iterator i = allExpressions.begin() ;
       i != allExpressions.end() ;
       ++i) {
    str << '{' << (*i)->getExpressionString() << '}' ;
  }
  str << ']' ;
  return patString(str.str()) ;
}

void bioArithMergedSum::getExpressionsToSimplify(vector<bioExpression*>* listOfExpr, 
						 patError*& err) {

  // The merged iterator will be added to the list if at least one of
  // its subpexression does not contain an iterator. At the time of
  // simplification, only expression without iterators will be
  // simplified

  patBoolean lowest(patFALSE) ;
  for (vector<bioExpression*>::const_iterator i = allExpressions.begin() ;
       i != allExpressions.end() ;
       ++i) {
    if ((*i)->containsAnIterator()) {
      (*i)->getExpressionsToSimplify(listOfExpr,err) ;
      if (err != NULL) {
	WARNING(err->describe()) ;
	return ;
      }
    }
    else {
      if (!lowest) {
	lowest = patTRUE ;
	listOfExpr->push_back(this) ;
      }
    }
  }
  
}

void bioArithMergedSum::tryToSimplify(patError*& err) {

  short simplifyExpressions = bioParameters::the()->getValueInt("simplifyExpressions",err) ;
  if (err != NULL) {
    WARNING(err->describe()) ;
    return ;
  }
  if (simplifyExpressions == 0) {
    return  ;
  }


  patULong originalNbrOper = getNumberOfOperations() ;
    patULong counter = 0 ;
  patBoolean finished = patFALSE ;
  while (!finished && counter <= 100) {
    ++counter ;

    map<patString,patULong> theSubExpressions ;
    map<patString,list<patULong> > theSubIds ;
    map<patString,list<bioExpression*> > thePointers ;
    for (vector<bioExpression*>::iterator expr = allExpressions.begin() ;
	 expr != allExpressions.end() ;
	 ++expr) {
      bioExpression* theExpression = *expr ;
      if (!theExpression->containsAnIterator()) {
	//	DEBUG_MESSAGE("No iterator. OK to simplify globally.") ;
	theExpression->countExpressions(&theSubExpressions,&theSubIds,&thePointers,err) ;
	if (err != NULL) {
	  WARNING(err->describe()) ;
	  return ;
	}
      }
      else {
	//	DEBUG_MESSAGE("Iterator. Simplify locally.") ;
	if (!theExpression->canBeSimplified()) {
	  vector<bioExpression*> expressionsToSimplify ;
	  theExpression->getExpressionsToSimplify(&expressionsToSimplify,err) ;
	  if (err != NULL) {
	    WARNING(err->describe()) ;
	    return ;
	  }
	  //	  DEBUG_MESSAGE("Number of expressions found: " << expressionsToSimplify.size()) ;
	  for (vector<bioExpression*>::iterator i = expressionsToSimplify.begin() ;
	       i != expressionsToSimplify.end() ;
	       ++i) {
	    (*i)->simplifyLowerIteratorOrSequence(err) ;
	    if (err != NULL) {
	      WARNING(err->describe()) ;
	      return ;
	    }
	  }
	}
	else {
	  // Simplify expression
	  bioExpression* newExpression = simplify(theExpression,err) ;
	  if (err != NULL) {
	    WARNING(err->describe()) ;
	    return ;
	  }
	  *expr = newExpression ;
	}
	
	if (err != NULL) {
	  WARNING(err->describe()) ;
	  return ;
	}
      }
    }
//   DEBUG_MESSAGE("SUBIDS") ;
//   for (map<patString,list<patULong> >::iterator k = theSubIds.begin() ;
//        k != theSubIds.end() ;
//        ++k) {
//     DEBUG_MESSAGE(k->first << ": ") ;
//     for (list<patULong>::iterator kk = k->second.begin() ;
// 	 kk != k->second.end() ;
// 	 ++kk) {
//       cout << *kk << " " ;
//     }
//     cout << endl ;
//   }
    
    patULong maxGain = 0 ;
    bioExpression* aRepeatedExpression(NULL) ;
    
    for (map<patString,patULong>::iterator i = theSubExpressions.begin() ;
	 i != theSubExpressions.end() ;
	 ++i) {
      
      if (i->second > 1) {
	//	DEBUG_MESSAGE(i->first << ": " << i->second) ;
	bioExpression* theExpr = *(thePointers[i->first].begin()) ;
	if (theExpr == NULL) {
	  err = new patErrNullPointer("bioExpression") ;
	  WARNING(err->describe()) ;
	  return ;
	}
	//	DEBUG_MESSAGE("Expression: " << *theExpr) ;
	patULong gain = theExpr->getNumberOfOperations() * (i->second-1) ;
	if (gain > maxGain) {
	  maxGain = gain ;
	  aRepeatedExpression = theExpr ;
	}
      }
    }
    if (maxGain > 0) {
      //      DEBUG_MESSAGE("Maxgain: " << maxGain << " Simplify: " << *aRepeatedExpression) ;
      repeatedExpressions.push_back(aRepeatedExpression) ;
      patULong theLiteralId = bioLiteralRepository::the()->addIntermediateValue(err) ;
      if (err != NULL) {
	WARNING(err->describe()) ;
	return ;
      }
      bioArithLiteral* theSimplifiedExpression =  new bioArithLiteral(NULL,theLiteralId) ;
      shortExpressions.push_back(theSimplifiedExpression) ;
      
      //      DEBUG_MESSAGE("SIMPLIFIED EXPRESSION:" << *theSimplifiedExpression) ;
      replaceExpression(aRepeatedExpression,theSimplifiedExpression,err) ;  
      if (err != NULL) {
	WARNING(err->describe()) ;
	return ;
      }
    }
    else {
      finished = patTRUE ;
    }
    patULong currentNbrOper = getNumberOfOperations() ;

  }
  
}

void bioArithMergedSum::generateGslFunctions(ostream& str_h,patString prefix, patError*& err) {
  for (vector<bioExpression*>::const_iterator i = allExpressions.begin() ;
       i != allExpressions.end() ;
       ++i) {
    (*i)->generateGslFunctions(str_h,prefix,err) ;
    if (err != NULL) {
      WARNING(err->describe()) ;
      return ;
    }
  }
}
