//-*-c++-*------------------------------------------------------------
//
// File name : bioExpression.cc
// Author :    Michel Bierlaire and Mamy Fetiarison
// Date :      Wed Apr 29 14:31:38 2009
//
//--------------------------------------------------------------------

#include <sstream>
#include "patMath.h"
#include "bioExpression.h"
#include "patDisplay.h"

#include "bioParameters.h"
#include "patErrNullPointer.h"
#include "patErrMiscError.h"
#include "bioArithConstant.h"
#include "bioArithSequence.h"

#include "bioExpressionRepository.h"
#include "bioLiteralRepository.h"

bioExpression::bioExpression(bioExpression* par,
			     bioExpression* left, 
                             bioExpression* right) : 
  parent(par),leftChild(left),rightChild(right),codeMustBeGenerated(patTRUE) {
  

  theId = bioExpressionRepository::the()->addExpression(this) ;

  if (left != NULL) {
    left->setParent(this) ;
  }
  if (right != NULL) {
    right->setParent(this) ;
  }
  
}

ostream& operator<<(ostream& str, const bioExpression& x) {
  patError* err = NULL ;
  str << x.getExpression(err) ;
  if (err != NULL) {
    str << err->describe() << '\t' ;
  }
  return str ;
}

// Memory management for the expression nodes is not well designed. So
// memory leak will appear but it is expected that it will not be
// major.
bioExpression::~bioExpression() {
  // Free memory !!!
}

patBoolean bioExpression::isTop() const {
  return parent == NULL;
}

bioExpression* bioExpression::getLeftChild() const {
  return leftChild;
}

bioExpression* bioExpression::getRightChild() const {
  return rightChild;
}

void bioExpression::setParent(bioExpression* par) {
  parent=par;
}
  
bioExpression* bioExpression::getRoot() const {
  if (this->isTop())
    return (bioExpression*)this;
  else
    return parent->getRoot();
}

bioExpression* bioExpression::getParent() const  {
  return parent;
}

patString bioExpression::getExpression(patError*& err) const {

  patString result ;
  patString leftResult ;
  patString rightResult ;
  switch (getOperatorType()) {
  case BINARY_OP :
  {
    if (leftChild == NULL || rightChild == NULL) {
      err = new patErrNullPointer("bioExpression") ;
      WARNING(err->describe()) ;
      return patString();
    }

    leftResult = leftChild->getExpression(err) ;
    if (err != NULL) {
      WARNING(err->describe()) ;
      return patString();
    }
    rightResult = rightChild->getExpression(err) ;
    if (err != NULL) {
      WARNING(err->describe()) ;
      return patString();
    }
    
      result = "" ;
      if (leftChild->getOperatorType() == BINARY_OP) {
	result += "(" ;
      }
      result += " " ;
      result += leftResult ;
      result += " " ;
      if (leftChild->getOperatorType() == BINARY_OP) {
	result += ")" ;
      }
      result += " " ;
      result += getOperatorName() ;
      result += " " ;
      if (rightChild->getOperatorType() == BINARY_OP) {
	result += "(" ;
      }
      result += " " ;
      result += rightResult ;
      result += " " ;
      if (rightChild->getOperatorType() == BINARY_OP) {
	result += ")" ;
      }
      return result ;
    }    
  case UNARY_OP :
    {
      if (leftChild == NULL) {
	err = new patErrNullPointer("bioExpression") ;
	WARNING(err->describe()) ;
	return patString();
      }

      leftResult = leftChild->getExpression(err) ;
      if (err != NULL) {
	WARNING(err->describe()) ;
	return patString();
      }
    
      result =  getOperatorName() ;
      result += '(' ;
      result += leftResult ;
      result += ')' ;
      return result ;
    }
  
  case LITERALS_OP :
    {
      patString res(getOperatorName()) ;
      res += " " ;
      return res ;
    }

  case SUM_OP :
  case PROD_OP:
  case INTEGRAL_OP:
  case DERIVATIVE_OP:
    {
      if (leftChild == NULL) {
	      err = new patErrNullPointer("bioExpression") ;
	      WARNING(err->describe()) ;
	      return patString();
      }

      leftResult = leftChild->getExpression(err) ;

      result =  getOperatorName() ;
      result += '(' ;
      result += leftResult ;
      result += ')' ;
      return result ;
    }

  case RANDOM_OP:
    return getOperatorName() ;

  case SEQUENCE_OP:
    // Overloaded
    break ;
  case DICTIONARY_OP:
    // Overloaded
    break ;
  case ENUM_OP:
    // Overloaded
    break ;
  case CONSTANT_OP :
    {
      stringstream str ;
      str << getValue(err) ;
      return patString(str.str()) ;
      break ;
    }

  default:
    {
    err = new patErrMiscError("Operator not defined");
    return patString();
    }
  }
}

patString bioExpression::getInfo() const {
  stringstream str ;
  if (isTop()) {
    str << "[" << *this << "]"  ;
    return patString(str.str());
  }
  str << "[" << *this << "] within [" << *getRoot() << "]" ;
  return patString(str.str()) ;
}

patBoolean bioExpression::dependsOf(patULong aLiteralId) const {
  patError* err(NULL) ;
  if (leftChild != NULL) {
    if (leftChild->dependsOf(aLiteralId)) {
      return patTRUE ;

    }
  }
  if (rightChild != NULL) {
    if (rightChild->dependsOf(aLiteralId)) {
      return patTRUE ;
    }
  }
  return patFALSE ;
}


patULong bioExpression::getId() const {
  return theId ;
}


patString bioExpression::getCppResult(patString prefix) const {
  stringstream str ;
  str << prefix << "_r" << theId ;
  return patString(str.str()) ;
}

patString bioExpression::getGslFunction(patString prefix, 
					patULong literalId, 
					patULong parentId,
					patError*& err)  {
  stringstream str ;
  str << " patReal  bioMyGhFunction_" << prefix << parentId << "::getValue(patReal x) {" << endl ;
  patString theLiteral = bioLiteralRepository::the()->getCppCode(literalId,err) ;
  if (err != NULL) {
    WARNING(err->describe()) ;
    return patString();
  }
  str << "double " << theLiteral << "= x ;" << endl ;
  str << "   patVariables* __x(theParams->__x) ;" << endl ;
  str << "   vector< vector<patReal> >* __draw(theParams->__draw);" << endl ;
  str << "   trVector* __beta(theParams->__beta) ;" << endl ;
  str << "   bioSample* sample(theParams->sample) ;" << endl ;
  str << "   bioIteratorSpan __theThreadSpan(theParams->__theThreadSpan) ;" << endl ;
  
  if (isSimple()) {
    str << " return " << getSimpleCppCode(err) << ";" << endl ;
    if (err != NULL) {
      WARNING(err->describe()) ;
      return patString();
    }
  }
  else {
    str << getCppCode(prefix,err) << endl ;
    if (err != NULL) {
      WARNING(err->describe()) ;
      return patString() ;
    }
    str << " return("<< getCppResult(prefix) <<") ;" << endl ;
    
  }
  str << "}" << endl  ;
  return patString(str.str()) ;
}

void bioExpression::generateGslFunctions(ostream& str_h,patString prefix,patError*& err) {
  
  if (leftChild != NULL) {
    leftChild->generateGslFunctions(str_h,prefix,err) ;
    if (err != NULL) {
      WARNING(err->describe()) ;
      return ;
    }
  }
  if (rightChild != NULL) {
    rightChild->generateGslFunctions(str_h,prefix,err) ;
    if (err != NULL) {
      WARNING(err->describe()) ;
      return ;
    }
  }
}



patString bioExpression::getCppResultDeclaration(patString prefix) const {
  stringstream str ;
  str << "patReal " << getCppResult(prefix) ;
  return patString(str.str()) ;
}

patBoolean bioExpression::isSimple() const {
  if (leftChild != NULL) {
    if (!leftChild->isSimple()) {
      return patFALSE ;
    }
  }
  if (rightChild != NULL) {
    if (!rightChild->isSimple()) {
      return patFALSE ;
    }
  }
  
  // ONLY FOR DEBUG
//   static patBoolean first = patTRUE ;
//   if (first) {
//     WARNING("******* MUST BE CHANGED. THIS IS FOR DEBUG ****") ;
//     first = patFALSE ;
//   }
//   return patFALSE ;
  /////

    return patTRUE ;
}

void bioExpression::forceCodeGeneration() {
  codeMustBeGenerated = patTRUE ;
  if (leftChild != NULL) {
    leftChild->forceCodeGeneration() ;
  }
  if (rightChild != NULL) {
    rightChild->forceCodeGeneration() ;
  }
}

patBoolean bioExpression::isStructurallyZero() const {
  return patFALSE ;
}

void bioExpression::simplifyZeros(patError*& err) {
  if (leftChild != NULL) {
    if (leftChild->isStructurallyZero()) {
      //      DELETE_PTR(leftChild) ;
      leftChild = new bioArithConstant(this,0) ;
    }
    else if (leftChild->isConstant()) {
      patReal value = leftChild->getValue(err) ;
      if (err != NULL) {
	WARNING(err->describe()) ;
	return ;
      }
      leftChild = new bioArithConstant(this,value) ;
    }
    else {
      leftChild->simplifyZeros(err) ;
      if (err != NULL) {
	WARNING(err->describe()) ;
	return ;
      }
    }
  }
  if (rightChild != NULL) {
    if (rightChild->isStructurallyZero()) {
      //      DELETE_PTR(rightChild) ;
      rightChild = new bioArithConstant(this,0) ;
    }
    else if (rightChild->isConstant()) {
      patReal value = rightChild->getValue(err) ;
      if (err != NULL) {
	WARNING(err->describe()) ;
	return ;
      }
      rightChild = new bioArithConstant(this,value) ;
    }
    else {
      rightChild->simplifyZeros(err) ;
      if (err != NULL) {
	WARNING(err->describe()) ;
	return ;
      }
    }
  }
}

pair<patULong,bioExpression*> bioExpression::identifySimplification(patError*& err)  {

  map<patString,patULong> theSubExpressions ;
  map<patString,list<patULong> > theSubIds ;
  map<patString,list<bioExpression*> > thePointers ;

  countExpressions(&theSubExpressions,&theSubIds,&thePointers,err) ;
  if (err != NULL) {
    WARNING(err->describe()) ;
    return pair<patULong,bioExpression*>()  ;
  }

  // Identify the most approriate formula to move forward
  patULong maxPotentialSaving(0) ;
  bioExpression* theRepeatedExpression(NULL) ;
  for (map<patString,patULong>::iterator i = theSubExpressions.begin() ;
       i != theSubExpressions.end() ;
       ++i) {
    bioExpression* theExpr = getExpressionFromId(*(theSubIds[i->first].begin())) ;
    patULong repetitionsMinusOne(patMax(patULong(theSubIds[i->first].size()-1),patULong(0))) ;
    if (theExpr == NULL) {
      stringstream str ;
      str << "Unknown ID for expression: " << *(theSubIds[i->first].begin()) ;
      err = new patErrMiscError(str.str()) ;
      WARNING(err->describe()) ;
      return pair<patULong,bioExpression*>() ;
    }
    if (theExpr->getExpressionString() != i->first) {
      stringstream str ;
      str << "There is an inconsistency wrt to the expressions ids. Expression " << *(theSubIds[i->first].begin()) << " is " << i->first << " and should be " << theExpr->getExpression(err) ;
      err = new patErrMiscError(str.str()) ;
      WARNING(err->describe()) ;
      return pair<patULong,bioExpression*>() ;
    }

    patULong nbrOper = theExpr->getNumberOfOperations() ;
    patULong potentialSaving =  (nbrOper <= 1)?0:nbrOper * repetitionsMinusOne ;
    //    DEBUG_MESSAGE("Expression [" << repetitionsMinusOne << ","<<nbrOper<<","<<potentialSaving<<"]=" << *theExpr) ;
    if (potentialSaving > maxPotentialSaving) {
      maxPotentialSaving = potentialSaving ;
      theRepeatedExpression = theExpr ;
    }
  }
  return pair<patULong,bioExpression*>(maxPotentialSaving, theRepeatedExpression) ;
}

bioIteratorInfo* bioExpression::theIterator() {
  return NULL ;
}

void bioExpression::setAdditionalCode(patString s) {
  additionalCode = s ;
}

patULong bioExpression::getNumberOfOperations() const {
  patULong result = 1 ;
  if (leftChild != NULL) {
    result += leftChild->getNumberOfOperations() ;
  }
  if (rightChild != NULL) {
    result += rightChild->getNumberOfOperations() ;
  }
  return result ;
}

patString bioExpression::getIncludeStatements(patString prefix) const {
  stringstream str ;
  if (leftChild != NULL) {
    str << leftChild->getIncludeStatements(prefix) ;
  }
  if (rightChild != NULL) {
    str << rightChild->getIncludeStatements(prefix) ;
  }
  return patString(str.str()) ;
}

patBoolean bioExpression::operator!=(const bioExpression& x) {
  return (!(*this == x)) ;
}

patBoolean bioExpression::operator==(const bioExpression& x) {
  patError* err(NULL) ;
  patString me = getExpressionString() ;
  patString him = x.getExpressionString() ;
  return (me == him) ;

}

// We consider the first expression that does not contain an
// iterator. Its ID is stored in "theRootId". It must be initialized
// to patBadId. Then we identify all subexpressions, count the number
// of times their appea, and store their IDs.

// It returns the number of operations in the expression

patULong bioExpression::countExpressions(map<patString,patULong>* theExpressions, 
					 map<patString,list<patULong> >* theIds, 
					 map<patString,list<bioExpression*> > * thePointers,
					 patError*& err) {
  
  if (containsAnIterator()) {
    stringstream str ;
    str << "Expression to simplify cannot contain an iterator: " << *this ;
    err = new patErrMiscError(str.str()) ;
    WARNING(err->describe()) ;
    return 0;
  }
  switch (getOperatorType()) {
  case bioExpression::SEQUENCE_OP: 
  case bioExpression::MERGED_OP: 
  case bioExpression::DICTIONARY_OP:
    {
      err = new patErrMiscError("Should not be called. Should be overloaded") ;
      WARNING(err->describe()) ;
      return patULong() ;
    }
  default:
    ;
    //    DEBUG_MESSAGE("***** Simplify: " << *theExprToSimplify) ;
    
  }
  patULong result(0) ;
  if (getNumberOfOperations() > 0) {
    stringstream str ;
    str << getExpressionString() ;
    (*theExpressions)[patString(str.str())] += 1 ;
    (*theIds)[patString(str.str())].push_back(getId()) ;
    (*thePointers)[patString(str.str())].push_back(this) ;
    result = getNumberOfOperations() ;
  }
  if (leftChild != NULL) {
    leftChild->countExpressions(theExpressions,theIds,thePointers,err) ;
    if (err != NULL) {
      WARNING(err->describe()) ;
      return patULong() ;
    }
  }
  if (rightChild != NULL) {
    rightChild->countExpressions(theExpressions,theIds,thePointers,err) ;
    if (err != NULL) {
      WARNING(err->describe()) ;
      return patULong() ;
    }
  }
  return result ;
}

patBoolean bioExpression::containsAnIntegral() const {
  bioOperatorType theType = getOperatorType() ;
  if (theType == INTEGRAL_OP) {
    return patTRUE ;
  }
  patBoolean result(patFALSE) ;
  if (leftChild != NULL) {
    if (leftChild->containsAnIntegral()) {
      result = patTRUE ;
      return result ;
    }
  }
  if (rightChild != NULL) {
    if (rightChild->containsAnIntegral()) {
      result = patTRUE ;
    }
  }
  return result ;
}


patBoolean bioExpression::containsAnIterator() const {

  bioOperatorType theType = getOperatorType() ;
  switch (theType) {
  case SUM_OP:
  case PROD_OP:
  case MERGED_OP:
  case INTEGRAL_OP:
    return patTRUE ;
  default:
    patBoolean result(patFALSE) ;
    if (leftChild != NULL) {
      if (leftChild->containsAnIterator()) {
	result = patTRUE ;
	return result ;
      }
    }
    if (rightChild != NULL) {
      if (rightChild->containsAnIterator()) {
	result = patTRUE ;
      }
    }
    return result ;
  }
}


patBoolean bioExpression::containsASequence() const {
  bioOperatorType theType = getOperatorType() ;
  switch (theType) {
  case SEQUENCE_OP:
    return patTRUE ;
  default:
    patBoolean result(patFALSE) ;
    if (leftChild != NULL) {
      if (leftChild->containsASequence()) {
	result = patTRUE ;
	return (result) ;
      }
    }
    if (rightChild != NULL) {
      if (rightChild->containsASequence()) {
	result = patTRUE ;
      }
    }
    return result ;
  }
}



bioExpression* bioExpression::getExpressionFromId(patULong theId) {
  if (theId == getId()) {
    return this ;
  }
  bioExpression* result(NULL) ;
  if (leftChild != NULL) {
    result = leftChild->getExpressionFromId(theId) ;
    if (result != NULL) {
      return result ;
    }
  }
  if (rightChild != NULL) {
    result = rightChild->getExpressionFromId(theId) ;
    if (result != NULL) {
      return result ;
    }
  }
  return NULL ;

}

void bioExpression::replaceExpression(bioExpression* theExpression,
				      bioExpression* anotherExpression,
				      patError*& err) {

  
  replaceLeftExpression(theExpression,anotherExpression,err) ;
  if (err != NULL) {
    WARNING(err->describe()) ;
    return ;
  }

  replaceRightExpression(theExpression,anotherExpression,err) ;
  if (err != NULL) {
    WARNING(err->describe()) ;
    return ;
  }


}

void bioExpression::replaceLeftExpression(bioExpression* theExpression,
				      bioExpression* anotherExpression,
				      patError*& err) {


  if (leftChild != NULL) {
    if (*leftChild == *theExpression) {
      //      DELETE_PTR(leftChild) ;
      leftChild = anotherExpression->getDeepCopy(err) ;
      if (err != NULL) {
	WARNING(err->describe()) ;
	return ;
      }
      leftChild->parent = this ;
      return ;
    }
    else {
      leftChild->replaceExpression(theExpression,anotherExpression,err) ;
      if (err != NULL) {
	WARNING(err->describe()) ;
	return ;
      }
    }
  }
}

void bioExpression::replaceRightExpression(bioExpression* theExpression,
				      bioExpression* anotherExpression,
				      patError*& err) {


  if (rightChild != NULL) {
    if (*rightChild == *theExpression) {
      //      DELETE_PTR(rightChild) ;
      rightChild = anotherExpression->getDeepCopy(err) ;
      if (err != NULL) {
	WARNING(err->describe()) ;
	return ;
      }
      rightChild->parent = this ;
      return ;
    }
    else {
      rightChild->replaceExpression(theExpression,anotherExpression,err) ;
      if (err != NULL) {
	WARNING(err->describe()) ;
	return ;
      }
    }
  }
}


patBoolean bioExpression::isConstant() const {

  switch (getOperatorType()) {
  case BINARY_OP :
  case UNARY_OP:
    if (leftChild != NULL) {
      if (!leftChild->isConstant()) {
	return patFALSE ;
      }
    }
    if (rightChild != NULL) {
      if (!rightChild->isConstant()) {
	return patFALSE ;
      }
    }
    return patTRUE ;
  case LITERALS_OP:
  case RANDOM_OP:
  case SUM_OP:
  case PROD_OP:
  case MERGED_OP:
  case SEQUENCE_OP:
  case DICTIONARY_OP:
  case UNDEFINED_OP:
    return patFALSE ;
  case CONSTANT_OP:
    return patTRUE ;
  default:
    return patFALSE ;
  }
}


patString bioExpression::getExpressionString() const {
  switch (getOperatorType()) {
  case LITERALS_OP:
  case RANDOM_OP:
  case SUM_OP:
  case PROD_OP:
  case MERGED_OP:
  case SEQUENCE_OP:
  case DICTIONARY_OP:
  case CONSTANT_OP:
    
  case BINARY_OP :
  case UNARY_OP:
  case UNDEFINED_OP:
  default:
  stringstream str ;
  str << getOperatorType() ;
  if (leftChild != NULL) {
    str << leftChild << getExpressionString() ;
  }
  if (rightChild != NULL) {
    str << rightChild << getExpressionString() ;
  }
  return patString(str.str()) ;
  }
}


bioExpression* simplify(bioExpression* theExprToSimplify, patError*& err) {
  
  short simplifyExpressions = bioParameters::the()->getValueInt("simplifyExpressions",err) ;
  if (err != NULL) {
    WARNING(err->describe()) ;
    return NULL ;
  }
  if (simplifyExpressions == 0) {
    return theExprToSimplify ;
  }

  if (theExprToSimplify == NULL) {
    err = new patErrNullPointer("bioExpression") ;
    WARNING(err->describe()) ;
    return NULL ;
  }

  // Keep original for debugging purposes
  patString originalExpression(theExprToSimplify->getExpression(err)) ;
  
  if (!theExprToSimplify->canBeSimplified()) {
    stringstream str ;
    str << "Expression to simplify cannot contain an iterator or a sequence: " << *theExprToSimplify;
    err = new patErrMiscError(str.str()) ;
    WARNING(err->describe()) ;
    return NULL;
  }

  patBoolean continueSimplifying(patTRUE) ;

  patULong savings(0) ;
  patULong potSavings(0) ;
  patULong before = theExprToSimplify->getNumberOfOperations() ;
  //  DEBUG_MESSAGE("Expression to simplify: " << *theExprToSimplify) ;
//      DEBUG_MESSAGE("Type: " << theExprToSimplify->getOperatorType()) ;
//      DEBUG_MESSAGE("String: " << theExprToSimplify->getExpressionString()) ;
  pair<patULong,bioExpression*> potential = theExprToSimplify->identifySimplification(err) ;
  if (err != NULL) {
    WARNING(err->describe()) ;
    return patULong() ;
  }
  
  if (potential.second != NULL) {
    potSavings = potential.first ;
    savings += potSavings ;
    if (potSavings > 0) {
      bioExpression* theSequence = 
	new bioArithSequence(NULL,potential.second,theExprToSimplify,err) ;
      if (err != NULL) {
	WARNING(err->describe()) ;
	return NULL ;
      }
      theSequence->simplifyLowerIteratorOrSequence(err) ;
      if (err != NULL) {
	WARNING(err->describe()) ;
	return NULL ;
      }
      theExprToSimplify = theSequence ;
      
    }
  }

  patULong after = theExprToSimplify->getNumberOfOperations() ;
  if (before < after) {
    stringstream str ;
    str << "Something's is wrong in the simplification procedure with potential savings: " << savings << endl ;
    str << "Before [" << before << "]: " << originalExpression << endl ;
    str << "After [" << after << "]: " << *theExprToSimplify << endl ;
    err = new patErrMiscError(str.str()) ;
    WARNING(err->describe()) ;
  }
  
  //DEBUG_MESSAGE("***** Simplified: " << *theExprToSimplify) ;
  return(theExprToSimplify) ;

}

void bioExpression::getExpressionsToSimplify(vector<bioExpression*>* listOfExpr, 
					     patError*& err) {
  
  if (listOfExpr == NULL) {
    err = new patErrNullPointer("vector<bioExpression*>") ;
    WARNING(err->describe()) ;
    return ;
  }

  switch (getOperatorType()) {
  case bioExpression::MERGED_OP:
  case bioExpression::DICTIONARY_OP:
    {
      err = new patErrMiscError("This function must be overloaded for this operator") ;
      WARNING(err->describe()) ;
      return ;
    }
  case bioExpression::UNDEFINED_OP:
    {
      stringstream str ;
      str << "Unknown operator in expression: " << *this ;
      err = new patErrMiscError(str.str()) ;
      WARNING(err->describe()) ;
      return ;
    }
  case bioExpression::SUM_OP:
  case bioExpression::PROD_OP:
    {
      if (leftChild == NULL) {
	err = new patErrNullPointer("bioExpression") ;
	WARNING(err->describe()) ;
	return ;
      }
      if (leftChild->canBeSimplified()) {
	// We have found one iterator at the lowest level
	listOfExpr->push_back(this) ;
	return ;
      }
      else {
	leftChild->getExpressionsToSimplify(listOfExpr,err) ;
	if (err != NULL) {
	  WARNING(err->describe()) ;
	  return ;
	}
	return ;
      }
      break ;
    }
  case bioExpression::SEQUENCE_OP:
    {
      if (leftChild == NULL) {
	err = new patErrNullPointer("bioExpression") ;
	WARNING(err->describe()) ;
	return ;
      }
      if (leftChild->canBeSimplified()) {
	// We have found one iterator at the lowest level
	listOfExpr->push_back(this) ;
	return ;
      }
      else {
	leftChild->getExpressionsToSimplify(listOfExpr,err) ;
	if (err != NULL) {
	  WARNING(err->describe()) ;
	  return ;
	}
      }
      if (rightChild == NULL) {
	err = new patErrNullPointer("bioExpression") ;
	WARNING(err->describe()) ;
	return ;
      }
      if (rightChild->canBeSimplified()) {
	// We have found one iterator at the lowest level
	listOfExpr->push_back(this) ;
	return ;
      }
      else {
	rightChild->getExpressionsToSimplify(listOfExpr,err) ;
	if (err != NULL) {
	  WARNING(err->describe()) ;
	  return ;
	}
      }
      return ;
    }
  default:
    {
      if (leftChild != NULL) {
	leftChild->getExpressionsToSimplify(listOfExpr,err) ;
	if (err != NULL) {
	  WARNING(err->describe()) ;
	  return ;
	}
      }
      if (rightChild != NULL) {
	rightChild->getExpressionsToSimplify(listOfExpr,err) ;
	if (err != NULL) {
	  WARNING(err->describe()) ;
	  return ;
	}
      }
    }
  }
}

void bioExpression::simplifyLowerIteratorOrSequence(patError*& err) {
  err = new patErrMiscError("This function can be called only from an iterator or a sequence") ;
  WARNING(err->describe()) ;
  return ;
}

patBoolean bioExpression::isSequence() const {
  return patFALSE ;
}
patBoolean bioExpression::canBeSimplified() const {
  if (containsAnIterator()) {
    return patFALSE ;
  }
  if (containsASequence()) {
    return patFALSE ;
  }
  return patTRUE ;
}


