//-*-c++-*------------------------------------------------------------
//
// File name : bioExpression.h
// Author :    Michel Bierlaire and Mamy Fetiarison
// Date :      Mon Apr 27 16:31:01 2009
//
//--------------------------------------------------------------------

#ifndef bioExpression_h
#define bioExpression_h

#include "patError.h"
#include <map>
#include <list>
#include <ostream>

#include "bioIteratorInfo.h"

/*!
 This virtual class represents a generic arithmetic expression.
 Most expressions are based on the leftChild/righChild representation, except for
 - bioArithMergedSum
 - bioArithElem

*/

class bioLiteral ;

class bioExpression {

  friend ostream& operator<<(ostream& str, const bioExpression& x) ;

 public:

  typedef  enum {
    UNARY_OP,
    BINARY_OP,
    LITERALS_OP,
    RANDOM_OP,
    CONSTANT_OP,
    SUM_OP,
    PROD_OP,
    ENUM_OP,
    MERGED_OP,
    SEQUENCE_OP,
    DICTIONARY_OP,
    INTEGRAL_OP,
    DERIVATIVE_OP,
    UNDEFINED_OP,
  } bioOperatorType ;


  bioExpression(bioExpression* par,
		bioExpression* left, 
		bioExpression* right) ;

  virtual patBoolean operator==(const bioExpression& x) ;
  virtual patBoolean operator!=(const bioExpression& x) ;

  virtual ~bioExpression() ;

  virtual bioOperatorType getOperatorType() const = PURE_VIRTUAL ;

  /*!
    \return patTRUE if the node has no parent
  */
  virtual patBoolean isTop() const  ;
  
  /*!
    return patTRUE if the expression does not contain iterators or vectors
    bioExpression returns patTRUE. The non trivial nodes overload the function.
   */
  virtual patBoolean isSimple() const ;

  /*!
    return patTRUE if the expression is structurally 0 so that there is no need to evaluate it.
  */
  virtual patBoolean isStructurallyZero() const ;
  
  /*!
    \return pointer to the node representing the left side of the expression
  */
  virtual bioExpression* getLeftChild() const ;

  /*!
    \return pointer to the node representing the right side of the expression
  */
  virtual bioExpression* getRightChild() const;

  /*!
    \param par pointer to the parent. 
  */
  void setParent(bioExpression* par) ;
  
  /*!
    \return pointer to the root of the current expression
  */
  virtual bioExpression* getRoot() const ;

  /**
     \return pointer to the iterator information if the top level node is an iterator, NULL otherwise. Not recursive. 
   */
  virtual bioIteratorInfo* theIterator() ;

  /*!
    \return pointer to the parent node
  */
  virtual bioExpression* getParent() const  ;

  /*!
    \return name of the operator
  */
  virtual patString getOperatorName() const = PURE_VIRTUAL ;
  
  /*!
    \return value of the expression
    \param err ref. of the pointer to the error object.
  */
  virtual patReal getValue(patError*& err) const = PURE_VIRTUAL ;

  /*!
    \return value of the derivative w.r.t literal
    \param index of the literal involved in the derivative
    \param err ref. of the pointer to the error object.
  */
  virtual bioExpression* getDerivative(patULong aLiteralId, 
				       patError*& err) const 
    = PURE_VIRTUAL ;

  /*!
     \return printed expression
   */
  virtual patString getExpression(patError*& err) const ;

 /*!
    Mainly for debugging
   */
  virtual patString getInfo() const ;

  /*!
     Get a deep copy of the expression. The user is in charge of
     releasing the memory.
   */
  virtual bioExpression* getDeepCopy(patError*& err) = PURE_VIRTUAL ;
  
  /*!
    code of the form "prefixresult = aFormula ;"
   */
  virtual patString getCppCode(patString prefix, patError*& err) = PURE_VIRTUAL ;

  /*!
    code with only the formula
   */
  virtual patString getSimpleCppCode(patError*& err) = PURE_VIRTUAL ;

  /*!  Defines a piece of code that will be appended to the generated
    code at the end of each loop of an iterator. If the node is not an
    iterator, it does nothing. It is typically designed to compute the
    BHHH matrix.
   */
  virtual void setAdditionalCode(patString s) ;

  /*!
   */
  virtual patBoolean dependsOf(patULong aLiteralId) const ;

  /*!
   */
  patULong getId() const;

  /*!
    Nameof the variable where the result of the expression in C++ is stored
   */
  patString getCppResult(patString prefix) const ;

  /*!
    Generate the code of all the GSL functions which are needed
  */
  virtual void generateGslFunctions(ostream& str_h,patString prefix,patError*& err) ;

  /*!
    Generate code for the GSL function. literalId identifies the literal corresponding to the variable of the function.
  */
  virtual patString getGslFunction(patString prefix, patULong literalId, patULong parentId, patError*& err)  ;

  /*!
    Declaration of the variable where the result of the expression in C++ is stored
   */
  patString getCppResultDeclaration(patString prefix) const ;

  /*!
    Force the generation of the cpp code for iterators
   */
  virtual void forceCodeGeneration() ;

  /*!
   */
  virtual patBoolean containsAnIterator() const ;

  /*!
   */
  virtual patBoolean containsAnIntegral() const ;

  /*!
   */
  virtual patBoolean containsASequence() const ;


  /*!
    An expression can be simplified if it does not contain an iterator or a sequence
   */
  virtual patBoolean canBeSimplified() const ;


  /**
     Identifies one potential simplification in the formula. For
     instance, log(3*a*b/c)+exp(3*a*b/c) involves 9 operations. We can
     replace it by y = 3*a*b/c ; log(y) + exp(y), which involves only 6
     operations.  This function would return the pair <2,3*a*b/c>,
     where the first member is the number of occurences of the repeated
     expression, and the second is the repeated expression.
   */
  pair<patULong,bioExpression*> identifySimplification(patError*& err)  ;

  /*
    Count the number of operations
   */
  virtual patULong getNumberOfOperations() const ;

  /*
    If an expression generates code requiring some include file, this function generates the include statements.
   */
  virtual patString getIncludeStatements(patString prefix) const ;

  virtual patULong countExpressions(map<patString,patULong>* theExpressions, 
				    map<patString,list<patULong> >* theIds,
				    map<patString,list<bioExpression*> >* thePointers,
				    patError*& err)  ; 
  

  /**
     The expressions to simplify must not contain any iterator. This
     function browse the expression tree and returns all iterators at
     the lowest level, meaning that the expression on which they
     iterate do not contain any iterator.
   */
  virtual void getExpressionsToSimplify(vector<bioExpression*>* theExpr, 
					patError*& err) ;


  virtual bioExpression* getExpressionFromId(patULong theId);

  /**
     Replace the expression by another expression. thisExpression
     cannot correspond to the current expression
   */
  virtual void replaceExpression(bioExpression* thisExpression, 
				 bioExpression* anotherExpression, 
				 patError*& err) ;

  /**
     Replace the expression in the left child by another expression.
   */
  virtual void replaceLeftExpression(bioExpression* thisExpression, 
				 bioExpression* anotherExpression, 
				 patError*& err) ;

  /**
     Replace the expression in the left child by another expression.
   */
  virtual void replaceRightExpression(bioExpression* thisExpression, 
				      bioExpression* anotherExpression, 
				      patError*& err) ;

  /*!
   */
  virtual void simplifyZeros(patError*& err) ;

  /**
   */
  virtual patBoolean isConstant() const ;

  /**
   */
  virtual patBoolean isSequence() const ;

  /**
     Compute a string that represents the expression. It is designed to replace the expression itself when used only for comparison purposes. 
     Code:
     +{expr1}{expr2}: binary plus
     -{expr1}{expr2}: binary minus
     *{expr1}{expr2}: multiplication
     /{expr1}{expr2}: division
     ^{expr1}{expr2}: power
     &{expr1}{expr2}: and
     |{expr1}{expr2}: or
     ={expr1}{expr2}: equal
     !={expr1}{expr2}: not equal
     <{expr1}{expr2}: lesser than
     <={expr1}{expr2}: lesser or equal to
     >{expr1}{expr2}: greater than
     >={expr1}{expr2}: greater or equal to 
     $A{expr}: abs
     $D[expr][{expr1}...{exprN}]: dictionary (bioArithElem)
     $E{expr}: exp
     $L{expr}: log
     $M{expr}: Unary minus
     $Piterator_name{expr}: prod
     $Q{string1}{string2}: sequence
     $Siterator_name{expr}: sum
     $Ziterator_name[{expr1}...{exprN}]: merged sum
     number: constant
     #id: literal
     &id: random
   */
  virtual patString getExpressionString() const = PURE_VIRTUAL ;
  virtual void simplifyLowerIteratorOrSequence(patError*& err) ;
protected:

  bioExpression* leftChild ;
  bioExpression* rightChild ;
  bioExpression* parent ;

  patULong theId ;
  patBoolean codeMustBeGenerated ;

  patString additionalCode ;

};

bioExpression* simplify(bioExpression* theExprToSimplify, patError*& err) ;

#endif
