//-*-c++-*------------------------------------------------------------
//
// File name : bioGenerateLikelihoodFunction.h
// Author :    Michel Bierlaire and Mamy Fetiarison
// Date :      Fri Jul 17 10:12:50 2009
//
//--------------------------------------------------------------------

#include "patLargeNumbers.h"
#include "patMath.h"
#include "patErrMiscError.h"
#include "patDisplay.h"
#include "bioGenerateLikelihoodFunction.h"
#include "bioLiteralRepository.h"
#include "bioExpressionRepository.h"
#include "bioVersion.h"
#include "bioExpression.h"
#include "bioIteratorInfoRepository.h"
#include "bioArithMergedSum.h"
#include "bioCountOperations.h"
#include "bioParameters.h"

bioGenerateLikelihoodFunction::bioGenerateLikelihoodFunction(map<patString,patULong>* lit, 
							     bioExpression* anExpression,
							     map<patString,bioExpression*>* constraints,
							     patString m, patString f, patString g):
  theExpression(anExpression), 
  theConstraints(constraints),
  literals(lit),
  mainFileName(m),
  fctFileName(f), 
  gradFileName(g) {
 DETAILED_MESSAGE("Generate expression for " << literals->size() << " parameters"<< " to files " << f << " and " << g) ;
}

void bioGenerateLikelihoodFunction::generate(patError*& err) {
  

  patBoolean buildAnalyticalGradient = (bioParameters::the()->getValueInt("buildAnalyticalGradient",err) != 0) ;
  if (err != NULL) {
    WARNING(err->describe()) ;
    return ;
  }

  vector<patULong> theIds = 
    bioLiteralRepository::the()->getBetaIds(patFALSE,err) ;
  if (err != NULL) {
    WARNING(err->describe()) ;
    return ;
  }

  // Compute gradient

  short simplifyExpressions = bioParameters::the()->getValueInt("simplifyExpressions",err) ;
  if (err != NULL) {
    WARNING(err->describe()) ;
    return ;
  }

  bioCountOperations countingOperations ;
  countingOperations.fctBeforeSimplification = theExpression->getNumberOfOperations() ;
  bioExpression* theOriginalExpression = theExpression->getDeepCopy(err) ;
  if (simplifyExpressions != 0) {
    // Simplify the expression
    
    if (err != NULL) {
      WARNING(err->describe()) ;
      return ;
    }
    
    //  DEBUG_MESSAGE("BEFORE ["<<theExpression->getNumberOfOperations()<<"]: " << *theExpression) ;
    if (!theExpression->canBeSimplified()) {
      vector<bioExpression*> expressionsToSimplify ;
      theExpression->getExpressionsToSimplify(&expressionsToSimplify,err) ;
      if (err != NULL) {
	WARNING(err->describe()) ;
	return ;
      }
      //      DEBUG_MESSAGE("Number of expressions found: " << expressionsToSimplify.size()) ;
      for (vector<bioExpression*>::iterator i = expressionsToSimplify.begin() ;
	   i != expressionsToSimplify.end() ;
	   ++i) {
	(*i)->simplifyLowerIteratorOrSequence(err) ;
	if (err != NULL) {
	  WARNING(err->describe()) ;
	  return ;
	}
      }
    }
    else {
      // Simplify expression
      bioExpression* newExpression = simplify(theExpression,err) ;
      if (err != NULL) {
	WARNING(err->describe()) ;
	return ;
      }
      theExpression = newExpression ;
    }
    //  DEBUG_MESSAGE("AFTER ["<<theExpression->getNumberOfOperations()<<"]: " << *theExpression) ;
    countingOperations.fctAfterSimplification = theExpression->getNumberOfOperations() ;

  }
  
  stringstream str ;
  ofstream mk("files.mk",ios::app) ;
  mk << mainFileName << " " ;
  mk.close() ;
  

  ofstream f(mainFileName.c_str(),ios::out|ios::app) ;
  writeHeaders(f,err) ;
  if (err != NULL) {
    WARNING(err->describe()) ;
    return ;
  }
  f << " #include <pthread.h> " << endl ; //POSIX Threads library
  f << " void *computeFunction( void *ptr );" << endl ;
  f << " void *computeFunctionAndGradient( void *ptr );" << endl ;
  f << "" << endl ;
  f << "// Definition of the class" << endl ;
  f << "#include \"bioMyObjects.h\"" << endl ;
  f << "" << endl ;
  f << "" << endl ;
  f << "// Implementation of the class" << endl ;
  f << "" << endl ;
  patULong n = bioLiteralRepository::the()->getNumberOfEstimatedParameters()  ;
  DEBUG_MESSAGE("Number of estimated parameters: " << n) ;
  f << "  myBiogeme::myBiogeme() : bhhh(theTrParameters,"<< n << ")" << endl ;
  f << ", hessian(theTrParameters,"<< bioLiteralRepository::the()->getNumberOfEstimatedParameters() << ")" << endl ;
  f << " , threadGrad(NUMTHREADS,trVector("<< bioLiteralRepository::the()->getNumberOfEstimatedParameters() << "))" << endl ;
  f << " , threadBhhh(NUMTHREADS,trHessian(theTrParameters," << bioLiteralRepository::the()->getNumberOfEstimatedParameters() << " ))" << endl ;
  f << " , threadHessian(NUMTHREADS,trHessian(theTrParameters," << bioLiteralRepository::the()->getNumberOfEstimatedParameters() << " ))" << endl ;
  f << "    { " << endl ;
  f << "  bhhh.setToZero() ;" << endl ;
  f << "  hessian.setToZero() ;" << endl ;
  f << "}" << endl ;
  f << "" << endl ;
  f << "" << endl ;
  
  f << "patReal myBiogeme::getFunction(trVector* x," << endl ;
  f << "			       patBoolean* success," << endl ;
  f << "			       patError*& err) {" << endl ;
  f << "" << endl ;
  f << "*success = patTRUE ;" << endl ;
  f << "  bioLiteralRepository::the()->setBetaValues(*x,err) ;" << endl ;
  f << "    if (err != NULL) {" << endl ;
  f << "      WARNING(err->describe()) ;" << endl ;
  f << "      return patReal();" << endl ;
  f << "    }" << endl ;
  f << "  trVector theBetas = bioLiteralRepository::the()->getBetaValues() ; " << endl ;
  
  f << "  patReal logLike = 0.0 ;" << endl ;
  f << "" << endl ;
  f << "  pthread_t *threads;" << endl ;
  f << "  threads = new pthread_t[NUMTHREADS];" << endl ;
  f << "  " << endl ;
  f << "  inputArg *input;" << endl ;
  f << "  input = new inputArg[NUMTHREADS];" << endl ;
  f << "  " << endl ;
  f << "  for(int i=0; i<NUMTHREADS; i++) {" << endl ;
  f << "  input[i].subsample = bioIteratorInfoRepository::the()->getSublist(i,err) ; " << endl ;  
  f << "    if (err != NULL) {" << endl ;
  f << "      WARNING(err->describe()) ;" << endl ;
  f << "      return patReal();" << endl ;
  f << "    }" << endl ;
  f << "  input[i].__beta = &theBetas ;" << endl ;
  f << "  input[i].grad = NULL ;" << endl ;
  f << "  input[i].bhhh = NULL ;" << endl ;
  f << "  input[i].success = success ;" << endl ;
  f << "  input[i].err = err ;" << endl ;
  f << "  input[i].sample = sample ;" << endl ;
  f << "  input[i].result = 0.0 ;" << endl ;
  f << "    if(pthread_create( &threads[i], NULL, computeFunction, (void*) &input[i])) {" << endl ;
    f << "      err = new patErrMiscError(\"Error creating Threads! Exit!\") ;" << endl ;
    f << "      WARNING(err->describe()) ;" << endl ;
    f << "      return 0.0;" << endl ;
    f << "    }" << endl ;
    f << "  }" << endl ;
    f << "  void *tmp;" << endl ;
    f << "  " << endl ;
    f << " logLike = 0.0 ;" << endl ;
    f << "  patBoolean maxreal(patFALSE) ;" << endl ;
    f << "  for(int i=0; i<NUMTHREADS; ++i) {" << endl ;
    f << "    pthread_join( threads[i], &tmp);" << endl ;
    f << " // Generated at " << __FILE__ << " " << __LINE__  << endl ;
    f << "      if (!maxreal) {" << endl ;
    f << "	if (input[i].result == -patMaxReal) {" << endl ;
    f << "	  logLike = patMaxReal ;" << endl ;
    f << "	  maxreal = patTRUE ;" << endl ;
    f << "	}" << endl ;
    f << "	else {" << endl ;
    f << "    logLike -= input[i].result ;" << endl ;
    f << "	}" << endl ;
    f << "      }" << endl ;
    f << "  }" << endl ;
    f << "	" << endl ;
    f << "    if (!finite(logLike)) {" << endl ;
    f << "      return patMaxReal ;" << endl ;
    f << "    }" << endl ;
    f << "  return logLike ;" << endl ;
  f << "}" << endl ;
  f << "  " << endl ;
  f << "" << endl ;
  f << "" << endl ;
  f << "trVector* myBiogeme::getGradient(trVector* x," << endl ;
  f << "				 trVector* grad," << endl ;
  f << "				 patBoolean* success," << endl ;
  f << "				 patError*& err) {" << endl ;
  f << "*success = patTRUE ;" << endl ;
  f << "  patReal result = getFunctionAndGradient(x,grad,success,err) ;" << endl ;
  f << "  if (err != NULL) {" << endl ;
  f << "    WARNING(err->describe()) ;" << endl ;
  f << "    return NULL  ;" << endl ;
  f << "  }" << endl ;
  f << "  return grad ;" << endl ;
  f << "}" << endl ;
  f << "  " << endl ;
  
  f << "patReal myBiogeme::getFunctionAndGradient(trVector* x," << endl ;
  f << "					  trVector* grad," << endl ;
  f << "					  patBoolean* success," << endl ;
  f << "					  patError*& err) {" << endl ;
  f << "*success = patTRUE ;" << endl ;


    f << "bhhh.setToZero() ;" << endl ;
    f << "  bioLiteralRepository::the()->setBetaValues(*x,err) ;" << endl ;
    f << "    if (err != NULL) {" << endl ;
    f << "      WARNING(err->describe()) ;" << endl ;
    f << "      return patReal();" << endl ;
    f << "    }" << endl ;

    if (buildAnalyticalGradient) {

    f << "  trVector theBetas = bioLiteralRepository::the()->getBetaValues() ; " << endl ;
    f << "  patReal logLike = 0.0 ;" << endl ;
    f << "" << endl ;
    f << "  fill(grad->begin(),grad->end(),0.0) ; " << endl ;
    f << "  pthread_t *threads;" << endl ;
    f << "  threads = new pthread_t[NUMTHREADS];" << endl ;
    f << "  " << endl ;
    f << "  inputArg *input;" << endl ;
    f << "  input = new inputArg[NUMTHREADS];" << endl ;
    f << "  " << endl ;
    f << "  for(int i=0; i<NUMTHREADS; i++) {" << endl ;
    f << "  input[i].subsample = bioIteratorInfoRepository::the()->getSublist(i,err) ; " << endl ;  
    f << "    if (err != NULL) {" << endl ;
    f << "      WARNING(err->describe()) ;" << endl ;
    f << "      return patReal();" << endl ;
    f << "    }" << endl ;
    f << "    input[i].__beta = &theBetas ;" << endl ;
    f << "    fill(threadGrad[i].begin(),threadGrad[i].end(),0.0) ;" << endl ;
    f << "    input[i].grad = &(threadGrad[i]) ;" << endl ;
    f << "    threadBhhh[i].setToZero() ;    input[i].bhhh = &(threadBhhh[i]) ;" << endl ;
    f << "    input[i].success = success ;" << endl ;
    f << "    input[i].err = err ;" << endl ;
    f << "    input[i].sample = sample ;" << endl ;
    f << "    input[i].result = 0.0 ;" << endl ;
    f << "    if(pthread_create( &threads[i], NULL, computeFunctionAndGradient, (void*) &input[i])) {" << endl ;
    f << "      err = new patErrMiscError(\"Error creating Threads! Exit!\") ;" << endl ;
    f << "      WARNING(err->describe()) ;" << endl ;
    f << "      return 0.0;" << endl ;
    f << "    }" << endl ;
    f << "  }" << endl ;
    f << "  " << endl ;
    f << "  void *tmp;" << endl ;
    f << "  " << endl ;
    f << "   patBoolean maxreal(patFALSE) ;" << endl; 
    f << "  logLike = 0.0 ; " << endl ;
    f << "  for(int i=0; i<NUMTHREADS; ++i) {" << endl ;
    f << "    pthread_join( threads[i], &tmp);" << endl ;
    f << " // Generated at " << __FILE__ << " " << __LINE__  << endl ;
    f << "    if (!maxreal) {" << endl ;
    f << "      if (input[i].result == -patMaxReal) {" << endl ;
    f << "	logLike = patMaxReal ;" << endl ;
    f << "	maxreal = patTRUE ;" << endl ;
    f << "      }" << endl ;
    f << "      else {" << endl ;
    f << "    logLike -= input[i].result ;" << endl ;
    f << "    (*grad) -= threadGrad[i] ;" << endl ;
    f << "    bhhh.add(1.0,threadBhhh[i],err) ;" << endl ;
    f << "    if (err != NULL) {" << endl ;
    f << "      WARNING(err->describe()) ;" << endl ;
    f << "      return 0.0;" << endl ;
    f << "    }" << endl ;
    f << "    }" << endl ;
    f << "    }" << endl ;
    f << "  }" << endl ;
    f << "	" << endl ;
    f << "    if (!finite(logLike)) {" << endl ;
    f << "      return patMaxReal ;" << endl ;
    f << "    }" << endl ;
    f << "  return logLike ;" << endl ;
    f << "" << endl ;
    }
    else {
      f << "  patReal result = getFunction(x,success,err) ;" << endl ;
      f << "  if (err != NULL) {" << endl ;
      f << "    WARNING(err->describe()) ;" << endl ;
      f << "    return NULL  ;" << endl ;
      f << "  }" << endl ;
      f << " getFinDiffGradient(x,grad,success,err) ;" << endl ;
      f << "  if (err != NULL) {" << endl ;
      f << "    WARNING(err->describe()) ;" << endl ;
      f << "    return NULL  ;" << endl ;
      f << "  }" << endl ;
      f << " return result ;" << endl ;
    }
    f << "}" << endl ;
    f << "" << endl ;
    f << "" << endl ;
  f << "trHessian* myBiogeme::computeHessian(patVariables* x," << endl ;
  f << "				     trHessian& hessian," << endl ;
  f << "				     patBoolean* success," << endl ;
  f << "				     patError*& err) {" << endl ;
  f << "    return computeFinDiffHessian(x,hessian,success,err) ;" << endl ;
  f << "}" << endl ;
  f << "" << endl ;
  f << "trHessian* myBiogeme::getCheapHessian(trVector* x," << endl ;
  f << "			     trHessian& hessian," << endl ;
  f << "			     patBoolean* success," << endl ;
  f << "				      patError*& err) {" << endl ;
  f << "    hessian = bhhh ;" << endl ;
  f << "    *success = patTRUE ;" << endl ;
  f << "    return &bhhh ;" << endl ;
  f << "}" << endl ;
  f << "" << endl ;
  f << "patBoolean myBiogeme::isCheapHessianAvailable() {" << endl ;
  f << "  return patTRUE ;" << endl ;
  f << "}" << endl ;
  f << "" << endl ;
  f << "trVector myBiogeme::getHessianTimesVector(trVector* x," << endl ;
  f << "					  const trVector* v," << endl ;
  f << "					  patBoolean* success," << endl ;
  f << "					  patError*& err) {" << endl ;
  f << "  return trVector() ;" << endl ;
  f << "}" << endl ;
  f << "" << endl ;
  f << "patBoolean myBiogeme::isGradientAvailable(patError*& err) const {" << endl ;
  f << "  return patTRUE ;" << endl ;
  f << "}" << endl ;
  f << "" << endl ;
  f << "patBoolean myBiogeme::isHessianAvailable(patError*& err)  const {" << endl ;
  f << "  return patTRUE ;" << endl ;
  f << "}" << endl ;
  f << "" << endl ;
  f << "patBoolean myBiogeme::isHessianTimesVectorAvailable(patError*& err) const {" << endl ;
  f << "  return patFALSE ;" << endl ;
  f << "}" << endl ;
  f << "" << endl ;
  f << "unsigned long myBiogeme::getDimension(patError*& err) const {" << endl ;
  f << "  return " << bioLiteralRepository::the()->getNumberOfEstimatedParameters() << " ;" << endl ;
  f << "}" << endl ;
  f << "" << endl ;
  f << "patBoolean myBiogeme::isUserBased() const {" << endl ;
  f << "  return patTRUE ;" << endl ;
  f << "}" << endl ;
  f << "" << endl ;
  f << "void myBiogeme::generateCppCode(patError*& err) {" << endl ;
  f << "  err = new patErrMiscError(\"This function should never be called\") ;" << endl ;
  f << "  WARNING(err->describe()) ;" << endl ;
  f << "  " << endl ;
  f << "}" << endl ;
  f << "" << endl ;
  f << "// Implementation of the default display  on a terminal" << endl ;
  f << "" << endl ;
  f << "patDisplay::patDisplay() : screenImportance(patImportance::patDEBUG), " << endl ;
  f << "			   logImportance(patImportance::patDEBUG)," << endl ;
  f << "			   logFile(patFileNames::the()->getLogFile().c_str()) {" << endl ;
  f << "" << endl ;
  f << "  patAbsTime now ;" << endl ;
  f << "  now.setTimeOfDay() ;" << endl ;
  f << "  logFile << \"This file has automatically been generated.\" << endl ;" << endl ;
  f << "  logFile << now.getTimeString(patTsfFULL) << endl ;" << endl ;
  f << "  logFile << bioVersion::the()->getCopyright() << endl ;" << endl ;
  f << "  logFile << endl ;" << endl ;
  f << "  logFile << bioVersion::the()->getVersionInfoDate() << endl ;" << endl ;
  f << "  logFile << bioVersion::the()->getVersionInfoAuthor() << endl ;" << endl ;
  f << "  logFile << endl ;" << endl ;
  f << "  logFile << flush ;" << endl ;
  f << "" << endl ;
  f << "}" << endl ;
  f << "" << endl ;
  f << "patDisplay::~patDisplay() {" << endl ;
  f << "  messages.erase(messages.begin(),messages.end()) ;" << endl ;
  f << "   logFile.close() ;" << endl ;
  f << "}" << endl ;
  f << "" << endl ;
  f << "patDisplay* patDisplay::the() {" << endl ;
  f << "  static patDisplay* singleInstance = NULL ;" << endl ;
  f << "  if (singleInstance == NULL) {" << endl ;
  f << "    singleInstance = new patDisplay ;" << endl ;
  f << "    assert(singleInstance != NULL) ;" << endl ;
  f << "  }" << endl ;
  f << "  return singleInstance ;" << endl ;
  f << "}" << endl ;
  f << "" << endl ;
  f << "void patDisplay::addMessage(const patImportance& aType," << endl ;
  f << "			    const patString& text," << endl ;
  f << "			    const patString& fileName," << endl ;
  f << "			    const patString& lineNumber) {" << endl ;
  f << "" << endl ;
  f << "  patMessage theMessage ;" << endl ;
  f << "" << endl ;
  f << "  theMessage.theImportance = aType ;" << endl ;
  f << "  theMessage.text = text ;" << endl ;
  f << "  theMessage.fileName = fileName ;" << endl ;
  f << "  theMessage.lineNumber = lineNumber ;" << endl ;
  f << "" << endl ;
  f << "  patAbsTime now ;" << endl ;
  f << "  now.setTimeOfDay() ;" << endl ;
  f << "  theMessage.theTime = now ;" << endl ;
  f << "" << endl ;
  f << "  if (aType <= screenImportance) {" << endl ;
  f << "    if (screenImportance < patImportance::patDEBUG) {" << endl ;
  f << "      cout << theMessage.shortVersion() << endl << flush  ;" << endl ;
  f << "    }" << endl ;
  f << "    else {" << endl ;
  f << "      cout << theMessage.fullVersion() << endl << flush  ;" << endl ;
  f << "    }" << endl ;
  f << "  }" << endl ;
  f << "  if (aType <= logImportance) {" << endl ;
  f << "    if (logImportance < patImportance::patDEBUG) {" << endl ;
  f << "      logFile << theMessage.shortVersion() << endl << flush  ;" << endl ;
  f << "    }" << endl ;
  f << "    else {" << endl ;
  f << "      logFile << theMessage.fullVersion() << endl << flush  ;" << endl ;
  f << "    }" << endl ;
  f << "  }" << endl ;
  f << "  messages.push_back(theMessage) ;" << endl ;
  f << "}" << endl ;
  f << "" << endl ;
  f << "void patDisplay::setScreenImportanceLevel(const patImportance& aType) {" << endl ;
  f << "  screenImportance = aType ;" << endl ;
  f << "}" << endl ;
  f << "  " << endl ;
  f << "void patDisplay::setLogImportanceLevel(const patImportance& aType) {" << endl ;
  f << "  logImportance = aType ;" << endl ;
  f << "}" << endl ;
  f << "" << endl ;
  f << "void patDisplay::initProgressReport(const patString message," << endl ;
  f << "			unsigned long upperBound) {" << endl ;
  f << "  " << endl ;
  f << "" << endl ;
  f << "}" << endl ;
  f << "" << endl ;
  f << "patBoolean patDisplay::updateProgressReport(unsigned long currentValue) {" << endl ;
  f << "  return patTRUE ;" << endl ;
  f << "}" << endl ;
  f << "" << endl ;
  f << "void patDisplay::terminateProgressReport() {" << endl ;
  f << "" << endl ;
  f << "}" << endl ;
  f << "" << endl ;
  f << "" << endl ;
  f << "" << endl ;
  f << "" << endl ;

  patString prefix("l") ;
  ofstream gslhfile("bioGslBiogeme.h",ios::app) ;

  theExpression->generateGslFunctions(gslhfile,prefix,err) ;
  if (err != NULL) {
    WARNING(err->describe()) ;
    return ;
  }
  gslhfile.close() ;

  
  f.close() ;
  mk.open("files.mk",ios::app) ;
  mk << fctFileName << " " ;
  mk.close() ;
  f.open(fctFileName.c_str()) ;
  writeHeaders(f,err) ;
  if (err != NULL) {
    WARNING(err->describe()) ;
    return ;
  }
  f << " void *computeFunction(void* fctPtr) {" << endl ;
  //  f << "   patVariables* __x ;" << endl ;
  f << "   vector< vector<patReal> >* __draw ;" << endl ;
  f << "   inputArg *input;" << endl ;
  f << "   input = (inputArg *) fctPtr;" << endl ;
  f << "   trVector* __beta =  input->__beta ;" << endl ;
  f << "   trVector* grad = input->grad ;" << endl ;
  f << "   patBoolean* success = input->success ;" << endl ;
  f << "   trHessian* theHessian = input->hessian ;" << endl ;
  f << "   trHessian* bhhh = input->bhhh ;" << endl ;
  f << "   patError*& err = input->err ;" << endl ;
  f << "   bioSample* sample = input->sample ;" << endl ;
  f << "  bioIteratorSpan __theThreadSpan = input->subsample ; " << endl ;
  f << "// The following is used to be sent to the bioGsl functions" << endl ;
  f << "  bioIteratorSpan __theCurrentSpan  ; " << endl ;
  


    
  if (theExpression->isSimple()){
    f << "   input->result = "<< theExpression->getSimpleCppCode(err) << " ;" << endl ;
    if (err != NULL) {
      WARNING(err->describe()) ;
    }
    f << "    if (!finite(input->result)) {" << endl ;
    f << "      input->result = patMaxReal ;" << endl ;
    f << "    }" << endl ;
  }
  else {
    f << theExpression->getCppCode(prefix,err) << endl ;
    if (err != NULL) {
      WARNING(err->describe()) ;
    }
      f << "   input->result = "<<theExpression->getCppResult(prefix) <<" ;" << endl ;
      f << "    if (!finite(input->result)) {" << endl ;
      f << "      input->result = patMaxReal ;" << endl ;
      f << "    }" << endl ;
  }

    f << " }" << endl ;

    f.close() ;

    if (buildAnalyticalGradient) {

      ofstream mk("files.mk",ios::app) ;
      mk << gradFileName << " " ;
      mk.close() ;
      f.open(gradFileName.c_str()) ;
      writeHeaders(f,err) ;
      if (err != NULL) {
	WARNING(err->describe()) ;
	return ;
      }
      
      theExpression->forceCodeGeneration() ;
      bioLiteralRepository::the()->resetAllFlags() ;
      
      patBoolean useSimplifiedExpressionForDerivatives(patFALSE) ;
      
      countingOperations.gradBeforeSimplification = 0 ;
      countingOperations.gradAfterSimplification = 0 ;
      vector<bioExpression*> theDerivatives ;
      for (patULong i = 0 ; i < theIds.size() ; ++i) {
	bioExpression* theDeriv(NULL) ;
	if (useSimplifiedExpressionForDerivatives) {
	  theDeriv = theExpression->getDerivative(theIds[i],err) ;
	}
	else {
	  theDeriv = theOriginalExpression->getDerivative(theIds[i],err) ;
	}
	//	GENERAL_MESSAGE("Derivative wrt " <<bioLiteralRepository::the()->getName(theIds[i],err) << " computed") ;
	if (err != NULL) {
	  WARNING(err->describe()) ;
	  return ;
	}

	countingOperations.gradBeforeSimplification +=  theDeriv->getNumberOfOperations() ;
	
	bioExpression* aClone = theDeriv->getDeepCopy(err) ;
	if (err != NULL) {
	  WARNING(err->describe()) ;
	  return ;
	}

	theDerivatives.push_back(aClone) ;
      }
      
      //      GENERAL_MESSAGE("Number of operations: " << endl << countingOperations) ;
      
      bioIteratorInfo* theIterator(NULL) ;
      patBoolean possibleToMerge(patTRUE) ;
      for (patULong i = 0 ; i < theIds.size() ; ++i) {
	if (theIterator == NULL && possibleToMerge) {
	  theIterator = theDerivatives[i]->theIterator() ;
	  if (theIterator == NULL) {
	    DEBUG_MESSAGE("The derivative is not an iterator. Do not merge. ") ;
	    possibleToMerge = patFALSE ;
	  }
	  else {
	    // Check that it is the same iterator
	    bioIteratorInfo* currentIter = theDerivatives[i]->theIterator() ;
	    if ((*currentIter) != (*theIterator)) {
	      WARNING("Different iterators. No merge") ;
	      possibleToMerge = patFALSE ;
	    }
	  }
	}
      }
      
      if (theIterator == NULL) {
	possibleToMerge = patFALSE ;
      }
      bioArithMergedSum* theMergedExpression(NULL) ;
      if (possibleToMerge) {
	vector<bioExpression*> theExpressions ;
	f << "// Computation of derivatives is merged" << endl ;
	vector<patString> theResults ;
	for (patULong i = 0 ; i < theIds.size() ; ++i) {
	  stringstream str ;
	  str << "(*input->grad)[" << i << "]" ;
	  theResults.push_back(patString(str.str())) ;
	  theExpressions.push_back(theDerivatives[i]->getLeftChild()) ;
	}
	patString prefix("g") ;
	
	
	theMergedExpression = 
	  new bioArithMergedSum(NULL,theIterator,theResults,theExpressions) ;
	
	// Additional code for the computation of BHHH
	stringstream str ;
	str << "    patULong n = " << bioLiteralRepository::the()->getNumberOfEstimatedParameters() <<" ;" << endl ;
	str << "	for (unsigned long k = 0 ; k < n  ; ++k) {" << endl ;
	str << "	  for (unsigned long kk = k ; kk < n ; ++kk) {" << endl ;
	str << "	    bhhh->addElement(k,kk,oneIterValues[k] *  oneIterValues[kk],err) ;" << endl ;
	str << "	  }" << endl ;
	str << "	}" << endl ;
	theMergedExpression->setAdditionalCode(patString(str.str())) ;
	
      }
    
      if (possibleToMerge) {
	f << theMergedExpression->getIncludeStatements(prefix) << endl ;
      }

      f << " void *computeFunction( void *ptr );" << endl ;
      f << "void *computeFunctionAndGradient(void* fctPtr) {" << endl ;
      f << "  " << endl ;
      //      f << "   patVariables* __x ;" << endl ;
      f << "     vector< vector<patReal> >* __draw ;" << endl ;
      f << "   inputArg *input;" << endl ;
      f << "   input = (inputArg *) fctPtr;" << endl ;
      f << "   trVector* __beta =  input->__beta ;" << endl ;
      f << "   trVector* grad = input->grad ;" << endl ;
      f << "   patBoolean* success = input->success ;" << endl ;
      f << "   trHessian* theHessian = input->hessian ;" << endl ;
      f << "   trHessian* bhhh = input->bhhh ;" << endl ;
      f << "   patError*& err = input->err ;" << endl ;
      f << "   bioSample* sample = input->sample ;" << endl ;
      f << "  bioIteratorSpan __theThreadSpan = input->subsample ; " << endl ;
      
      f << "  computeFunction(fctPtr) ;" << endl ;
      
      if (possibleToMerge) {
	f << theMergedExpression->getCppCode(prefix,err) ;
	if (err != NULL) {
	  WARNING(err->describe()) ;
	  return ;
	}
	// Add functions to the file for GSL integration

	ofstream gslhfile("bioGslBiogeme.h",ios::app) ;
	theMergedExpression->generateGslFunctions(gslhfile,prefix,err) ;
	if (err != NULL) {
	  WARNING(err->describe()) ;
	  return ;
	}
	gslhfile.close() ;
      

      }
      else {
	for (patULong i = 0 ; i < theIds.size() ; ++i) {
	  f << "// Derivative with respect to " << bioLiteralRepository::the()->getName(theIds[i],err) << endl ;
	  stringstream str ;
	  str << "g" <<  bioLiteralRepository::the()->getName(theIds[i],err) ;
	  prefix = patString(str.str()) ;
	  if (theDerivatives[i]->isSimple()) {
	    f << "   (*input->grad)[" << i << "] = "<<theDerivatives[i]->getSimpleCppCode(err) <<" ;" << endl ;
	    if (err != NULL) {
	      WARNING(err->describe()) ;
	      return ;
	    }
	    f << "    if (!finite(*input->grad[" << i << "])) {" << endl ;
	    f << "      input->result = patMaxReal ;" << endl ;
	    f << "    }" << endl ;
	  }
	  else {
	    f << theDerivatives[i]->getCppCode(prefix,err) << endl ;
	    if (err != NULL) {
	      WARNING(err->describe()) ;
	      return ;
	    }
	    f << "   (*input->grad)[" << i << "] = "<<theDerivatives[i]->getCppResult(prefix) <<" ;" << endl ;
	    f << "    if (!finite(*input->grad[" << i << "])) {" << endl ;
	    f << "      input->result = patMaxReal ;" << endl ;
	    f << "    }" << endl ;
	  }

	  // Add functions to the file for GSL integration
	  
	  ofstream gslhfile("bioGslBiogeme.h",ios::app) ;
	  theDerivatives[i]->generateGslFunctions(gslhfile,prefix,err) ;
	  if (err != NULL) {
	    WARNING(err->describe()) ;
	    return ;
	  }
	  gslhfile.close() ;


	}
      }
    
      f << "}  " << endl ;


      f.close() ;


    }
    else {
      // Create an empty file for the gadient
      f.open(gradFileName.c_str()) ;
      f.close() ;
    }
    // Constraints

    f.open("bio__constraints.h") ;
    patAbsTime now ;
    now.setTimeOfDay() ;
    f << "// File: bio__constraints.h" << endl ;
    f << "// This file has automatically been generated." << endl ;
    f << "// " << now.getTimeString(patTsfFULL) << endl ;
    f << "// " << bioVersion::the()->getCopyright() << endl ;
    f << endl ;
    f << "// " << bioVersion::the()->getVersionInfoDate() << endl ;
    f << "// " << bioVersion::the()->getVersionInfoAuthor() << endl ;
    f << endl ;
      
    if (theConstraints != NULL) {
      
      f << "#include \"bioConstraintWrapper.h\"" << endl ;
      for (map<patString,bioExpression*>::iterator c = theConstraints->begin() ;
	   c != theConstraints->end() ;
	   ++c) {
	f << "///////////////////////////////" << endl ;
	f << "// Constraint: " << *(c->second) << endl ;
	f << "class bio__constraint" << c->first << " : public bioConstraintWrapper { " << endl ;
	f << "    public:" << endl ;
	f << "      patReal getFunction(trVector* x," << endl ;
	f << "			  patBoolean* success," << endl ;
	f << "			  patError*& err) ;" << endl ;
	f << "      trVector* getGradient(trVector* x," << endl ;
	f << "			    trVector* grad," << endl ;
	f << "			    patBoolean* success," << endl ;
	f << "			    patError*& err) ;" << endl ;
	f << "      patBoolean isGradientAvailable(patError*& err) const  ;" << endl ;
	f << "" << endl ;
	f << "      unsigned long getDimension(patError*& err) const ;" << endl ;
	f << "" << endl ;
	f << "};" << endl ;
      }
    }
    else {
      f << endl ;
      f << "// No constraint." << endl ;
    }
    f.close() ;

    mk.open("files.mk",ios::app) ;
    mk << "bio__constraints.cc " ;
    mk.close() ;
    
    f.open("bio__constraints.cc") ;
    if (theConstraints != NULL) {
      patString contrPrefix("contr") ;
      patAbsTime now ;
      now.setTimeOfDay() ;
      f << "// File: bio__constraints.cc" << endl ;
      f << "// This file has automatically been generated." << endl ;
      f << "// " << now.getTimeString(patTsfFULL) << endl ;
      f << "// " << bioVersion::the()->getCopyright() << endl ;
      f << endl ;
      f << "// " << bioVersion::the()->getVersionInfoDate() << endl ;
      f << "// " << bioVersion::the()->getVersionInfoAuthor() << endl ;
      f << endl ;
    
      f << "#include \"bioMyObjects.h\"" << endl ;
      f << "#include \"bio__constraints.h\"" << endl ;
      f << "#include \"bioLiteralRepository.h\"" << endl ;
      for (map<patString,bioExpression*>::iterator c = theConstraints->begin() ;
	   c != theConstraints->end() ;
	   ++c) {
	f << "///////////////////////////////" << endl ;
	f << "// Constraint: " << *(c->second) << endl ;
	f << "      patReal bio__constraint" << c->first << "::getFunction(trVector* x," << endl ;
	f << "			  patBoolean* success," << endl ;
	f << "			  patError*& err) {" << endl ;
	f << "*success = patTRUE ;" << endl ;
	

	f << "  bioLiteralRepository::the()->setBetaValues(*x,err) ;" << endl ;
	f << "    if (err != NULL) {" << endl ;
	f << "      WARNING(err->describe()) ;" << endl ;
	f << "      return patReal();" << endl ;
	f << "    }" << endl ;
	f << "  trVector theBetas = bioLiteralRepository::the()->getBetaValues() ; " << endl ;
    
	f << " trVector* __beta = &theBetas ; " << endl ;
	f << "// Constraint: " << *(c->second) << endl ;
	bioExpression* theExpression = c->second ;
	f << "	patReal result ;" << endl ;
	if (theExpression->isSimple()){
	  f << "    result = " << theExpression->getSimpleCppCode(err) << " ;" << endl ;
	  if (err != NULL) {
	    WARNING(err->describe()) ;
	  }
	}
	else {
	  f << theExpression->getCppCode(contrPrefix,err) << endl ;
	  if (err != NULL) {
	    WARNING(err->describe()) ;
	  }
	  f << "   result = "<<theExpression->getCppResult(contrPrefix) <<" ;" << endl ;
	}

	  f << "return result ; " << endl ;
    
	f << "}" << endl ;

	// The following IDs correspond to the position in the
	// gradient vector. The "theIds" do not.
	map<patString,patULong>* theParamIds = bioLiteralRepository::the()->getIdsOfParameters() ;
	
	
	f << "      trVector* bio__constraint" << c->first << "::getGradient(trVector* x," << endl ;
	f << "			    trVector* grad," << endl ;
	f << "			    patBoolean* success," << endl ;
	f << "			    patError*& err) {" << endl ;
	f << "*success = patTRUE ;" << endl ;
	f << "  bioLiteralRepository::the()->setBetaValues(*x,err) ;" << endl ;
	f << "    if (err != NULL) {" << endl ;
	f << "      WARNING(err->describe()) ;" << endl ;
	f << "      return grad;" << endl ;
	f << "    }" << endl ;
	f << "  trVector theBetas = bioLiteralRepository::the()->getBetaValues() ; " << endl ;
	f << " trVector* __beta = &theBetas ; " << endl ;
	
	f << "	patReal result ;" << endl ;
	f << "// Constraint: " << *(c->second) << endl ;
	for (patULong i = 0 ; i < theIds.size() ; ++i) {
	  f << "// i = " << i << " Id = " << theIds[i] << endl ;
	  patString theName = bioLiteralRepository::the()->getName(theIds[i],err) ;	  f << "// Derivative wrt " << theName << endl ;
	  
	  bioExpression* theExpression = c->second->getDerivative(theIds[i],err) ;
	  // Should simplfiy the expression here
      
	  if (theExpression->isSimple()){
	    f << "    result = " << theExpression->getSimpleCppCode(err) << " ;" << endl ;
	    if (err != NULL) {
	      WARNING(err->describe()) ;
	    }
	  }
	  else {
	    f << theExpression->getCppCode(contrPrefix,err) << endl ;
	    if (err != NULL) {
	      WARNING(err->describe()) ;
	    }
	    f << "   result = "<<theExpression->getCppResult(contrPrefix) <<" ;" << endl ;
	  }
	  f << "	  ((*grad)["<<i<<"]) = result ;" << endl ;
	}
	f << "return grad ; " << endl ;
	f << "      }" << endl ;
	f << "      patBoolean bio__constraint" << c->first << "::isGradientAvailable(patError*& err) const {" << endl ;
	f << "    return patTRUE ;" << endl ;
	f << "}" << endl ;
	f << "" << endl ;
	f << "      unsigned long bio__constraint" << c->first << "::getDimension(patError*& err) const {" << endl ;
	f << "return " <<theIds.size() << ";" << endl ;
	f << "}" << endl ;
	f << "" << endl ;
      }
    }

    f << "void myBioConstraints::addConstraints() {" << endl ;
    
    f << "// Constraints" << endl ;
    f << "" << endl ;
    if (theConstraints != NULL) {
      for (map<patString,bioExpression*>::iterator c = theConstraints->begin() ;
	   c != theConstraints->end() ;
	   ++c) {
	f << "      bioConstraintWrapper* aConstraint"<<c->first<<" = new bio__constraint" << c->first << "() ;" << endl ;
	f << "	constraints.push_back(aConstraint" << c->first <<") ;" << endl ;
      }    
    }
  f << "    };" << endl ;

    f.close() ;
}




void bioGenerateLikelihoodFunction::writeHeaders(ofstream& f, patError*& err) {
  
  patULong nThreads = bioParameters::the()->getValueInt("numberOfThreads",err) ;
  if (err != NULL) {
    WARNING(err->describe()) ;
    return ;
  }
  f << " #define NUMTHREADS " << nThreads << endl ;
  f << " " << endl ;


  f << "#include <assert.h>" << endl ;
  f << "#include \"bioGaussHermite.h\"" << endl ;
  f << "#include \"bioPythonWrapper.h\"" << endl ;
  f << "#include \"bioSample.h\"" << endl ;
  f << "#include \"patFileNames.h\"" << endl ;
  f << "#include \"patMath.h\"" << endl ;
  f << "#include \"bioVersion.h\"" << endl ;
  f << "#include \"bioIterates.h\"" << endl ;
  f << "#include \"bioIteratorInfoRepository.h\"" << endl ;
  f << "#include \"bioLiteralRepository.h\"" << endl ;
  f << "#include \"bio__constraints.h\"" << endl ;
  f << "#include \"bioGslParams.h\"" << endl ;

  f << " // Data structure for the threads" << endl ;
  f << " typedef struct{" << endl ;
  f << "   trVector* __beta ;" << endl ;
  f << "   trVector* grad;" << endl ;
  f << "   trHessian* hessian ;" << endl ;
  f << "   trHessian* bhhh ;" << endl ;
  f << "   patBoolean* success ;" << endl ;
  f << "   patError* err ;" << endl ;
  f << "   bioSample* sample ;" << endl ;
  f << "   patReal result ;" << endl ;
  f << "   bioIteratorSpan subsample ; " << endl ;
  f << " } inputArg ;" << endl ;
  f << " " << endl ;
  f << "   #include \"bioMetaIterator.h\"" << endl ;
  f << "   #include \"bioDrawIterator.h\"" << endl ;
  f << "   #include \"bioRandomDraws.h\"" << endl ;
  f << "   #include \"bioRowIterator.h\"" << endl ;
  f << "   #include \"bioMain.h\"" << endl ;
  f << "   #include \"bioGslBiogeme.h\"" << endl ;
  f << "   #include \"patMath.h\"" << endl ;
  f << "   #include \"patNormalCdf.h\"" << endl ;
  f << " " << endl ;
  //  f << " #define DEBUG_MESSAGE(message)" << endl ;
  //f << " #define DEBUG_MESSAGE(message) std::cout << message << endl ;" << endl ;

}


