//-*-c++-*------------------------------------------------------------
//
// File name : bioGenerateSimulation.h
// Author :    Michel Bierlaire
// Date :      Thu May 13 09:39:03 2010
//
//--------------------------------------------------------------------

#include "patLargeNumbers.h"
#include "patMath.h"
#include "patErrMiscError.h"
#include "patDisplay.h"
#include "bioGenerateSimulation.h"
#include "bioLiteralRepository.h"
#include "bioExpressionRepository.h"
#include "bioVersion.h"
#include "bioExpression.h"
#include "bioIteratorInfoRepository.h"
#include "bioArithMergedSum.h"
#include "bioCountOperations.h"
#include "bioParameters.h"

bioGenerateSimulation::bioGenerateSimulation(map<patString,patULong>* lit, 
							     bioExpression* anExpression,
							     patString m, patString f, patString g):
  theExpression(anExpression), 
  literals(lit),
  mainFileName(m),
  fctFileName(f), 
  gradFileName(g) {
 DEBUG_MESSAGE("Generate expression for " << literals->size() << " parameters"<< " to files " << f << " and " << g) ;
}

void bioGenerateSimulation::generate(patError*& err) {
  

  vector<patULong> theIds = 
    bioLiteralRepository::the()->getBetaIds(patFALSE,err) ;
  if (err != NULL) {
    WARNING(err->describe()) ;
    return ;
  }

  // Compute gradient


  // Simplify the expression

  bioExpression* theOriginalExpression = theExpression->getDeepCopy(err) ;
  if (err != NULL) {
    WARNING(err->describe()) ;
    return ;
  }

  bioCountOperations countingOperations ;
  countingOperations.fctBeforeSimplification = theExpression->getNumberOfOperations() ;
  //  DEBUG_MESSAGE("BEFORE ["<<theExpression->getNumberOfOperations()<<"]: " << *theExpression) ;
  if (!theExpression->canBeSimplified()) {
    vector<bioExpression*> expressionsToSimplify ;
    theExpression->getExpressionsToSimplify(&expressionsToSimplify,err) ;
    if (err != NULL) {
      WARNING(err->describe()) ;
      return ;
    }
    DEBUG_MESSAGE("Number of expressions found: " << expressionsToSimplify.size()) ;
    for (vector<bioExpression*>::iterator i = expressionsToSimplify.begin() ;
	 i != expressionsToSimplify.end() ;
	 ++i) {
      (*i)->simplifyLowerIteratorOrSequence(err) ;
      if (err != NULL) {
	WARNING(err->describe()) ;
return ;
      }
    }
  }
  else {
    // Simplify expression
    bioExpression* newExpression = simplify(theExpression,err) ;
    if (err != NULL) {
      WARNING(err->describe()) ;
      return ;
    }
    theExpression = newExpression ;
  }
  //  DEBUG_MESSAGE("AFTER ["<<theExpression->getNumberOfOperations()<<"]: " << *theExpression) ;
  countingOperations.fctAfterSimplification = theExpression->getNumberOfOperations() ;


  
  stringstream str ;
  ofstream f(mainFileName.c_str(),ios::out|ios::app) ;
  writeHeaders(f,err) ;
  if (err != NULL) {
    WARNING(err->describe()) ;
    return ;
  }
  f << " #include <pthread.h> " << endl ; //POSIX Threads library
  f << " void *computeFunction( void *ptr );" << endl ;
  f << " void *computeFunctionAndGradient( void *ptr );" << endl ;
  f << "" << endl ;
  f << "// Definition of the class" << endl ;
  f << "#include \"bioMyObjects.h\"" << endl ;
  f << "" << endl ;
  f << "" << endl ;
  f << "// Implementation of the class" << endl ;
  f << "" << endl ;
  patULong n = bioLiteralRepository::the()->getNumberOfEstimatedParameters()  ;
  DEBUG_MESSAGE("Number of estimated parameters: " << n) ;
  f << "  myBiogeme::myBiogeme() : bhhh("<< n << ")" << endl ;
  f << ", hessian("<< bioLiteralRepository::the()->getNumberOfEstimatedParameters() << ")" << endl ;
  f << " , threadGrad(NUMTHREADS,trVector("<< bioLiteralRepository::the()->getNumberOfEstimatedParameters() << "))" << endl ;
  f << " , threadBhhh(NUMTHREADS,trHessian(" << bioLiteralRepository::the()->getNumberOfEstimatedParameters() << " ))" << endl ;
  f << " , threadHessian(NUMTHREADS,trHessian(" << bioLiteralRepository::the()->getNumberOfEstimatedParameters() << " ))" << endl ;
  f << "    { " << endl ;
  f << "  bhhh.setToZero() ;" << endl ;
  f << "  hessian.setToZero() ;" << endl ;
  f << "}" << endl ;
  f << "" << endl ;
  f << "" << endl ;
  
  f << "patReal myBiogeme::getFunction(trVector* x," << endl ;
  f << "			       patBoolean* success," << endl ;
  f << "			       patError*& err) {" << endl ;
  f << "" << endl ;
  f << "*success = patTRUE ;" << endl ;
  f << "  bioLiteralRepository::the()->setBetaValues(*x,err) ;" << endl ;
  f << "    if (err != NULL) {" << endl ;
  f << "      WARNING(err->describe()) ;" << endl ;
  f << "      return patReal();" << endl ;
  f << "    }" << endl ;
  f << "  trVector theBetas = bioLiteralRepository::the()->getBetaValues() ; " << endl ;
  
  f << "  patReal logLike = 0.0 ;" << endl ;
  f << "" << endl ;
  f << "  pthread_t *threads;" << endl ;
  f << "  threads = new pthread_t[NUMTHREADS];" << endl ;
  f << "  " << endl ;
  f << "  inputArg *input;" << endl ;
  f << "  input = new inputArg[NUMTHREADS];" << endl ;
  f << "  " << endl ;
  f << "  for(int i=0; i<NUMTHREADS; i++) {" << endl ;
  f << "  input[i].subsample = bioIteratorInfoRepository::the()->getSublist(i,err) ; " << endl ;  
  f << "    if (err != NULL) {" << endl ;
  f << "      WARNING(err->describe()) ;" << endl ;
  f << "      return patReal();" << endl ;
  f << "    }" << endl ;
  f << "  input[i].__beta = &theBetas ;" << endl ;
  f << "  input[i].grad = NULL ;" << endl ;
  f << "  input[i].bhhh = NULL ;" << endl ;
  f << "  input[i].success = success ;" << endl ;
  f << "  input[i].err = err ;" << endl ;
  f << "  input[i].sample = sample ;" << endl ;
  f << "  input[i].result = 0.0 ;" << endl ;
  f << "    if(pthread_create( &threads[i], NULL, computeFunction, (void*) &input[i])) {" << endl ;
  f << "      err = new patErrMiscError(\"Error creating Threads! Exit!\") ;" << endl ;
  f << "      WARNING(err->describe()) ;" << endl ;
  f << "      return 0.0;" << endl ;
  f << "    }" << endl ;
  f << "  }" << endl ;
  f << "  void *tmp;" << endl ;
  f << "  " << endl ;
  f << " logLike = 0.0 ;" << endl ;
  f << "  patBoolean maxreal(patFALSE) ;" << endl ;
  f << "  for(int i=0; i<NUMTHREADS; ++i) {" << endl ;
  f << "    pthread_join( threads[i], &tmp);" << endl ;
  f << " // Generated at " << __FILE__ << " " << __LINE__  << endl ;
  f << "      if (!maxreal) {" << endl ;
  f << "	if (input[i].result == -patMaxReal) {" << endl ;
  f << "	  logLike = patMaxReal ;" << endl ;
  f << "	  maxreal = patTRUE ;" << endl ;
  f << "	}" << endl ;
  f << "	else {" << endl ;
  f << "    logLike -= input[i].result ;" << endl ;
  f << "	}" << endl ;
  f << "      }" << endl ;
  f << "  }" << endl ;
  f << "	" << endl ;
  f << "  return logLike ;" << endl ;
  f << "}" << endl ;
  f << "  " << endl ;
  f << "" << endl ;
  f << "" << endl ;
  f << "trVector* myBiogeme::getGradient(trVector* x," << endl ;
  f << "				 trVector* grad," << endl ;
  f << "				 patBoolean* success," << endl ;
  f << "				 patError*& err) {" << endl ;
  f << "*success = patTRUE ;" << endl ;
  f << "  patReal result = getFunctionAndGradient(x,grad,success,err) ;" << endl ;
  f << "  if (err != NULL) {" << endl ;
  f << "    WARNING(err->describe()) ;" << endl ;
  f << "    return NULL  ;" << endl ;
  f << "  }" << endl ;
  f << "  return grad ;" << endl ;
  f << "}" << endl ;
  f << "  " << endl ;
  
  f << "patReal myBiogeme::getFunctionAndGradient(trVector* x," << endl ;
  f << "					  trVector* grad," << endl ;
  f << "					  patBoolean* success," << endl ;
  f << "					  patError*& err) {" << endl ;
  f << "*success = patTRUE ;" << endl ;


  f << "bhhh.setToZero() ;" << endl ;
  f << "  bioLiteralRepository::the()->setBetaValues(*x,err) ;" << endl ;
  f << "    if (err != NULL) {" << endl ;
  f << "      WARNING(err->describe()) ;" << endl ;
  f << "      return patReal();" << endl ;
  f << "    }" << endl ;
  
  f << "  patReal result = getFunction(x,success,err) ;" << endl ;
  f << "  if (err != NULL) {" << endl ;
  f << "    WARNING(err->describe()) ;" << endl ;
  f << "    return NULL  ;" << endl ;
  f << "  }" << endl ;
  f << " getFinDiffGradient(x,grad,success,err) ;" << endl ;
  f << "  if (err != NULL) {" << endl ;
  f << "    WARNING(err->describe()) ;" << endl ;
  f << "    return NULL  ;" << endl ;
  f << "  }" << endl ;
  f << " return result ;" << endl ;
  f << "}" << endl ;
  f << "" << endl ;
  f << "" << endl ;
  f << "trHessian* myBiogeme::computeHessian(patVariables* x," << endl ;
  f << "				     trHessian& hessian," << endl ;
  f << "				     patBoolean* success," << endl ;
  f << "				     patError*& err) {" << endl ;
  f << "    return computeFinDiffHessian(x,hessian,success,err) ;" << endl ;
  f << "}" << endl ;
  f << "" << endl ;
  f << "trHessian* myBiogeme::getCheapHessian(trVector* x," << endl ;
  f << "			     trHessian& hessian," << endl ;
  f << "			     patBoolean* success," << endl ;
  f << "				      patError*& err) {" << endl ;
  f << "    hessian = bhhh ;" << endl ;
  f << "    *success = patTRUE ;" << endl ;
  f << "    return &bhhh ;" << endl ;
  f << "}" << endl ;
  f << "" << endl ;
  f << "patBoolean myBiogeme::isCheapHessianAvailable() {" << endl ;
  f << "  return patTRUE ;" << endl ;
  f << "}" << endl ;
  f << "" << endl ;
  f << "trVector myBiogeme::getHessianTimesVector(trVector* x," << endl ;
  f << "					  const trVector* v," << endl ;
  f << "					  patBoolean* success," << endl ;
  f << "					  patError*& err) {" << endl ;
  f << "  return trVector() ;" << endl ;
  f << "}" << endl ;
  f << "" << endl ;
  f << "patBoolean myBiogeme::isGradientAvailable(patError*& err) const {" << endl ;
  f << "  return patFALSE ;" << endl ;
  f << "}" << endl ;
  f << "" << endl ;
  f << "patBoolean myBiogeme::isHessianAvailable(patError*& err)  const {" << endl ;
  f << "  return patFALSE;" << endl ;
  f << "}" << endl ;
  f << "" << endl ;
  f << "patBoolean myBiogeme::isHessianTimesVectorAvailable(patError*& err) const {" << endl ;
  f << "  return patFALSE ;" << endl ;
  f << "}" << endl ;
  f << "" << endl ;
  f << "unsigned long myBiogeme::getDimension(patError*& err) const {" << endl ;
  f << "  return " << bioLiteralRepository::the()->getNumberOfEstimatedParameters() << " ;" << endl ;
  f << "}" << endl ;
  f << "" << endl ;
  f << "patBoolean myBiogeme::isUserBased() const {" << endl ;
  f << "  return patTRUE ;" << endl ;
  f << "}" << endl ;
  f << "" << endl ;
  f << "void myBiogeme::generateCppCode(patError*& err) {" << endl ;
  f << "  err = new patErrMiscError(\"This function should never be called\") ;" << endl ;
  f << "  WARNING(err->describe()) ;" << endl ;
  f << "  " << endl ;
  f << "}" << endl ;
  f << "" << endl ;
  f << "// Implementation of the default display  on a terminal" << endl ;
  f << "" << endl ;
  f << "patDisplay::patDisplay() : screenImportance(patImportance::patDEBUG), " << endl ;
  f << "			   logImportance(patImportance::patDEBUG)," << endl ;
  f << "			   logFile(patFileNames::the()->getLogFile().c_str()) {" << endl ;
  f << "" << endl ;
  f << "  patAbsTime now ;" << endl ;
  f << "  now.setTimeOfDay() ;" << endl ;
  f << "  logFile << \"This file has automatically been generated.\" << endl ;" << endl ;
  f << "  logFile << now.getTimeString(patTsfFULL) << endl ;" << endl ;
  f << "  logFile << bioVersion::the()->getCopyright() << endl ;" << endl ;
  f << "  logFile << endl ;" << endl ;
  f << "  logFile << bioVersion::the()->getVersionInfoDate() << endl ;" << endl ;
  f << "  logFile << bioVersion::the()->getVersionInfoAuthor() << endl ;" << endl ;
  f << "  logFile << endl ;" << endl ;
  f << "  logFile << flush ;" << endl ;
  f << "" << endl ;
  f << "}" << endl ;
  f << "" << endl ;
  f << "patDisplay::~patDisplay() {" << endl ;
  f << "  messages.erase(messages.begin(),messages.end()) ;" << endl ;
  f << "   logFile.close() ;" << endl ;
  f << "}" << endl ;
  f << "" << endl ;
  f << "patDisplay* patDisplay::the() {" << endl ;
  f << "  static patDisplay* singleInstance = NULL ;" << endl ;
  f << "  if (singleInstance == NULL) {" << endl ;
  f << "    singleInstance = new patDisplay ;" << endl ;
  f << "    assert(singleInstance != NULL) ;" << endl ;
  f << "  }" << endl ;
  f << "  return singleInstance ;" << endl ;
  f << "}" << endl ;
  f << "" << endl ;
  f << "void patDisplay::addMessage(const patImportance& aType," << endl ;
  f << "			    const patString& text," << endl ;
  f << "			    const patString& fileName," << endl ;
  f << "			    const patString& lineNumber) {" << endl ;
  f << "" << endl ;
  f << "  patMessage theMessage ;" << endl ;
  f << "" << endl ;
  f << "  theMessage.theImportance = aType ;" << endl ;
  f << "  theMessage.text = text ;" << endl ;
  f << "  theMessage.fileName = fileName ;" << endl ;
  f << "  theMessage.lineNumber = lineNumber ;" << endl ;
  f << "" << endl ;
  f << "  patAbsTime now ;" << endl ;
  f << "  now.setTimeOfDay() ;" << endl ;
  f << "  theMessage.theTime = now ;" << endl ;
  f << "" << endl ;
  f << "  if (aType <= screenImportance) {" << endl ;
  f << "    if (screenImportance < patImportance::patDEBUG) {" << endl ;
  f << "      cout << theMessage.shortVersion() << endl << flush  ;" << endl ;
  f << "    }" << endl ;
  f << "    else {" << endl ;
  f << "      cout << theMessage.fullVersion() << endl << flush  ;" << endl ;
  f << "    }" << endl ;
  f << "  }" << endl ;
  f << "  if (aType <= logImportance) {" << endl ;
  f << "    if (logImportance < patImportance::patDEBUG) {" << endl ;
  f << "      logFile << theMessage.shortVersion() << endl << flush  ;" << endl ;
  f << "    }" << endl ;
  f << "    else {" << endl ;
  f << "      logFile << theMessage.fullVersion() << endl << flush  ;" << endl ;
  f << "    }" << endl ;
  f << "  }" << endl ;
  f << "  messages.push_back(theMessage) ;" << endl ;
  f << "}" << endl ;
  f << "" << endl ;
  f << "void patDisplay::setScreenImportanceLevel(const patImportance& aType) {" << endl ;
  f << "  screenImportance = aType ;" << endl ;
  f << "}" << endl ;
  f << "  " << endl ;
  f << "void patDisplay::setLogImportanceLevel(const patImportance& aType) {" << endl ;
  f << "  logImportance = aType ;" << endl ;
  f << "}" << endl ;
  f << "" << endl ;
  f << "void patDisplay::initProgressReport(const patString message," << endl ;
  f << "			unsigned long upperBound) {" << endl ;
  f << "  " << endl ;
  f << "" << endl ;
  f << "}" << endl ;
  f << "" << endl ;
  f << "patBoolean patDisplay::updateProgressReport(unsigned long currentValue) {" << endl ;
  f << "  return patTRUE ;" << endl ;
  f << "}" << endl ;
  f << "" << endl ;
  f << "void patDisplay::terminateProgressReport() {" << endl ;
  f << "" << endl ;
  f << "}" << endl ;
  f << "" << endl ;
  f << "" << endl ;
  f << "" << endl ;
  f << "" << endl ;

  patString prefix("l") ;
  ofstream gslhfile("bioGslBiogeme.h",ios::app) ;

  theExpression->generateGslFunctions(gslhfile,prefix,err) ;
  if (err != NULL) {
    WARNING(err->describe()) ;
    return ;
  }
  gslhfile.close() ;

  
  f.close() ;
  ofstream mk("files.mk",ios::app) ;
  mk << fctFileName << " " ;
  mk.close() ;
  f.open(fctFileName.c_str()) ;
  writeHeaders(f,err) ;
  if (err != NULL) {
    WARNING(err->describe()) ;
    return ;
  }
  f << " void *computeFunction(void* fctPtr) {" << endl ;
  //  f << "   patVariables* __x ;" << endl ;
  f << "   vector< vector<patReal> >* __draw ;" << endl ;
  f << "   inputArg *input;" << endl ;
  f << "   input = (inputArg *) fctPtr;" << endl ;
  f << "   trVector* __beta =  input->__beta ;" << endl ;
  f << "   trVector* grad = input->grad ;" << endl ;
  f << "   patBoolean* success = input->success ;" << endl ;
  f << "   trHessian* theHessian = input->hessian ;" << endl ;
  f << "   trHessian* bhhh = input->bhhh ;" << endl ;
  f << "   patError*& err = input->err ;" << endl ;
  f << "   bioSample* sample = input->sample ;" << endl ;
  f << "  bioIteratorSpan __theThreadSpan = input->subsample ; " << endl ;
  


    
  if (theExpression->isSimple()){
    f << "   input->result = "<< theExpression->getSimpleCppCode(err) << " ;" << endl ;
    if (err != NULL) {
      WARNING(err->describe()) ;
    }
  }
  else {
    f << theExpression->getCppCode(prefix,err) << endl ;
    if (err != NULL) {
      WARNING(err->describe()) ;
    }
      f << "   input->result = "<<theExpression->getCppResult(prefix) <<" ;" << endl ;
    }

    f << " }" << endl ;

    f.close() ;

    // Create an empty file for the gadient
    f.open(gradFileName.c_str()) ;
    f.close() ;
    
    // Constraints
    
    f.open("bio__constraints.h") ;
    patAbsTime now ;
    now.setTimeOfDay() ;
    f << "// File: bio__constraints.h" << endl ;
    f << "// This file has automatically been generated." << endl ;
    f << "// " << now.getTimeString(patTsfFULL) << endl ;
    f << "// " << bioVersion::the()->getCopyright() << endl ;
    f << endl ;
    f << "// " << bioVersion::the()->getVersionInfoDate() << endl ;
    f << "// " << bioVersion::the()->getVersionInfoAuthor() << endl ;
    f << endl ;
    
    f << endl ;
    f << "// No constraint." << endl ;
    f.close() ;

    mk.open("files.mk",ios::app) ;
    mk << "bio__constraints.cc " ;
    mk.close() ;
    
    f.open("bio__constraints.cc") ;
    
    f << "#include \"bioMyObjects.h\"" << endl ;
    f << "void myBioConstraints::addConstraints() {" << endl ;
    
    f << "// Constraints" << endl ;
    f << "" << endl ;
    f << "    };" << endl ;

    f.close() ;
}




void bioGenerateSimulation::writeHeaders(ofstream& f, patError*& err) {
  
  patULong nThreads = bioParameters::the()->getValueInt("numberOfThreads",err) ;
  if (err != NULL) {
    WARNING(err->describe()) ;
    return ;
  }
  f << " #define NUMTHREADS " << nThreads << endl ;
  f << " " << endl ;


  f << "#include <assert.h>" << endl ;
  f << "#include \"bioGaussHermite.h\"" << endl ;
  f << "#include \"bioPythonWrapper.h\"" << endl ;
  f << "#include \"bioSample.h\"" << endl ;
  f << "#include \"patFileNames.h\"" << endl ;
  f << "#include \"patMath.h\"" << endl ;
  f << "#include \"bioVersion.h\"" << endl ;
  f << "#include \"bioIterates.h\"" << endl ;
  f << "#include \"bioIteratorInfoRepository.h\"" << endl ;
  f << "#include \"bioLiteralRepository.h\"" << endl ;
  f << "#include \"bio__constraints.h\"" << endl ;
  f << "#include \"bioGslParams.h\"" << endl ;

  f << " // Data structure for the threads" << endl ;
  f << " typedef struct{" << endl ;
  f << "   trVector* __beta ;" << endl ;
  f << "   trVector* grad;" << endl ;
  f << "   trHessian* hessian ;" << endl ;
  f << "   trHessian* bhhh ;" << endl ;
  f << "   patBoolean* success ;" << endl ;
  f << "   patError* err ;" << endl ;
  f << "   bioSample* sample ;" << endl ;
  f << "   patReal result ;" << endl ;
  f << "   bioIteratorSpan subsample ; " << endl ;
  f << " } inputArg ;" << endl ;
  f << " " << endl ;
  f << "   #include \"bioMetaIterator.h\"" << endl ;
  f << "   #include \"bioDrawIterator.h\"" << endl ;
  f << "   #include \"bioRandomDraws.h\"" << endl ;
  f << "   #include \"bioRowIterator.h\"" << endl ;
  f << "   #include \"bioMain.h\"" << endl ;
  f << "   #include \"bioGslBiogeme.h\"" << endl ;
  f << "   #include \"patMath.h\"" << endl ;
  f << " " << endl ;
  //  f << " #define DEBUG_MESSAGE(message)" << endl ;
  //f << " #define DEBUG_MESSAGE(message) std::cout << message << endl ;" << endl ;

}


