//-*-c++-*------------------------------------------------------------
//
// File name : bioLiteralRepository.cc
// Author :    Michel Bierlaire and Mamy Fetiarison
// Date :      Tue Apr 28 16:28:07 2009
//
//--------------------------------------------------------------------

#include <algorithm>
#include <sstream>
#include "patDisplay.h"
#include "patErrMiscError.h"
#include "patErrNullPointer.h"
#include "patErrOutOfRange.h"
#include  "bioLiteralRepository.h"

bioLiteralRepository* bioLiteralRepository::the() {
  static bioLiteralRepository* singleInstance = NULL ;

  if (singleInstance==NULL) {
    singleInstance = new bioLiteralRepository ;
  }
  return singleInstance ;
}

bioLiteralRepository::bioLiteralRepository() {
}

patULong bioLiteralRepository::addVariable(patString theName, 
					   patULong colId,
                                              patError*& err) {
  patULong id = listOfLiterals.size() ;

  map<patString,patULong>::iterator found = 
    listOrganizedByNames.find(theName) ;
  if (found != listOrganizedByNames.end()) {
    stringstream str ;
    str << "Literal " << theName << " already used with ID " << found->second ;
    err = new patErrMiscError(str.str()) ;
    WARNING(err->describe()) ;
    return patULong()  ;
  }

  bioVariable variable(theName,id, colId) ;
  listOfVariables.push_back(variable) ;
  listOfLiterals.push_back(theName) ;
  listOrganizedByNames[theName] = id ;
  return id ;
}

patULong bioLiteralRepository::addRandomVariable(patString theName, 
						 patError*& err) {
  patULong id = listOfLiterals.size() ;

  map<patString,patULong>::iterator found = 
    listOrganizedByNames.find(theName) ;
  if (found != listOrganizedByNames.end()) {
    stringstream str ;
    str << "Literal " << theName << " already used with ID " << found->second ;
    err = new patErrMiscError(str.str()) ;
    WARNING(err->describe()) ;
    return patULong()  ;
  }
  
  bioRandomVariable variable(theName,id) ;
  listOfRandomVariables.push_back(variable) ;
  listOfLiterals.push_back(theName) ;
  listOrganizedByNames[theName] = id ;
  return id ;
}

patULong bioLiteralRepository::addIntermediateValue(patError*& err) {
  patULong id = listOfLiterals.size() ;
  // Name created for debugging purposes only.
  stringstream str ;
  str << "_bioKeyword_" << id ;
  patString name = patString(str.str()) ;
  map<patString,patULong>::iterator found = 
    listOrganizedByNames.find(name) ;
  if (found != listOrganizedByNames.end()) {
    stringstream str ;
    str << "Literal " << name << " already used with ID " << found->second ;
    err = new patErrMiscError(str.str()) ;
    WARNING(err->describe()) ;
    return NULL ;
  }
  listOfIntermediateVariables.push_back(name) ;
  listOfLiterals.push_back(name) ;
  listOrganizedByNames[name] = id ;
  flags[id] = patFALSE ;
  return id ;

}

patULong bioLiteralRepository::addFixedParameter(patString theName, 
						 patReal val,
						 patReal lowerBound,
						 patReal upperBound, 
						 patBoolean fixed, 
						 patString description, 
						 patError*& err) {
  patULong id = listOfLiterals.size() ;

  map<patString,patULong>::iterator found = 
    listOrganizedByNames.find(theName) ;
  if (found != listOrganizedByNames.end()) {
    stringstream str ;
    str << "Literal " << theName << " already used with ID " << found->second ;
    err = new patErrMiscError(str.str()) ;
    WARNING(err->describe()) ;
    return NULL ;
  }

  bioFixedParameter parameter(theName,id, val,lowerBound, upperBound, fixed, description) ;
  listOfFixedParameters.push_back(parameter) ;
  listOfLiterals.push_back(theName) ;
  listOrganizedByNames[theName] = id ;
  flags[id] = patFALSE ;
  return id ;

}
patULong bioLiteralRepository::addFixedParameter(patString theName, patReal val, patError*& err) {
  patULong id = listOfLiterals.size() ;

  map<patString,patULong>::iterator found = 
    listOrganizedByNames.find(theName) ;
  if (found != listOrganizedByNames.end()) {
    stringstream str ;
    str << "Literal " << theName << " already used with ID " << found->second ;
    err = new patErrMiscError(str.str()) ;
    WARNING(err->describe()) ;
    return NULL ;
  }

  bioFixedParameter parameter(theName,id, val) ;
  listOfFixedParameters.push_back(parameter) ;
  listOfLiterals.push_back(theName) ;
  listOrganizedByNames[theName] = id ;
  flags[id] = patFALSE ;
  return id ;
}


patULong bioLiteralRepository::getLiteralId(patString theName,
					  patError*& err) const {
  
  map<patString,patULong>::const_iterator found = 
    listOrganizedByNames.find(theName) ;
  if (found != listOrganizedByNames.end()) {
    return found->second ;
  }
  else {
    return patBadId ;
  }
  
}

patULong bioLiteralRepository::getLiteralId(const bioLiteral* aLiteral,
					    patError*& err) const {
  if (aLiteral == NULL) {
    return patBadId ;
  }
  patULong result = getLiteralId(aLiteral->getName(),err) ;
  if (err != NULL) {
    WARNING(err->describe()) ;
    return patBadId ;
  }
  return result ;
}

void bioLiteralRepository::computeIds() {
  sort(listOfFixedParameters.begin(),listOfFixedParameters.end()) ;
  sort(listOfVariables.begin(),listOfVariables.end()) ;
  sort(listOfRandomVariables.begin(),listOfRandomVariables.end()) ;
  DETAILED_MESSAGE("Computed IDs of " << listOfFixedParameters.size() << "param, " << listOfVariables.size() << " variables and " << listOfRandomVariables.size() << " random variables") ;
  for (patULong i = 0 ; i < listOfVariables.size() ; ++i) {
    bioLiteral* ptr = &(listOfVariables[i]) ;  
    idsOfVariables[listOfVariables[i].getName()] = i ;
    listOfVariables[i].theVariableId = i ;
  }
  for (patULong i = 0 ; i < listOfRandomVariables.size() ; ++i) {
    bioLiteral* ptr = &(listOfRandomVariables[i]) ;  
    idsOfRandomVariables[listOfRandomVariables[i].getName()] = i ;
    listOfRandomVariables[i].theRandomVariableId = i ;
  }
  patULong estimatedIndex = 0 ;
  for (patULong i = 0 ; i < listOfFixedParameters.size() ; ++i) {
    bioLiteral* ptr = &(listOfFixedParameters[i]) ;  
    idsOfParameters[listOfFixedParameters[i].getName()] = i ;
    listOfFixedParameters[i].theParameterId = i ;
    if (!listOfFixedParameters[i].isFixed) {
      parametersToBeEstimated[i] = estimatedIndex ;
      ++estimatedIndex ;
    }
  }
}

patString bioLiteralRepository::getName(patULong theId, patError*& err) {
  if (theId >= listOfLiterals.size()) {
    err = new patErrOutOfRange<patULong>(theId,0,listOfLiterals.size()-1) ;
    WARNING(err->describe()) ;
    return patString() ;
  }
  return listOfLiterals[theId] ;
}

patString bioLiteralRepository::getCppCode(patULong theId, patError*& err) {
  if (theId >= listOfLiterals.size()) {
    err = new patErrOutOfRange<patULong>(theId,0,listOfLiterals.size()-1) ;
    WARNING(err->describe()) ;
    return patString() ;
  }

  patString name = listOfLiterals[theId] ;
  map<patString,patULong>::iterator found = idsOfVariables.find(name) ;
  if (found != idsOfVariables.end()) {
    stringstream str ;
    str << "(*__x)[" << listOfVariables[found->second].columnId << "]" ;
    return(patString(str.str())) ;
  }
  found = idsOfParameters.find(name) ;
  if (found != idsOfParameters.end()) {
    stringstream str ;
    str << "(*__beta)[" << found->second << "]" ;
    return(patString(str.str())) ;
  }
  return name ;

}

map<patString,patULong>* bioLiteralRepository::getIdsOfVariables()  {
  return &idsOfVariables ;
}
map<patString,patULong>* bioLiteralRepository::getIdsOfRandomVariables()  {
  return &idsOfRandomVariables ;
}

map<patString,patULong>* bioLiteralRepository::getIdsOfParameters()  {
  return &idsOfParameters ;
}

patULong bioLiteralRepository::getNumberOfVariables() const {
  return listOfVariables.size() ;
}
patULong bioLiteralRepository::getNumberOfRandomVariables() const {
  return listOfRandomVariables.size() ;
}
patULong bioLiteralRepository::getNumberOfParameters() const {
  return listOfFixedParameters.size() ;
}

patULong bioLiteralRepository::getNumberOfEstimatedParameters()  {
  return parametersToBeEstimated.size() ;
}

patVariables bioLiteralRepository::getBetaValues(patBoolean all) const {
  if (all) {
    patVariables result ;
    for (vector<bioFixedParameter>::const_iterator i = listOfFixedParameters.begin() ;
	 i != listOfFixedParameters.end() ;
	 ++i) {
      result.push_back(i->currentValue) ;
    }
    return result ;
    
  }
  else {
     patVariables result(parametersToBeEstimated.size()) ;
    for (map<patULong,patULong>::const_iterator i = parametersToBeEstimated.begin() ;
	 i != parametersToBeEstimated.end() ;
	 ++i) {
      result[i->second] = listOfFixedParameters[i->first].currentValue ;
    }
  return result ;
  }
}

patVariables bioLiteralRepository::getLowerBounds() const {
  patVariables result ;
  for (vector<bioFixedParameter>::const_iterator i = 
	 listOfFixedParameters.begin() ;
       i != listOfFixedParameters.end() ;
       ++i) {
    if (!i->isFixed) {
      result.push_back(i->lowerBound) ;
    }
  }
  return result ;
  
}

patVariables bioLiteralRepository::getUpperBounds() const {
  patVariables result ;
  for (vector<bioFixedParameter>::const_iterator i = listOfFixedParameters.begin() ;
       i != listOfFixedParameters.end() ;
       ++i) {
    if (!i->isFixed) {
      result.push_back(i->upperBound) ;
    }
  }
  return result ;
  
}


vector<patULong> bioLiteralRepository::getBetaIds(patBoolean all,
						  patError*& err) const {
  vector<patULong> result ;
  for (vector<bioFixedParameter>::const_iterator i = listOfFixedParameters.begin() ;
       i != listOfFixedParameters.end() ;
       ++i) {
    patULong id = getLiteralId(&(*i),err) ;
    if (err != NULL) {
      WARNING(err->describe()) ;
      return vector<patULong>() ;
    }
    if (all || !i->isFixed) {
      result.push_back(id) ;
    }
  }
  return result ;
}

void bioLiteralRepository::setBetaValues(patVariables b, patError*& err) {

  for (map<patULong,patULong>::iterator i = parametersToBeEstimated.begin() ;
       i != parametersToBeEstimated.end() ;
       ++i) {
    listOfFixedParameters[i->first].currentValue = b[i->second] ;
  }
  return ;
}

patString bioLiteralRepository::getBetaName(patULong betaId, patError*& err) {
  if (betaId >= listOfFixedParameters.size()) {
    err = new patErrOutOfRange<patULong>(betaId,0, listOfFixedParameters.size()-1) ;
    WARNING(err->describe()) ;
    return patString() ;
  }
  return listOfFixedParameters[betaId].name ;
}

bioFixedParameter* bioLiteralRepository::getBetaParameter(patULong id, patError*& err) {
  patString name = listOfLiterals[id] ;
  map<patString,patULong>::iterator found = idsOfParameters.find(name) ;
  if (found == idsOfParameters.end()) {
    stringstream str ;
    str << "Parameter " << name << " with ID " << id << " unknown" ;
    err = new patErrMiscError(str.str()) ;
    WARNING(err->describe()) ;
    return NULL ;
  }
  return &(listOfFixedParameters[found->second]) ;
}


void bioLiteralRepository::setFlag(patULong id) {
  flags[id] = patTRUE ;
}

void bioLiteralRepository::unsetFlag(patULong id) {
  flags[id] = patFALSE ;
}

patBoolean bioLiteralRepository::isFlagSet(patULong id) const {
  map<patULong,patBoolean>::const_iterator found = flags.find(id) ;
  if (found == flags.end()) {
    return patFALSE ;
  }
  return found->second ;
}

void bioLiteralRepository::resetAllFlags() {
  for (map<patULong,patBoolean>::iterator i = flags.begin() ;
       i != flags.end() ;
       ++i) {
    i->second = patFALSE ;
  }
}
ostream& operator<<(ostream &str, const bioLiteralRepository& x) {
  str << "Ids of variables" << endl ;
  str << "++++++++++++++++" << endl ;
  for (map<patString,patULong>::const_iterator i = x.idsOfVariables.begin() ;
       i != x.idsOfVariables.end() ;
       ++i) {
    str << i->second << ": " << i->first << endl ;
  }
  str << "Ids of parameters" << endl ;
  str << "+++++++++++++++++" << endl ;
  for (map<patString,patULong>::const_iterator i = x.idsOfParameters.begin() ;
       i != x.idsOfParameters.end() ;
       ++i) {
    str << i->second << ": " << i->first << endl ;
  }
  str << "Ids of estimated parameters" << endl ;
  str << "+++++++++++++++++++++++++++" << endl ;
  for (map<patULong,patULong>::const_iterator i = x.parametersToBeEstimated.begin() ;
       i != x.parametersToBeEstimated.end() ;
       ++i) {
    str << "x[" << i->second << "]:" << i->first << endl ;
  }
  str << "Ids of literals" << endl ;
  str << "+++++++++++++++" << endl ;
  for (unsigned long i = 0 ; i < x.listOfLiterals.size() ; ++i) {
    str << i << ": " << x.listOfLiterals[i] << endl ;
  }
  return str ;
}

patString bioLiteralRepository::printFixedParameters(patBoolean estimation) const {
  stringstream str ;
  for (vector<bioFixedParameter>::const_iterator i = listOfFixedParameters.begin() ;
       i != listOfFixedParameters.end() ;
       ++i) {
    str << i->printPythonCode(estimation) << endl ;
  }
  return patString(str.str()) ;
}
