//-*-c++-*------------------------------------------------------------
//
// File name : bioModelParser.cc
// Author :    Michel Bierlaire and Mamy Fetiarison
// Date :      Thu May  7 09:53:57 2009
//
//--------------------------------------------------------------------

#include <iostream>
#include <list>

#include "patString.h"
#include "patError.h"
#include "patDisplay.h"
#include "patErrMiscError.h"

#include "bioParameters.h"

#include "bioModelParser_i.h"
#include "bioModelParser.h"

#include "bioArithAbs.h"
#include "bioArithAnd.h"
#include "bioArithUnaryMinus.h"
#include "bioArithBinaryMinus.h"
#include "bioArithBinaryPlus.h"
#include "bioArithDivide.h"
#include "bioArithEqual.h"
#include "bioArithExp.h"
#include "bioArithGreater.h"
#include "bioArithGreaterEqual.h"
#include "bioArithLess.h"
#include "bioArithLessEqual.h"
#include "bioArithLiteral.h"
#include "bioArithLog.h"
#include "bioArithNormalCdf.h"
#include "bioArithMult.h"
#include "bioArithNotEqual.h"
#include "bioArithOr.h"
#include "bioArithPower.h"
#include "bioArithConstant.h"
#include "bioArithSum.h"
#include "bioArithProd.h"
#include "bioArithPrint.h"
#include "bioArithIntegral.h"
#include "bioArithDerivative.h"
#include "bioArithElem.h"
#include "bioArithRandom.h"
#include "bioArithMax.h"
#include "bioArithMin.h"

#include "bioRandomDraws.h"

#include "bioRowIterator.h"
#include "bioIteratorInfo.h"
#include "bioIteratorInfoRepository.h"

// #define DEBUG_MESSAGE(s) GENERAL_MESSAGE("[INFO] " << s)
// #define DEBUG_MESSAGE(s) 

/**
   Constructor
   \fname python module file (no .py extension) which contains the
   model specification
*/
bioModelParser::bioModelParser(const patString& fname) :  filename(fname) {
  // Filename must not contain any path
  if (filename.find('/')!=string::npos) {
    FATAL("Filename contains directory path!") ;
  }
  repository = bioLiteralRepository::the() ;
}


bioModel *bioModelParser::readModel(patError*& err) {
  PyObject* pModule, *pBioObject ;
  PyObject* pDraws ;
  PyObject* pOperator ;
  PyObject* pModel ;
  bioExpression* theFormula ;


  // Remove .py extension from the file name if present
  patString fn(filename) ;
  patString pythonFileExtension(".py") ;
  int extensionPos = fn.size()-pythonFileExtension.size() ;
  if (fn.find(pythonFileExtension, extensionPos)!=string::npos) {
    fn.erase(extensionPos) ;
  }

  // Init Python
  Py_Initialize() ;
  //printf( "Path = ",Py_GetPath()) ;

  // Load the module and check if it has loaded
  pModule = PyImport_ImportModule(const_cast<char *>(fn.c_str())) ;
  if (PY_FAIL(pModule)) {
    PY_ERR_PRINT() ;
    FATAL("Failed to load " << filename) ;
  }

  GET_PY_OBJ(pModule, OPERATOR_OBJECT, pOperator) ;
  setOperators(pOperator) ;

  buildIteratorInfo(pModule,err) ;
  if (err != NULL) {
    WARNING(err->describe()) ;
    return NULL ;
  }

  GET_PY_OBJ(pModule, BIOGEME_OBJECT, pBioObject) ;

  theModel = new bioModel() ;

  map<patString, patString>* parameters = getParameters(pBioObject,err) ;
  if (err != NULL) {
    WARNING(err->describe()) ;
    return NULL ;
  }
  
  bioParameters::the()->setParameters(parameters,err) ;
  if (err != NULL) {
    WARNING(err->describe()) ;
    return NULL ;
  }
  
  // Get expression for statistics
  map<patString, bioExpression*>* statistics = getStatistics(pBioObject,err) ;
  if (err != NULL) {
    WARNING(err->describe()) ;
    return NULL ;
  }
  theModel->setStatistics(statistics) ;

  // Get expression for exclusion
  bioExpression* excludeExpr = getExclude(pBioObject,err) ;
  if (err != NULL) {
    WARNING(err->describe()) ;
    return NULL ;
  }
  theModel->setExcludeExpr(excludeExpr) ;

  // The likelihood function
  theFormula = getFormula(pBioObject,err) ;
  if (err != NULL) {
    WARNING(err->describe()) ;
    return NULL ;
  }


  if (theFormula != NULL) {
    
    // A formula for the likelihood function is present. Estimation will be performed
    theModel->setFormula(theFormula) ;
    
    
    // Get expressions for CONSTRAINTS 
    map<patString, bioExpression*>* constraints = getConstraints(pBioObject,err) ;
    if (err != NULL) {
      WARNING(err->describe()) ;
      return NULL ;
    }
    theModel->setConstraints(constraints) ;
    
  }
  else {
    theFormula = getSimulation(pBioObject,err) ;
    if (err != NULL) {
      WARNING(err->describe()) ;
      return NULL ;
    }
    if (theFormula != NULL) {
      DEBUG_MESSAGE("PERFORM SIMULATION HERE") ;
    }
    theModel->setSimulation(theFormula) ;
  }



  Py_DECREF(pModule) ;

  // Stop Python

  PyErr_Clear(); 
  Py_Finalize() ;
  

  DEBUG_MESSAGE("Model specification read!") ;

  bioLiteralRepository::the()->computeIds() ;

  return theModel ;
}


bioExpression* bioModelParser::getFormula(PyObject* pBioObject,patError*& err) {
  bioExpression* theFormula ;

  PyObject* pFormula ;

  GET_PY_OBJ(pBioObject, ESTIMATE, pFormula) ;
  if (pFormula != NULL) {
    if (pFormula != Py_None) {
      theFormula = buildExpression(pFormula,err) ;
      if (err != NULL) {
	WARNING(err->describe()) ;
	return NULL ;
      }
      return theFormula ;
    }
    else {
      DEBUG_MESSAGE("No formula (None) for the loglikelihood has been provided") ;
      return NULL ;

    }
  }
  else {
    DEBUG_MESSAGE("No formula for the loglikelihood has been provided") ;
    return NULL ;
  }
}


bioExpression* bioModelParser::getExclude(PyObject* pBioObject,patError*& err) {
  bioExpression* theFormula ;

  PyObject* pFormula ;

  GET_PY_OBJ(pBioObject, EXCLUDE, pFormula) ;
  if (pFormula != NULL) {
    if (pFormula != Py_None) {
      theFormula = buildExpression(pFormula,err) ;
      if (err != NULL) {
	WARNING(err->describe()) ;
	return NULL ;
      }
      return theFormula ;
    }
    else {
      DEBUG_MESSAGE("No formula (None) for the exclude condition has been provided") ;
      return NULL ;

    }
  }
  else {
    DEBUG_MESSAGE("No formula for the exclude condition has been provided") ;
    return NULL ;
  }
}


map<patString, bioExpression*>* bioModelParser::getStatistics(PyObject* pBioObject, patError*& err ) {
  PyObject* pStatistics ;
  PyObject* pKeyList ;
  map<patString,bioExpression*>* statisticsDic = new map<patString,bioExpression*> ;
  
  GET_PY_OBJ(pBioObject, STATISTICS, pStatistics) ;
  pKeyList = PyDict_Keys(pStatistics) ;

  Py_ssize_t listSize = PyList_Size(pKeyList) ;
  for (Py_ssize_t pos=0 ; pos<listSize ; ++pos) {
    PyObject* pName = PyList_GetItem(pKeyList, pos) ;
    PyObject* pStat = PyDict_GetItem(pStatistics, pName) ;

    patString statName = patString(PyBytes_AsString(PyUnicode_AsASCIIString(pName))) ;
    bioExpression *expr = buildExpression(pStat,err) ;
    if (err != NULL) {
      WARNING(err->describe()) ;
      return NULL ;
    }
    statisticsDic->insert(pair<patString,bioExpression*>(statName, expr)) ;
  }
  return statisticsDic ;
}


bioExpression* bioModelParser::getSimulation(PyObject* pBioObject, patError*& err ) {
  bioExpression* theFormula ;

  PyObject* pFormula ;

  GET_PY_OBJ(pBioObject, SIMULATE, pFormula) ;
  //  DEBUG_MESSAGE("Python object collected") ;
  if (pFormula != NULL) {
    if (pFormula != Py_None) {
      //      DEBUG_MESSAGE("Build expresion") ;
      theFormula = buildExpression(pFormula,err) ;
      if (err != NULL) {
	WARNING(err->describe()) ;
	return NULL ;
      }
      //      DEBUG_MESSAGE("getSimulation : Expression built") ;
      
      return theFormula ;
    }
    else {
      //      DEBUG_MESSAGE("No formula (None) for the simulation has been provided") ;
      return NULL ;

    }
  }
  else {
    //    DEBUG_MESSAGE("No formula for the simulation has been provided") ;
    return NULL ;
  }

}


map<patString, bioExpression*>* bioModelParser::getConstraints(PyObject* pBioObject, patError*& err) {
  PyObject* pConstraints ;
  PyObject* pKeyList ;
  map<patString,bioExpression*>* constraintsDic = new map<patString,bioExpression*> ;
  
  GET_PY_OBJ(pBioObject, CONSTRAINTS, pConstraints) ;
  pKeyList = PyDict_Keys(pConstraints) ;

  Py_ssize_t listSize = PyList_Size(pKeyList) ;
  for (Py_ssize_t pos=0 ; pos<listSize ; ++pos) {
    PyObject* pName = PyList_GetItem(pKeyList, pos) ;
    PyObject* pConstraint = PyDict_GetItem(pConstraints, pName) ;

    patString constraintName = patString(PyBytes_AsString(PyUnicode_AsASCIIString(pName))) ;
    bioExpression *expr = buildExpression(pConstraint,err) ;
    if (err != NULL) {
      WARNING(err->describe()) ;
      return NULL ;
    }
    constraintsDic->insert(pair<patString,bioExpression*>(constraintName, expr)) ;
  }
  return constraintsDic ;
}


map<patString, patString>* bioModelParser::getParameters(PyObject* pBioObject, patError*& err) {
  PyObject* pParameters ;
  PyObject* pKeyList ;
  map<patString,patString>* parametersDic = new map<patString,patString> ;
  
  GET_PY_OBJ(pBioObject, PARAMETERS, pParameters) ;
  pKeyList = PyDict_Keys(pParameters) ;

  Py_ssize_t listSize = PyList_Size(pKeyList) ;
  for (Py_ssize_t pos=0 ; pos<listSize ; ++pos) {
    PyObject* pName = PyList_GetItem(pKeyList, pos) ;
    PyObject* pParam = PyDict_GetItem(pParameters, pName) ;

    patString paramName = patString(PyBytes_AsString(PyUnicode_AsASCIIString(pName))) ;

    if (!PyUnicode_Check(pParam)) {
      stringstream str ;
      str << "Parameters must be specified as strings. It is not the case for parameter " << paramName ;
      err = new patErrMiscError(str.str()) ;
      WARNING(err->describe()) ;
      return NULL;
    }

    patString paramValue = patString(PyBytes_AsString(PyUnicode_AsASCIIString(pParam))) ;
    parametersDic->insert(pair<patString,patString>(paramName, paramValue)) ;
  }
  return parametersDic ;
}


void bioModelParser::buildIteratorInfo(PyObject* pModule, patError*& err) {
  
  PyObject *pIteratorList ;
  GET_PY_OBJ(pModule, ITERATOR_LIST, pIteratorList) ;
  int li = PyList_Size(pIteratorList) ;

  DETAILED_MESSAGE("Get infos for " << li << " iterators") ;
  // Process each iterator in the list
  for (int pos=0 ; pos<li ; ++pos) {

    bioIteratorInfo *iterInfo ; 
    PyObject *pIteratorName, *pColName, *pTypeIter ;
    patString iteratorName, colName, typeIter ;
    PyObject* pIterator ;

    pIterator = PyList_GetItem(pIteratorList, pos) ;
    if (PY_FAIL(pIterator)) {
      PY_ERR_PRINT() ;
      FATAL ("pIterator is NULL") ;
    }

    //DEBUG_MESSAGE("Get Iterators " << ITERATOR_NAME) ;
    GET_PY_OBJ(pIterator, ITERATOR_NAME, pIteratorName ) ;

    iteratorName = patString(PyBytes_AsString(PyUnicode_AsASCIIString(pIteratorName))) ;

    GET_PY_OBJ(pIterator, INDEX_VARIABLE, pColName) ;
    colName = patString(PyBytes_AsString(PyUnicode_AsASCIIString(pColName))) ;

    GET_PY_OBJ(pIterator, ITERATOR_TYPE, pTypeIter) ;
    typeIter = patString(PyBytes_AsString(PyUnicode_AsASCIIString(pTypeIter))) ;

    //    DEBUG_MESSAGE("[" << pos << "] Iterator " << typeIter << "[" << colName << "]") ;

    if (typeIter == patString(DRAW_TYPE)) {
      iterInfo = new bioIteratorInfo(patString("")) ;
      //DEBUG_MESSAGE("*** Adding to the repository: " << *iterInfo) ;
      bioIteratorInfoRepository::the()->addIteratorInfo(iteratorName, *iterInfo,err) ;
      if (err != NULL) {
	WARNING(err->describe()) ;
	return  ;
      }
    }
    else if (typeIter == patString(META_TYPE)) {
      //      DEBUG_MESSAGE("[" << pos << "] Build MetaIterator ");
      PyObject* pChild ;
      PyObject* pName ;

      GET_PY_OBJ(pIterator, CHILD, pChild) ;
      patString childIteratorName(PyBytes_AsString(PyUnicode_AsASCIIString(pChild))) ;
      GET_PY_OBJ(pIterator, STRUCTURE_NAME, pName) ;  // pName may be a filename or an Iterator

      if (PyUnicode_Check(pName)) {  // Iterator
        patString iteratorParentName(PyBytes_AsString(PyUnicode_AsASCIIString(pName))) ;
	if (iteratorParentName == "__dataFile__") {
	  iterInfo = new bioIteratorInfo(theDataFile,childIteratorName, colName, META) ;
	  DEBUG_MESSAGE("iterInfo created: " << iterInfo->getInfo()) ;
	}
	else {
	  bioIteratorInfo* iterParent = 
	    bioIteratorInfoRepository::the()->getIteratorInfo(iteratorParentName) ;
	  if (iterParent == NULL) {
	    FATAL("Iterator " << iteratorParentName << " does not exist!") ;
	    return ;
	  }
	  else {
          //DEBUG_MESSAGE("Parent iterator " << iteratorParentName << " is " << *iterParent) ;
	  }
	  iterInfo = new bioIteratorInfo(iterParent, childIteratorName, colName, META) ;
	}
      } 
      else {  
	FATAL("Argument of metaIterator must be a string") ;
      }
      //      DEBUG_MESSAGE("*** Adding to the repository: " << *iterInfo) ;
      bioIteratorInfoRepository::the()->addIteratorInfo(iteratorName, *iterInfo,err) ;
      if (err != NULL) {
	WARNING(err->describe()) ;
	return ;
      }
    }
    else if (typeIter == patString(ROW_TYPE)) {
      //DEBUG_MESSAGE("[" << pos << "] Build RowIterator ");
      PyObject* pChild ;
      GET_PY_OBJ(pIterator, CHILD, pChild) ;
      patString childIteratorName(PyBytes_AsString(PyUnicode_AsASCIIString(pChild))) ;
      PyObject* pName ;
      GET_PY_OBJ(pIterator, STRUCTURE_NAME, pName) ;  // pName can be a filename or an Iterator object

      if (PyUnicode_Check(pName)) {   // Iterator
        patString iteratorParentName(PyBytes_AsString(PyUnicode_AsASCIIString(pName))) ;
	if (iteratorParentName == "__dataFile__") {
	  iterInfo = new bioIteratorInfo(theDataFile,childIteratorName, colName, ROW) ;
	  DEBUG_MESSAGE("iterInfo created: " << iterInfo->getInfo()) ;
	}
	else {
	  bioIteratorInfo* iterParent = bioIteratorInfoRepository::the()->getIteratorInfo(iteratorParentName) ;
	  if (iterParent == NULL) {
	    FATAL("Iterator " << iteratorParentName << " does not exist!") ;
	    return ;
	  }
	  else {
	    DEBUG_MESSAGE("Parent iterator " << iteratorParentName << " is " << *iterParent) ;
	  }
	  iterInfo = new bioIteratorInfo(iterParent,childIteratorName, colName, ROW) ;
	  DEBUG_MESSAGE("iterInfo created: " << iterInfo->getInfo()) ;
	} 
      }
      else {  // Iterator
	FATAL("Argument of rowIterator must be a string") ;
      }
      //      DEBUG_MESSAGE("*** Adding to the repository: " << *iterInfo) ;
      bioIteratorInfoRepository::the()->addIteratorInfo(iteratorName, *iterInfo, err) ;
      if (err != NULL) {
	WARNING(err->describe()) ;
	return ;
      }
    }
  }
}


// Build the corresponding bioExpression object from the object "Expression" (in the python module)
bioExpression* bioModelParser::buildExpression(PyObject *pExpression, patError*& err) {
  PyObject *pIndexType ;
  int indexType ;

  GET_PY_OBJ(pExpression, OPERATOR_INDEX, pIndexType) ;
  indexType = PyLong_AsLong(pIndexType) ;

  // Numeric: 0
  if (indexType==operators.indexByNames[OP_NUM]) {
    PyObject *pValue ;
    GET_PY_OBJ(pExpression, NUMERIC_VALUE, pValue) ;
    patReal val(PyFloat_AsDouble(pValue)) ;
    bioArithConstant* constantNode = new bioArithConstant(NULL,val) ;
    return constantNode ;
  }
  // Variable
  else if (indexType==operators.indexByNames[OP_VAR]) { 
    bioArithLiteral *variable ;
    PyObject *pVariableName ;
    GET_PY_OBJ(pExpression, VARIABLE_NAME, pVariableName) ;
    patString name = patString(PyBytes_AsString(PyUnicode_AsASCIIString(pVariableName))) ;
    patULong aLiteralId = repository->getLiteralId(name, err) ;

    // The literal does not exist yet. It must be added in the literal repository.
    if (aLiteralId==patBadId) {
      stringstream str ;
      str << "Variable " << name << " undefined in the data file" ;
      err = new patErrMiscError(str.str()) ;
      WARNING(err->describe()) ;
      return NULL;
    }

    variable = new bioArithLiteral(NULL, aLiteralId) ;
    return variable ;
  }
  // Random variable Exactly the same as a regular variable, except
  // that it does not have to be declared by an expression ior in the
  // data file. It is used only for integration.
  else if (indexType==operators.indexByNames[OP_RV]) { 
    bioArithLiteral *variable ;
    PyObject *pVariableName ;
    GET_PY_OBJ(pExpression, VARIABLE_NAME, pVariableName) ;
    patString name = patString(PyBytes_AsString(PyUnicode_AsASCIIString(pVariableName))) ;
    patULong aLiteralId = repository->getLiteralId(name, err) ;
    if (err != NULL) {
      WARNING(err->describe()) ;
      return NULL ;
    }

    // The literal does not exist yet. It must be added in the literal repository.
    if (aLiteralId == patBadId) {
      aLiteralId = repository->addRandomVariable(name,err) ;
      if (err != NULL) {
	WARNING(err->describe()) ;
	return NULL ;
      }
    }
    variable = new bioArithLiteral(NULL, aLiteralId) ;
    return variable ;
  }
  // Normal Random number
  else if (indexType==operators.indexByNames[OP_NORMAL]) { 

    PyObject *pVariableName ;
    PyObject *pIndexName ;

    GET_PY_OBJ(pExpression, RAND_NORMAL_NAME, pVariableName) ;
    patString name = patString(PyBytes_AsString(PyUnicode_AsASCIIString(pVariableName))) ;
    GET_PY_OBJ(pExpression, RAND_NORMAL_INDEX, pIndexName) ;
    patString index = patString(PyBytes_AsString(PyUnicode_AsASCIIString(pIndexName))) ;

    patULong id = bioRandomDraws::the()->addRandomVariable(name,bioRandomDraws::NORMAL,index,err) ;
    if (err != NULL) {
      WARNING(err->describe()) ;
      return NULL ;
    }
    patULong indexId = repository->getLiteralId(index, err) ;
    bioArithRandom* theNormal = new bioArithRandom(NULL, id,patString("Normal"), indexId) ;
    return theNormal ;
  }
  // Uniform Random number
  else if (indexType==operators.indexByNames[OP_UNIFORM]) { 

    PyObject *pVariableName ;
    PyObject *pIndexName ;

    GET_PY_OBJ(pExpression, RAND_UNIFORM_NAME, pVariableName) ;
    patString name = patString(PyBytes_AsString(PyUnicode_AsASCIIString(pVariableName))) ;
    GET_PY_OBJ(pExpression, RAND_UNIFORM_INDEX, pIndexName) ;
    patString index = patString(PyBytes_AsString(PyUnicode_AsASCIIString(pIndexName))) ;

    patULong id = bioRandomDraws::the()->addRandomVariable(name,bioRandomDraws::UNIFORM,index,err) ;
    if (err != NULL) {
      WARNING(err->describe()) ;
      return NULL ;
    }
    patULong indexId = repository->getLiteralId(index, err) ;
    bioArithRandom* theUniform = new bioArithRandom(NULL, id,patString("Uniform"), indexId) ;
    return theUniform ;
  }
  // Uniform Random number
  else if (indexType==operators.indexByNames[OP_UNIFORMSYM]) { 

    PyObject *pVariableName ;
    PyObject *pIndexName ;

    GET_PY_OBJ(pExpression, RAND_UNIFORMSYM_NAME, pVariableName) ;
    patString name = patString(PyBytes_AsString(PyUnicode_AsASCIIString(pVariableName))) ;
    GET_PY_OBJ(pExpression, RAND_UNIFORMSYM_INDEX, pIndexName) ;
    patString index = patString(PyBytes_AsString(PyUnicode_AsASCIIString(pIndexName))) ;

    patULong id = bioRandomDraws::the()->addRandomVariable(name,bioRandomDraws::UNIFORMSYM,index,err) ;
    if (err != NULL) {
      WARNING(err->describe()) ;
      return NULL ;
    }
    patULong indexId = repository->getLiteralId(index, err) ;
    bioArithRandom* theUniformSym = new bioArithRandom(NULL, id,patString("UniformSym"), indexId) ;
    return theUniformSym ;
  }
  // Parameters (Beta)
  else if (indexType==operators.indexByNames[OP_PARAM]) { 
    bioArithLiteral *beta ;
    PyObject *pParamName ;
    PyObject *pLowerBound ;
    PyObject *pUpperBound ;
    PyObject *pFixed ;
    PyObject *pCurVal ;
    PyObject *pDesc ;

    GET_PY_OBJ(pExpression, "name", pParamName) ;
    patString name = patString(PyBytes_AsString(PyUnicode_AsASCIIString(pParamName))) ;

    GET_PY_OBJ(pExpression,"lb", pLowerBound) ;
    patReal lb = PyFloat_AsDouble(pLowerBound) ;
    
    GET_PY_OBJ(pExpression, "ub", pUpperBound) ;
    patReal ub = PyFloat_AsDouble(pUpperBound) ;
    
    GET_PY_OBJ(pExpression, "st", pFixed) ;
    patBoolean fixed = patBoolean(PyLong_AsLong(pFixed)==1) ;

    GET_PY_OBJ(pExpression, "val", pCurVal) ;
    patReal val = PyFloat_AsDouble(pCurVal) ;

    GET_PY_OBJ(pExpression, "desc", pDesc) ;
    patString desc = patString(PyBytes_AsString(PyUnicode_AsASCIIString(pDesc))) ;
    
    patULong aLiteralId  = repository->getLiteralId(name, err) ;

    // The literal does not exist yet. It must be added in the literal repository.
    if (aLiteralId == patBadId) {
      aLiteralId = repository->addFixedParameter(name, val, lb, ub, fixed, desc, err) ;
    }

    beta = new bioArithLiteral(NULL, aLiteralId) ;
    return beta ;
  }
  else if ((indexType>=operators.minBinopIndex) && (indexType<=operators.maxBinopIndex)) { // Bin Op

    bioExpression *binOpExpr ;
    PyObject *pLeft ;
    PyObject *pRight ;

    GET_PY_OBJ(pExpression, "left", pLeft) ;
    GET_PY_OBJ(pExpression, "right", pRight) ;

    bioExpression *left = buildExpression(pLeft,err) ;
    if (err != NULL) {
      WARNING(err->describe()) ;
      return NULL ;
    }
    bioExpression *right = buildExpression(pRight,err) ;
    if (err != NULL) {
      WARNING(err->describe()) ;
      return NULL ;
    }

    if (indexType==operators.indexByNames[OP_ADD]) { // add
      binOpExpr = new bioArithBinaryPlus(NULL, left, right) ; }
    else if (indexType==operators.indexByNames[OP_SUB]) { // sub
      binOpExpr = new bioArithBinaryMinus(NULL, left, right) ; }
    else if (indexType==operators.indexByNames[OP_MUL]) { // mul
      binOpExpr = new bioArithMult(NULL, left, right) ; } 
    else if (indexType==operators.indexByNames[OP_DIV]) { // div
      binOpExpr = new bioArithDivide(NULL, left, right) ; }
    else if (indexType==operators.indexByNames[OP_POW]) { // power
      binOpExpr = new bioArithPower(NULL, left, right) ; }
    else if (indexType==operators.indexByNames[OP_AND]) { // and
      binOpExpr = new bioArithAnd(NULL, left, right) ; }
    else if (indexType==operators.indexByNames[OP_OR]) { // or
      binOpExpr = new bioArithOr(NULL, left, right) ; } 
    else if (indexType==operators.indexByNames[OP_EQ]) { // ==
      binOpExpr = new bioArithEqual(NULL, left, right) ; } 
    else if (indexType==operators.indexByNames[OP_NEQ]) { // <>
      binOpExpr = new bioArithNotEqual(NULL, left, right) ; } 
    else if (indexType==operators.indexByNames[OP_GT]) { // >
      binOpExpr = new bioArithGreater(NULL, left, right) ; } 
    else if (indexType==operators.indexByNames[OP_GE]) { // >=
      binOpExpr = new bioArithGreaterEqual(NULL, left, right) ; } 
    else if (indexType==operators.indexByNames[OP_LT]) { // <
      binOpExpr = new bioArithLess(NULL, left, right) ; } 
    else if (indexType==operators.indexByNames[OP_LE]) { // <=
      binOpExpr = new bioArithLessEqual(NULL, left, right) ; } 
    else if (indexType==operators.indexByNames[OP_MAX]) {
      binOpExpr = new bioArithMax(NULL,left,right) ;
    }
    else if (indexType==operators.indexByNames[OP_MIN]) {
      binOpExpr = new bioArithMin(NULL,left,right) ;
    }
    else if (indexType==operators.indexByNames[OP_MOD]) {
      FATAL("mod operator not yet implemented") ;
      //      binOpExpr = new bioArithMod(NULL,left,right) ;
    }
    else {
      FATAL("Binary Operator not yet implemented. indexType [" << indexType << "]") ;
    }
    left->setParent(binOpExpr) ;
    right->setParent(binOpExpr) ;
    return binOpExpr ;
  }
  else if ((indexType>=operators.minUnopIndex)&&(indexType<=operators.maxUnopIndex)) { // UnOp

    bioExpression *unOpExpr ;
    PyObject *pExp ;
    GET_PY_OBJ(pExpression, "expression", pExp) ;
    bioExpression *exp = buildExpression(pExp,err) ;
    if (err != NULL) {
      WARNING(err->describe()) ;
      return NULL ;
    }

    if (indexType==operators.indexByNames[OP_ABS]) { // abs
      //      DEBUG_MESSAGE("BuildExpression : Abs") ;
      unOpExpr = new bioArithAbs(NULL, exp) ;
    }
    else if (indexType==operators.indexByNames[OP_NEG]) { // negOp (unaryMinus)
      //      DEBUG_MESSAGE("BuildExpression : UnaryMinus") ;
      unOpExpr = new bioArithUnaryMinus(NULL, exp) ;
    }
    else if (indexType==operators.indexByNames[OP_LOG]) { // Logarithm function
      //      DEBUG_MESSAGE("BuildExpression : Logarithm") ;
      unOpExpr = new bioArithLog(NULL, exp) ;
    }
    else if (indexType==operators.indexByNames[OP_NORMALCDF]) { // Normal CDF function
      //      DEBUG_MESSAGE("BuildExpression : Normal CDF") ;
      unOpExpr = new bioArithNormalCdf(NULL, exp) ;
    }
    else if (indexType==operators.indexByNames[OP_EXP]) { // Exponential function
      //      DEBUG_MESSAGE("BuildExpression : Exponential") ;
      unOpExpr = new bioArithExp(NULL, exp) ;
    }
    else {
      FATAL("Unary Operator not yet implemented. indexType [" << indexType << "]") ;
    }
    exp->setParent(unOpExpr) ;
    return unOpExpr ;
  }
  else if (indexType==operators.indexByNames[OP_INTEGRAL]) {
    //      DEBUG_MESSAGE("[bioModelParser::buildExpression] Create new bioArithIntegral") ;
    PyObject* pExpr, *pVariable ;
    GET_PY_OBJ(pExpression, "function", pExpr ) ;
    bioExpression* expr = buildExpression(pExpr,err) ;
    if (err != NULL) {
      WARNING(err->describe()) ;
      return NULL ;
    }
    GET_PY_OBJ(pExpression,"variable", pVariable) ;
    patString variableName = patString(PyBytes_AsString(PyUnicode_AsASCIIString(pVariable))) ;
    bioExpression* theExpression = new bioArithIntegral(NULL, expr, variableName,err) ;
    if (err != NULL) {
      WARNING(err->describe()) ;
      return NULL ;
    }
    expr->setParent(theExpression) ;
    return theExpression ;
  }
  else if (indexType==operators.indexByNames[OP_DERIVATIVE]) {
    //      DEBUG_MESSAGE("[bioModelParser::buildExpression] Create new bioArithDerivative") ;
    PyObject* pExpr, *pVariable ;
    GET_PY_OBJ(pExpression, "function", pExpr ) ;
    bioExpression* expr = buildExpression(pExpr,err) ;
    if (err != NULL) {
      WARNING(err->describe()) ;
      return NULL ;
    }
    GET_PY_OBJ(pExpression,"variable", pVariable) ;
    patString variableName = patString(PyBytes_AsString(PyUnicode_AsASCIIString(pVariable))) ;
    bioExpression* theExpression = new bioArithDerivative(NULL, expr, variableName,err) ;
    if (err != NULL) {
      WARNING(err->describe()) ;
      return NULL ;
    }
    expr->setParent(theExpression) ;
    return theExpression ;
  }
  else if ((indexType>=operators.minIteratoropIndex)&&
           (indexType<=operators.maxIteratoropIndex)) {
    bioExpression *theExpression ;
    PyObject *pExpr, *pVariable, *pIteratorInfo;
    bioExpression *expr ; 
    bioLiteral *aVariable ;
    bioIteratorInfo *iterInfo ;

    GET_PY_OBJ(pExpression, "function", pExpr ) ;
    expr = buildExpression(pExpr,err) ;
    if (err != NULL) {
      WARNING(err->describe()) ;
      return NULL ;
    }

    GET_PY_OBJ(pExpression, "iteratorName", pIteratorInfo) ;
    patString iteratorName = patString(PyBytes_AsString(PyUnicode_AsASCIIString(pIteratorInfo))) ;
    iterInfo = bioIteratorInfoRepository::the()->getIteratorInfo(iteratorName) ;
    
    if (iterInfo == NULL) {
      stringstream str ;
      str << "Undefined iterator: " << iteratorName ;
      err = new patErrMiscError(str.str()) ;
      WARNING(err->describe()) ;
      return NULL ;
    }
    if (indexType==operators.indexByNames[OP_SUM]) {
      //      DEBUG_MESSAGE("[bioModelParser::buildExpression] Create new bioArithSum") ;
      //      DEBUG_MESSAGE("Iter info: " << *iterInfo) 
      theExpression = new bioArithSum(NULL, expr, iterInfo) ;
    }
    else if (indexType==operators.indexByNames[OP_PROD]) {
      //      DEBUG_MESSAGE("[bioModelParser::buildExpression] Create new bioArithProd") ;
      theExpression = new bioArithProd(NULL, expr, iterInfo) ;
    }
    else {
      FATAL("Operator not yet implemented. indexType [" << indexType << "]") ;
    }
    expr->setParent(theExpression) ;
    return theExpression ;

  }
  else if (indexType == operators.indexByNames[OP_ENUM]) {
    bioExpression *theExpression ;
    PyObject* pTerm ;
    PyObject* pKeyList ;
    map<patString,bioExpression*> theDict ;
    Py_ssize_t length ;
    


    GET_PY_OBJ(pExpression, "theDict", pTerm) ;
    pKeyList = PyDict_Keys(pTerm) ;
    length = PyList_Size(pKeyList) ;
    for (Py_ssize_t pos=0 ; pos<length ; ++pos) {
      patString theKey ;
      PyObject* pKey = PyList_GetItem(pKeyList, pos) ;
      if (pKey == NULL) {
	FATAL("NUll pointer") ;
      }
      PyObject* pExpression = PyDict_GetItem(pTerm, pKey) ;
      if (pExpression == NULL) {
	FATAL("Should never happen !!!") ;
      }
      theKey = PyBytes_AsString(PyUnicode_AsASCIIString(pKey)) ;
      //      DEBUG_MESSAGE("key = " << theKey) ;
      theDict[theKey] = buildExpression(pExpression,err) ; // expr ;
      if (err != NULL) {
	WARNING(err->describe()) ;
	return NULL ;
      }
    }

    //    DEBUG_MESSAGE("ENUM: dictionary has been built") ;
    PyObject* pIteratorInfo ;
    GET_PY_OBJ(pExpression, "iteratorName", pIteratorInfo) ;
    bioIteratorInfo *iterInfo ;
    patString iteratorName = patString(PyBytes_AsString(PyUnicode_AsASCIIString(pIteratorInfo))) ;
    iterInfo = bioIteratorInfoRepository::the()->getIteratorInfo(iteratorName) ;
    
    if (iterInfo == NULL) {
      stringstream str ;
      str << "Undefined iterator: " << iteratorName ;
      err = new patErrMiscError(str.str()) ;
      WARNING(err->describe()) ;
      return NULL ;
    }
    theExpression = new bioArithPrint(NULL,theDict,iterInfo) ;
    return theExpression ;
  }
  else if (indexType==operators.indexByNames[OP_ELEM]) {
    PyObject* pProb ;
    PyObject* pChoice ;
    PyObject* pDefault ;
    
    bioExpression* choice ;
    bioExpression* defExpression ;
    PyObject* pKeyList ;
    map<patULong,bioExpression*> theDict ;
    Py_ssize_t length ;
    bioExpression* theExpression = NULL ;
    
    GET_PY_OBJ(pExpression, "prob", pProb) ;
    pKeyList = PyDict_Keys(pProb) ;
    GET_PY_OBJ(pExpression, "default", pDefault) ;
    defExpression = buildExpression(pDefault,err) ;
    if (err != NULL) {
      WARNING(err->describe()) ;
      return NULL ;
    }
    GET_PY_OBJ(pExpression, "choice", pChoice) ;
    choice = buildExpression(pChoice,err) ;
    if (err != NULL) {
      WARNING(err->describe()) ;
      return NULL ;
    }

    length = PyList_Size(pKeyList) ;
    //    DEBUG_MESSAGE("Length = " << length) ;
    for (Py_ssize_t pos=0 ; pos<length ; ++pos) {
      patULong theKey ;
      PyObject* pKey = PyList_GetItem(pKeyList, pos) ;
      if (pKey == NULL) {
	FATAL("NUll pointer") ;
      }
      // DEBUG_MESSAGE("Key=" << PyLong_AsLong(pKey)) ;
      //       DEBUG_MESSAGE("Key = " << PyBytes_AsString(PyUnicode_AsASCIIString(PyObject)_Str(pKey)) );
      //       DEBUG_MESSAGE("Type Key = " << PyBytes_AsString(PyUnicode_AsASCIIString(PyObject)_Str(PyObject_Type(pKey))));

      //      if (PyLong_Check(pKey)) {
      PyObject* pExpression = PyDict_GetItem(pProb, pKey) ;
      if (pExpression == NULL) {
	FATAL("Should never happen !!!") ;
      }
      // bioExpression* expr = buildExpression(pExpression) ;
      theKey = PyLong_AsLong(pKey) ;
      theDict[theKey] = buildExpression(pExpression,err) ; // expr ;
      if (err != NULL) {
	WARNING(err->describe()) ;
	return NULL ;
      }
      //       }
      //       else {
      //         stringstream str ;
      //         str << "Problem with key: " << PyBytes_AsString(PyUnicode_AsASCIIString(pKey)) ;
      //         err = new patErrMiscError(str.str()) ;
      //         WARNING(err->describe()) ;
      //         return NULL;
      //       }
    }

    theExpression = new bioArithElem(NULL,choice,theDict,defExpression) ;
    return theExpression ;
  }
  else {
    FATAL("Not yet implemented node indexType [" << indexType << "]") ;
    return NULL ;
  }
  return NULL ;
}


void bioModelParser::setSampleFile(patString f) {
  theDataFile = f ;
}
