//-*-c++-*------------------------------------------------------------
//
// File name : bioReporting.cc
// Author :    Michel Bierlaire and Mamy Fetiarison
// Date :      Mon Aug 17 18:56:15 2009
//
//--------------------------------------------------------------------

#include <set>
#include "patCompareCorrelation.h"
#include "patCorrelation.h"
#include "patErrNullPointer.h"
#include "patMath.h"
#include "bioReporting.h"
#include "bioRawResults.h"
#include "patDisplay.h"
#include "patSvd.h"
#include "patPValue.h"
#include "bioLiteralRepository.h"
#include "bioVersion.h"
#include "bioParameters.h"
#include "patFormatRealNumbers.h"
#include "bioSample.h"
#include "bioRandomDraws.h"

bioReporting::bioReporting() :
  theRawResults(NULL),
  nbrDraws(patBadId),
  nbrParameters(patBadId),
  sampleSize(patBadId),
  initLoglikelihood(0.0),
  diagnostic(""),
  iterations(patBadId),
  runTime(""),
  sampleFile("") {

}

void bioReporting::computeEstimationResults(bioRawResults* rr, patError*& err) {
  theRawResults = rr ;
  if (rr == NULL) {
    err = new patErrNullPointer("bioRawResults") ;
    WARNING(err->describe()) ;
    return ;
  }
  compute(err) ;
  if (err != NULL) {
    WARNING(err->describe()) ;
    return ;
  }
}

void bioReporting::compute(patError*& err) {

  // Set the beta values
  bioLiteralRepository::the()->setBetaValues(theRawResults->solution,err) ;
  if (err != NULL) {
    WARNING(err->describe()) ;
    return ;
  }
  
  // Standard errors

  if (theRawResults->varianceCovariance != NULL) {
    for (patULong i = 0 ; i < theRawResults->varianceCovariance->nRows() ; ++i) {
      if ((*theRawResults->varianceCovariance)[i][i] < 0) {
	stdErr.push_back(patMaxReal) ;
      }
      else{
	stdErr.push_back(sqrt((*theRawResults->varianceCovariance)[i][i])) ; 
      }
    }
  }

  
  if (theRawResults->robustVarianceCovariance != NULL) {
    for (patULong i = 0 ; i < theRawResults->robustVarianceCovariance->nRows() ; ++i) {
      if ((*theRawResults->robustVarianceCovariance)[i][i] < 0) {
	robustStdErr.push_back(patMaxReal) ;
      } 
      else {
	robustStdErr.push_back(sqrt((*theRawResults->robustVarianceCovariance)[i][i])) ; 
      }
    }
  }

  nbrDraws = bioRandomDraws::the()->nbrOfDraws() ;
  nbrParameters = bioLiteralRepository::the()->getNumberOfEstimatedParameters() ;
  
}

patBoolean bioReporting::isRaoCramerAvailable() const {
  if (theRawResults == NULL) {
    return patFALSE ;
  }
  return (theRawResults->varianceCovariance != NULL) ;
}

patBoolean bioReporting::isSandwichAvailable() const {
  if (theRawResults == NULL) {
    return patFALSE ;
  }
  return (theRawResults->robustVarianceCovariance != NULL) ;
}


void bioReporting::writeHtml(patString fileName, patError*& err) {
  patFormatRealNumbers theNumber ;
  DEBUG_MESSAGE("Write " << fileName) ;
  ofstream htmlFile(fileName.c_str()) ;
  patAbsTime now ;
  now.setTimeOfDay() ;

  htmlFile << "<!DOCTYPE HTML PUBLIC \"-//W3C//DTD HTML 4.01 Transitional//EN\">" << endl ;
    htmlFile << "" << endl ;
  htmlFile << "<html>" << endl ;
  htmlFile << "<head>" << endl ;
  htmlFile << "<script src=\"http://transp-or2.epfl.ch/biogeme/sorttable.js\"></script>" << endl ;
  htmlFile << "<meta http-equiv=\"content-type\" content=\"text/html;charset=ISO-8859-1\">" << endl ;
  htmlFile << "<title>"<< fileName <<" - Report from " << bioVersion::the()->getVersionInfoDate() << "</title>" << endl ;
  htmlFile << "<meta name=\"keywords\" content=\"biogeme, discrete choice, random utility\">" << endl ;
  htmlFile << "<meta name=\"description\" content=\"Report from " << bioVersion::the()->getVersionInfoDate() << "\">" << endl ;
  htmlFile << "<meta name=\"author\" content=\"Michel Bierlaire\">" << endl ;
  htmlFile << "<style type=text/css>" << endl ;
  htmlFile << "<!--table" << endl ;
  htmlFile << ".biostyle" << endl ;
  htmlFile << "	{font-size:10.0pt;" << endl ;
  htmlFile << "	font-weight:400;" << endl ;
  htmlFile << "	font-style:normal;" << endl ;
  htmlFile << "	font-family:Courier;}" << endl ;
  htmlFile << ".boundstyle" << endl ;
  htmlFile << "	{font-size:10.0pt;" << endl ;
  htmlFile << "	font-weight:400;" << endl ;
  htmlFile << "	font-style:normal;" << endl ;
  htmlFile << "	font-family:Courier;" << endl ;
  htmlFile << "        color:red}" << endl ;
  htmlFile << "-->" << endl ;
  htmlFile << "</style>" << endl ;
  htmlFile << "</head>" << endl ;
  htmlFile << "" << endl ;
  htmlFile << "<body bgcolor=\"#ffffff\">" << endl ;

  htmlFile << "<p>" << bioVersion::the()->getVersionInfoDate() << "</p>" << endl ;
  htmlFile << "<p>" << bioVersion::the()->getVersionInfoAuthor() << "</p>" << endl ;

  htmlFile << "<h1>Estimation report</h1>" << endl ;
  htmlFile << "<p>This file has automatically been generated.</p>" << endl ;
  htmlFile << "<p> " << now.getTimeString(patTsfFULL) << "</p>" << endl ;
  htmlFile << endl ;


  htmlFile << "<table border=\"0\">" << endl ;


  if (nbrDraws != patBadId && nbrDraws != 0) {
    htmlFile << "<tr class=biostyle><td align=right ><strong>Number of draws</strong>:</td> <td>"<< nbrDraws <<"</td></tr>" << endl ;
  }
  if (nbrParameters != patBadId) {
    htmlFile << "<tr class=biostyle><td align=right ><strong>Number of estimated parameters</strong>: </td> <td>"<< nbrParameters <<"</td></tr>" << endl ;
  }
  if (sampleSize != patBadId) {
    htmlFile << "<tr class=biostyle><td align=right ><strong>Sample size</strong>:</td> <td>"<< sampleSize <<"</td></tr>" << endl ;
  }
  if (initLoglikelihood != 0.0) {
    htmlFile << "<tr class=biostyle><td align=right><strong>Init log-likelihood</strong>:	</td> <td>"<< theNumber.formatStats(initLoglikelihood) 
	     <<"</td></tr>" << endl ;
  }
  if (theRawResults != NULL) {
    htmlFile << "<tr class=biostyle><td align=right><strong>Final log-likelihood</strong>:	</td> <td>"
	     << theNumber.formatStats(theRawResults->valueAtSolution) 
	     <<"</td></tr>" << endl ;
    htmlFile << "<tr class=biostyle><td align=right><strong>Likelihood ratio test for the init. model</strong>:	</td> <td>"
	     << theNumber.formatStats(-2.0 * (initLoglikelihood  - theRawResults->valueAtSolution))  
	     <<"</td></tr>" << endl ;
    htmlFile << "<tr class=biostyle><td align=right><strong>Rho for the init. model</strong>:	</td> <td>"
	     << theNumber.formatStats(1.0-(theRawResults->valueAtSolution/initLoglikelihood))
	     <<"</td></tr>" << endl ;
    htmlFile << "<tr class=biostyle><td align=right><strong>Rho bar for the init. model</strong>:	</td> <td>"
	     << theNumber.formatStats(1.0-((theRawResults->valueAtSolution-patReal(nbrParameters))/initLoglikelihood)) 
	     <<"</td></tr>" << endl ;
    
    
    htmlFile << "<tr class=biostyle><td align=right><strong>Final gradient norm</strong>:	</td> <td>"
	     << theNumber.format(patTRUE,
				 patFALSE,
				 3,
				 theRawResults->gradientNorm) 
	     <<"</td></tr>" << endl ;
    htmlFile << "<tr class=biostyle><td align=right><strong>Diagnostic</strong>:	</td> <td>"
	     << diagnostic 
	     <<"</td></tr>" << endl ;
  }
  if (iterations != patBadId) {
    htmlFile << "<tr class=biostyle><td align=right><strong>Iterations</strong>:	</td> <td>"
	     << iterations
	     <<"</td></tr>" << endl ;
  }
  if (runTime != "") {
    htmlFile << "<tr class=biostyle><td align=right><strong>Run time</strong>:	</td> <td>"
	     << runTime
	     <<"</td></tr>" << endl ;
    htmlFile << "</td></tr>" << endl ;
  }
  if (sampleFile != "") {
    htmlFile << "<tr class=biostyle><td align=right><strong>Sample file</strong>:	</td>" ;
    htmlFile << "<td>" << sampleFile  << "</td></tr>" ;
  }
  for (map<patString, patReal>::iterator s = statistics.begin() ;
       s != statistics.end() ;
       ++s) {
    htmlFile << "<tr class=biostyle><td align=right><strong><em>"<<s->first<<":</em></strong>	</td>" ;
  htmlFile << "<td><em>" << s->second  << "</em></td></tr>" ;

  }

  htmlFile << "</table>" << endl ;

  vector<patULong> betaIds = bioLiteralRepository::the()->getBetaIds(patFALSE,err) ;
  if (theRawResults != NULL) {
    htmlFile << "<h2>Estimated parameters</h2>" << endl ;
    htmlFile << "<p><font size='-1'>Click on the headers of the columns to sort the table  [<a href='http://www.kryogenix.org/code/browser/sorttable/' target='_blank'>Credits</a>]</font></p>" << endl ;
    htmlFile << "<table border=\"1\" class=\"sortable\">" << endl ;
    htmlFile << "<tr class=biostyle>" ;
    // Col 1
    htmlFile << "<th>Name </th>" ;
    // Col 2
    htmlFile << "<th>Value</th>" ;
    // Col 3
    htmlFile << "<th>Std err</th>" ;
    // Col 4
    htmlFile << "<th>t-test</th>" ;
    int printPValue = bioParameters::the()->getValueInt("printPValue",err) ;
    if (err != NULL) {
      WARNING(err->describe()) ;
      return ;
    }
    if (printPValue != 0) {
    // Col 5
      htmlFile << "<th>p-value</th>" ;
    }
    // Col 6
    htmlFile << "<th></th>" ;
    if (isSandwichAvailable()) {
    // Col 7
      htmlFile << "<th>Robust Std err</th>" ;
    // Col 8
      htmlFile << "<th>Robust t-test</th>" ;
      if (printPValue) {
    // Col 9
	htmlFile << "<th>p-value</th>" ;
      }
    // Col 10
      htmlFile << "<th></th>" ;
    }
    htmlFile << "</tr>" ;
    htmlFile << endl ;
    
    if (err != NULL) {
      WARNING(err->describe()) ;
      return ;
    }
    
    patReal tTestThreshold = bioParameters::the()->getValueReal("tTestThreshold",err) ;
    if (err != NULL) {
      WARNING(err->describe()) ;
      return ;
    }
    patString warningSign = bioParameters::the()->getValueString("warningSign",err) ; 
    if (err != NULL) {
      WARNING(err->describe()) ;
      return ;
    }
    

    for (patULong i = 0 ;
	 i < betaIds.size() ;
	 ++i) { // loop on beta  parameters
      
      bioFixedParameter* beta = bioLiteralRepository::the()->getBetaParameter(betaIds[i],err) ;
      if (err != NULL) {
	WARNING(err->describe()) ;
	return ;
      }
      if (beta == NULL) {
	err = new patErrNullPointer("bioFixedParameter") ;
	WARNING(err->describe()) ;
	return ;
      }
      
      if (patAbs(beta->getValue() - beta->getLowerBound()) <= patEPSILON ||
	  patAbs(beta->getValue() - beta->getUpperBound()) <= patEPSILON) {
	htmlFile << "<tr class=boundstyle>" << endl ;
      }
      else {
	htmlFile << "<tr class=biostyle>" << endl ;
      }
      // Col 1 
      htmlFile << "<td>" << beta->getName() << "</td>" ;
      // Col 2
      htmlFile << "<td>" << theNumber.formatParameters(beta->getValue()) << "</td>" ;
    
      if (isRaoCramerAvailable()) {
	patReal ttest = beta->getValue() / stdErr[i] ; 
      // Col 3
	htmlFile << "<td>" << theNumber.formatParameters(stdErr[i]) << "</td>";
      // Col 4
	htmlFile << "<td>" << theNumber.formatTTests(ttest) << "</td>" ;
	if (printPValue) {
	  patReal pvalue = patPValue(patAbs(ttest),err) ;
	  if (err != NULL) {
	    WARNING(err->describe()) ;
	    return ;
	  }
	  // Col 5
	  htmlFile << "<td>" << theNumber.formatTTests(pvalue) << "</td>" ;
	}
	// Col 6
	if (patAbs(ttest) < tTestThreshold || !finite(ttest)) {
	  htmlFile << "<td>" << warningSign << "</td>";
	}
	else {
	  htmlFile << "<td></td>" ;
	}
      }
      else {
	// Col 3
	htmlFile << "<td>var-covar unavailable</td>" ;
      }
      if (isSandwichAvailable()) {
	patReal rttest = beta->getValue()  / robustStdErr[i] ; 
	// Col 7
	htmlFile << "<td>" << theNumber.formatParameters(robustStdErr[i]) << "</td>" ;
	// Col 8
	htmlFile << "<td>" << theNumber.formatTTests(rttest) << "</td>" ; 
	if (printPValue) {
	  
	  patReal pvalue = patPValue(patAbs(rttest),err) ;
	  if (err != NULL) {
	    WARNING(err->describe()) ;
	    return ;
	  }
	// Col 9
	  htmlFile << "<td>" << theNumber.formatTTests(pvalue) << "</td>" ;
	}
	// Col 10
	if (patAbs(rttest) < tTestThreshold ||
	    !finite(rttest)) {
	  htmlFile << "<td>" << warningSign << "</td>";
	}
	else {
	  htmlFile << "<td></td>" ;
	}
      }
      else {
	// Col 7
	htmlFile << "<td></td>" ;
	// Col 8
	if (printPValue) {
	  htmlFile << "<td></td>" ;
	}
	// Col 9
	htmlFile << "<td></td>" ;
	// Col 10
	htmlFile << "<td></td>" ;
      }
      htmlFile << "</tr>" << endl ;
    }
    htmlFile << "</table>" << endl ;
    
    
    std::set<patCorrelation,patCompareCorrelation> correlation ;
    for (patULong i = 0 ;
	 i < betaIds.size() ;
	 ++i) { // loop on beta  parameters
      
      bioFixedParameter* betai = bioLiteralRepository::the()->getBetaParameter(betaIds[i],err) ;
      if (err != NULL) {
	WARNING(err->describe()) ;
	return ;
      }
      if (betai == NULL) {
	err = new patErrNullPointer("bioFixedParameter") ;
	WARNING(err->describe()) ;
	return ;
      }
      patReal varI(0) ;
      if (isRaoCramerAvailable()) {
	varI = (*theRawResults->varianceCovariance)[i][i] ;
      }
      patReal robustVarI(0) ;
      if (isSandwichAvailable()) {
	robustVarI = (*theRawResults->robustVarianceCovariance)[i][i] ;
      }    
      for (patULong j = i+1 ;
	   j < betaIds.size() ;
	   ++j) { // loop on beta  parameters
	
	bioFixedParameter* betaj = 
	  bioLiteralRepository::the()->getBetaParameter(betaIds[j],err) ;
	if (err != NULL) {
	  WARNING(err->describe()) ;
	  return ;
	}
	if (betaj == NULL) {
	  err = new patErrNullPointer("bioFixedParameter") ;
	  WARNING(err->describe()) ;
	  return ;
	}
	patCorrelation c ;
	c.firstCoef = betai->getName() ;
	c.secondCoef = betaj->getName() ;
	if (isRaoCramerAvailable()) {
	  patReal varJ = (*theRawResults->varianceCovariance)[j][j] ;
	  c.covariance = (*theRawResults->varianceCovariance)[i][j] ;
	  patReal tmp = varI * varJ ;
	  c.correlation = (tmp > 0) 
	    ? c.covariance / sqrt(tmp) 
	    : 0.0 ;
	  tmp = varI + varJ - 2.0 * c.covariance ;
	  c.ttest = (tmp > 0)
	    ? (betai->getValue() - betaj->getValue()) / sqrt(tmp) 
	    : 0.0  ;
	}
	if (isSandwichAvailable()) {
	  patReal robustVarJ = (*theRawResults->robustVarianceCovariance)[j][j] ;
	  c.robust_covariance = (*theRawResults->robustVarianceCovariance)[i][j] ;
	  patReal tmp = robustVarI * robustVarJ ;
	  c.robust_correlation = (tmp > 0) 
	    ? c.robust_covariance / sqrt(tmp) 
	    : 0.0 ;
	  tmp = robustVarI + robustVarJ - 2.0 * c.robust_covariance ;
	  c.robust_ttest = (tmp > 0)
	    ? (betai->getValue() - betaj->getValue()) / sqrt(tmp) 
	    : 0.0  ;
	}

	correlation.insert(c) ;
      }
    
    }
    htmlFile << "<h2>Correlation of coefficients</h2>" << endl ;
    htmlFile << "<p><font size='-1'>Click on the headers of the columns to sort the table [<a href='http://www.kryogenix.org/code/browser/sorttable/' target='_blank'>Credits</a>]</font></p>" << endl ;
    htmlFile << "<table border=\"1\" class=\"sortable\">" << endl ;
    htmlFile << "<tr class=biostyle>" ;
    // Col 1
    htmlFile << "<th>Coefficient1</th>" ;
    // Col 2
    htmlFile << "<th>Coefficient2</th>" ;
    // Col 3
    htmlFile << "<th>Covariance</th>" ;
    // Col 4
    htmlFile << "<th>Correlation</th>" ;
    // Col 5
    htmlFile << "<th>t-test</th>" ;
    // Col 6
    if (printPValue) {
      htmlFile << "<th>p-value</th>" ;
    }
    // Col 7
    htmlFile << "<th></th>" ;
    // Col 8
    htmlFile << "<th>Rob. cov.</th>" ;
    // Col 9
    htmlFile << "<th>Rob. corr.</th>" ;
    // Col 10
    htmlFile << "<th>Rob. t-test</th>" ;
    if (printPValue) {
      htmlFile << "<th>p-value</th>" ;
    }
    // Col 11
    htmlFile << "<th></th>";
    // Col 12
    htmlFile << "</tr>"  << endl ;
  
    for (std::set<patCorrelation,patCompareCorrelation>::iterator i = correlation.begin() ;
	 i != correlation.end() ;
	 ++i) {
      htmlFile << "<tr class=biostyle>" << endl ;
      // Col 1
      htmlFile << "<td>" << i->firstCoef << "</td>" ;
      // Col 2
      htmlFile << "<td>" << i->secondCoef << "</td>" ;
      // Col 3
      htmlFile << "<td>" << theNumber.formatParameters(i->covariance) << "</td>" ;
      // Col 4
      htmlFile << "<td>" << theNumber.formatParameters(i->correlation) << "</td>" ;
      // Col 5
      htmlFile << "<td>" << theNumber.formatTTests(i->ttest) << "</td>" ;
      // Col 6
      if (printPValue) {
	patReal pvalue = patPValue(patAbs(i->ttest),err) ;
	if (err != NULL) {
	  WARNING(err->describe()) ;
	  return ;
	}
	htmlFile << "<td>" << theNumber.formatTTests(pvalue) << "</td>" ;
      }
      // Col 7
      if (patAbs(i->ttest) < tTestThreshold) {
	htmlFile << "<td>" << warningSign << "</td>" ;
      }
      else {
	htmlFile << "<td></td>" ;
      }
      if (isSandwichAvailable()) {
	// Col 8
	htmlFile << "<td>" << theNumber.formatParameters(i->robust_covariance) << "</td>" ;
	// Col 9
	htmlFile << "<td>" << theNumber.formatParameters(i->robust_correlation) << "</td>" ;
	// Col 10
	htmlFile << "<td>" << theNumber.formatTTests(i->robust_ttest) << "</td>";
	// Col 11
	if (printPValue) {
	  
	  patReal pvalue = patPValue(patAbs(i->robust_ttest),err) ;
	  if (err != NULL) {
	    WARNING(err->describe()) ;
	    return ;
	  }
	  htmlFile << "<td>" << theNumber.formatTTests(pvalue) << "</td>"  ;
	}
	// Col 12
	if (patAbs(i->robust_ttest) < tTestThreshold) {
	  htmlFile << "<td>" << warningSign  << "</td>" ;
	}
	else {
	  htmlFile << "<td></td>" ;
	}
	  
      }
      htmlFile << "</tr>" ;
    }
    htmlFile << "</table>" << endl ;

    htmlFile << "<p>Smallest singular value: " << theRawResults->smallestSingularValue << "</p>" ;
    if (!theRawResults->eigenVectors.empty()) {
      htmlFile << "<h2>Unidentifiable model</h2>" << endl ;
      htmlFile << "<p>The log-likelihood is (almost) flat along the following combinations of parameters</p>" << endl ;
      htmlFile << "<table border=\"0\">" << endl ;
    
      patReal threshold = bioParameters::the()->getValueReal("singularValueThreshold",err) ;
      if (err != NULL) {
	WARNING(err->describe()) ;
	return ;
      }
      //    patIterator<patBetaLikeParameter>* theParamIter = createAllParametersIterator() ;
    
      for(map<patReal,patVariables>::iterator iter = 
	    theRawResults->eigenVectors.begin() ;
	  iter != theRawResults->eigenVectors.end() ;
	  ++iter) {
	htmlFile << "<tr class=biostyle><td>Sing. value</td><td>=</td><td>"
		 << iter->first 
		 << "</td></tr>" << endl ; 
      
	vector<patBoolean> printed(iter->second.size(),patFALSE) ;
	for (patULong j = 0 ;
	     j < betaIds.size() ;
	     ++j) { // loop on beta  parameters
	  bioFixedParameter* beta = 
	    bioLiteralRepository::the()->getBetaParameter(betaIds[j],err) ;
	  if (err != NULL) {
	    WARNING(err->describe()) ;
	    return ;
	  }
	  if (beta == NULL) {
	    err = new patErrNullPointer("bioFixedParameter") ;
	    WARNING(err->describe()) ;
	    return ;
	  }
	  if (patAbs(iter->second[j]) >= threshold) {
	    htmlFile << "<tr class=biostyle><td>" 
		     << iter->second[j] 
		     << "</td><td>*</td><td>" 
		     << beta->getName() << "</td></tr>" << endl ;
	    printed[j] = patTRUE ;
	  }
	}
	for (int j = 0 ; j < iter->second.size() ; ++j) {
	  if (patAbs(iter->second[j]) >= threshold && !printed[j]) {
	    htmlFile << "<tr class=biostyle><td>" 
		     << iter->second[j] 
		     << "</td><td>*</td><td>Param[" 
		     << j << "]</td></tr>" << endl ;
	  
	  }
	}
      }


      htmlFile << "</table>" << endl ;
    }
    
  }
  htmlFile << "</body>" << endl;
  htmlFile << "</html>" << endl;

  GENERAL_MESSAGE("File " << fileName << " has been generated") ;
  
}


void bioReporting::computeFromSample(bioSample* s, patError*& err) {

  if (s == NULL) {
    err = new patErrNullPointer("bioSample") ;
    WARNING(err->describe()) ;
    return ;
  }
  sampleFile = s->getFileName() ;
  sampleSize = s->size() ;

}

void bioReporting::setDiagnostic(patString diag) {
  diagnostic = diag ;
}

void bioReporting::setIterations(patULong i) {
  iterations = i ;
}

void bioReporting::setRunTime(patString t) {
  runTime = t ;
}

void bioReporting::setInitLL(patReal l) {
  initLoglikelihood = l ;
}

void bioReporting::addStatistic(patString name, patReal value) {
  statistics[name] = value ;
}
