
//-*-c++-*------------------------------------------------------------
//
// File name : bioSample.cc
// Author :    Michel Bierlaire and Mamy Fetiarison
// Date :      Fri Jun 19 10:35:31  2009
//
//--------------------------------------------------------------------


#include <set>
#include <iostream>
#include <fstream>
#include "bioSample.h"
#include "bioParameters.h"
#include "patDisplay.h"
#include "patErrMiscError.h"
#include "bioIteratorInfoRepository.h"
#include "bioMetaIterator.h"
#include "bioRowIterator.h"
#include "bioLiteralRepository.h"
#include "bioLiteralValues.h"
#include "bioExpression.h"



bioSample::bioSample(patString n, patError*& err) : fileName(n) {
  // Read the headers
  // Open the file
  patString line ;
  patString str ;
  ifstream in(fileName.c_str());
  if (!in || in.fail())
  {
    stringstream str ;
    str << "Cannot open the file " << fileName ;
    err = new patErrMiscError(str.str()) ;
    return ;
  }



  // Read headers
  int count = 0 ;
  getline(in, line) ;
  istringstream iss(line);
  while (iss >> str) {
    this->headers.push_back(str) ;
    indexFromHeader[str] = count ;
    bioLiteralRepository::the()->addVariable(str,count,err) ;
    if (err != NULL) {
      WARNING(err->describe()) ;
      return ;
    }
    ++count ;
  }

  patULong nbrOfColumns = count ;
  DETAILED_MESSAGE("There are " << nbrOfColumns << " columns in the file") ;
  
  patString rowIdHeader = 
    bioParameters::the()->getValueString(patString("HeaderOfRowId"),err) ;
  if (err != NULL) {
    WARNING(err->describe()) ;
    return ;
  }
  headers.push_back(rowIdHeader) ;
  indexFromHeader[rowIdHeader] = count ;
  bioLiteralRepository::the()->addVariable(rowIdHeader,count,err) ;
  ++count ;


  DETAILED_MESSAGE("Headers from the data file:") ;
  printHeaders() ;
  in.close() ;

  generatePythonHeaders(err) ;
   if (err != NULL) {
     WARNING(err->describe()) ;
     return ;
   }

}

bioSample::~bioSample() {
}

void bioSample::readFile(bioExpression* exclude, patError*& err) {
  ifstream in(fileName.c_str());
  if (!in || in.fail()) {
    stringstream str ;
    str << "Cannot open the file " << fileName ;
    err = new patErrMiscError(str.str()) ;
    return ;
  }
  // Skip headers
  patString line ;
  getline(in, line) ;

  GENERAL_MESSAGE("Read sample file: " << fileName) ;

  patReal rowNumber(0.0) ;
  bioLiteralValues::the()->eraseValues() ;
  vector<patReal> listOfV(headers.size()) ;
  // Read and store values
  while(in && !in.eof()) {
    in >> listOfV[0] ;
    bioLiteralValues::the()->setValue(headers[0],listOfV[0],err) ;
    if (err != NULL) {
      WARNING(err->describe()) ;
      return ;
    }
    if (!in.eof()) {
      for (unsigned long i = 1 ;  i < listOfV.size()-1 ; ++i) {
	if (!in.eof()) {
	  in >> listOfV[i] ;
	  bioLiteralValues::the()->setValue(headers[i],listOfV[i],err) ;
	  if (err != NULL) {
	    WARNING(err->describe()) ;
	    return ;
	  }
	}
      }

      patReal ex ;
      if (exclude != NULL) {
	ex = exclude->getValue(err) ;
	if (err != NULL) {
	  WARNING(err->describe()) ;
	  return ;
	}
      }
      else {
	ex = 0 ;
      }


      if (ex == 0) {
	listOfV[listOfV.size()-1] = rowNumber ;
	rowNumber += 1.0 ;
	samples.push_back(listOfV) ;
      }
    }
  }

  // Close file
  in.close() ;


}


int bioSample::size() {
  return samples.size() ;
}


vector<patString> bioSample::getHeaders() const {
  return headers ;
}


patVariables* bioSample::at(int index) {
  if (index >= samples.size()) {
    return NULL ;
  }
  return &(samples[index]) ;
}


patULong bioSample::getIndexFromHeader(patString h) {
  if (indexFromHeader.find(h) == indexFromHeader.end())
    return patBadId ;
  return indexFromHeader[h] ;
}


patString bioSample::getFileName() {
  return fileName ;
}


patBoolean bioSample::isEqual(bioSample s) {
  return s.getFileName() == this->fileName ;
}


void bioSample::printSample() {
  DEBUG_MESSAGE("*****************") ;
  printHeaders() ;
  int cpt = 0;
  for (vector<patVariables>::const_iterator it=samples.begin() ;
       it != samples.end() ;
       it++ ) {
    std::cout << "[" << cpt++ << "] " << *it << endl ;
  }
  std::cout << std::endl ;
}


void bioSample::addObservationData(patVariables data) {
  samples.push_back(data) ;
}


void bioSample::printHeaders() {
  for (vector<patString>::const_iterator it=headers.begin() ;
       it != headers.end();
       it++ ) {
     std::cout << *it << "    ";
  }
  std::cout << std::endl ;
}


void bioSample::computeMapOfDatabase(patError*& err) {
  map<patString,bioIteratorInfo> rep = 
    bioIteratorInfoRepository::the()->getRepository() ;
  map<patULong, patReal> currentValue ;
  
  for (map<patString,bioIteratorInfo>::iterator i = rep.begin() ;
       i != rep.end() ;
       ++i) {
    if (i->second.isRowIterator() || i->second.isMetaIterator()) {
      patULong colId = getIndexFromHeader(i->second.indexName) ;
      if (colId == patBadId) {
	stringstream str ;
	str << "Column " << i->second.indexName << " characterizing iterator " << i->first << " is unknown in file " << fileName ;
	err = new patErrMiscError(str.str()) ;
	WARNING(err->describe()) ;
	return ;
      }
      currentValue[colId] = patReal() ;
    }
  }

  for (patULong row = 0 ;
       row < samples.size() ;
       ++row) {
    if (row == 0) {
      for (map<patULong, patReal>::iterator col = currentValue.begin() ;
	   col != currentValue.end() ;
	   ++col) {
	bioIteratorInfoRepository::the()->addNewRowId(headers[col->first],0) ;
        col->second = samples[0][col->first] ;
      }
    }
    else {
      for (map<patULong, patReal>::iterator col = currentValue.begin() ;
	   col != currentValue.end() ;
	   ++col) {
	if (col->second != samples[row][col->first]) {
	  bioIteratorInfoRepository::the()->addNewRowId(headers[col->first],row) ;
	  col->second = samples[row][col->first] ;
	}
      }
    }
  }

  bioIteratorInfoRepository::the()->computeRowPointers(samples.size(),err) ;
  if (err != NULL) {
    WARNING(err->describe()) ;
    return ;
  }
}

bioMetaIterator* bioSample::createMetaIterator(bioIteratorSpan theSpan, bioIteratorSpan theThreadSpan, patError*& err) {
  
  //  DEBUG_MESSAGE("CREATE META ITERATOR FOR THREAD " << theThreadSpan) ;

  if (theSpan.lastRow == patBadId) {
    // theSpan.lastRow = size()-1 ;
    theSpan.lastRow = size() ;
  }
  bioMetaIterator* ptr = new bioMetaIterator(theSpan, theThreadSpan,err) ;
  if (err != NULL) {
    WARNING(err->describe()) ;
  }
  return ptr ;
}

bioRowIterator* bioSample::createRowIterator(bioIteratorSpan theSpan, bioIteratorSpan theThreadSpan, patError*& err) {
  
  if (theSpan.lastRow == patBadId) {
    // theSpan.lastRow = size()-1 ;
    theSpan.lastRow = size() ;
  }
  bioRowIterator* ptr = new bioRowIterator(&samples,theSpan,theThreadSpan,err) ;
  if (err != NULL) {
    WARNING(err->describe()) ;
    return NULL ;
  }
  return ptr ;
}

void bioSample::prepare(bioExpression* exclude, patError*& err) {
  DEBUG_MESSAGE("Prepare data...") ;
  if (exclude != NULL) {
    DEBUG_MESSAGE("Exclude expression: " << exclude->getExpression(err)) ;
  }
  readFile(exclude,err) ;
  if (err != NULL) {
    WARNING(err->describe()) ;
    return ;
  }
  computeMapOfDatabase(err) ;
  if (err != NULL) {
    WARNING(err->describe()) ;
    return ;
  }
  
}

void bioSample::generatePythonHeaders(patError*& err) {
  ofstream f("headers.py") ;
  f << "from biogeme import *" << endl ;
  for (vector<patString>::iterator i = headers.begin() ;
       i != headers.end() ;
       ++i) {
    f << *i << "=Variable('" << *i << "')" << endl ;
  }
  f.close() ;
  ofstream g("headersSimulation.py") ;
  for (unsigned long i = 0 ; i < headers.size() ; ++i) {
    g << headers[i] << " = currentRow[" << i << "]" << endl ;
  }
  g.close() ;
}

vector<patULong> bioSample::getListOfIds(patString header, patError*& err) {
  patULong indexId = getIndexFromHeader(header) ;
  if (indexId == patBadId) {
    stringstream str ;
    str << "Header " << header << " is unknown" ;
    err = new patErrMiscError(str.str()) ;
    WARNING(err->describe()) ;
    return vector<patULong>() ;
  }
  set<patULong> listOfUniqueIds ;
  for (vector<patVariables>::iterator row = samples.begin();
       row != samples.end() ;
       ++row) {
    listOfUniqueIds.insert(patULong((*row)[indexId])) ;
  }
  DETAILED_MESSAGE("Variable " << header << " has " << listOfUniqueIds.size() << " different values in the sample") ;
  vector<patULong> result ;
  for (set<patULong>::iterator i = listOfUniqueIds.begin() ;
       i != listOfUniqueIds.end() ;
       ++i) {
    result.push_back(*i) ;
  }
  return result ;
}

